/*  TODO:
    + clean up global Stack handling?
 */

#include "mace4.h"

#define MACE_NAME     "ICGNS"
#define MACE_VERSION  "2002b"
#define VERSION_DATE  "July 2002"

#define DEBUGno

Symbol_data Symbols;  /* List of symbols */

/* This maps OPS symbol IDs to MACE symbol IDs, which start with 0. */

int Sn_map_size;
Symbol_data *Sn_to_mace_sn;

/* Cell table is indexed by eterm IDs. */

int Number_of_cells;
struct cell *Cells;           /* the table of cells (dynamically allocated) */
struct cell **Ordered_cells;  /* (pointers to) permutation of Cells */

/* Misc */

int Max_domain_element_in_input;  /* For Least Number Heuristic */

int Domain_size;
Term *Domain;    /* array of terms representing (shared) domain elements  */

Glist Ground_clauses;

int Relation_flag;  /* term flag */
int Negation_flag;  /* term flag */

int Eq_sn;
int Or_sn;
int Not_sn;

/* ---------------------- Statistics ------------------------- */

/* stats for entire run */

unsigned All_models;

/* stats for the current domain size */

int Mace4_clock;

unsigned Current_models;
unsigned Selections;
unsigned Propagations;
unsigned Backups;
unsigned Cross_offs;
unsigned Rewrite_terms;
unsigned Rewrite_bools;
unsigned Indexes;
unsigned Ground_clauses_seen;
unsigned Ground_clauses_kept;
unsigned Rules_from_neg;

unsigned Neg_elim_attempts;
unsigned Neg_elim;
unsigned Neg_elim_agone;
unsigned Neg_elim_egone;

unsigned Neg_assign_attempts;
unsigned Neg_assign;
unsigned Neg_assign_agone;
unsigned Neg_assign_egone;

unsigned Neg_near_assign_attempts;
unsigned Neg_near_assign;
unsigned Neg_near_assign_agone;
unsigned Neg_near_assign_egone;

unsigned Neg_near_elim_attempts;
unsigned Neg_near_elim;
unsigned Neg_near_elim_agone;
unsigned Neg_near_elim_egone;

/* Ground terms.  MACE4 operates on ground clauses, which are
   represented by the structure mclause.  Ground terms (and
   atoms) are represented with ordinary OPS terms.  There are
   a few tricks:

   (1) We use upward pointers from terms to superterms
   and from atoms to clauses.
 
   (2) We need to mark atoms with a termflag, so that we know
   when to stop when following the upward pointers.  Also,
   a termflag is used to indicate that an atom is negated.

   (3) Domain elements are represented by variables (sorry,
   but it is very convenient to do so).  Also, there is only
   one actual copy of each domain element (structure sharing
   of domain elements).  Global array *Domain contains them.

   IDs.  If all of the arguments of a term (including atoms) are
   domain elements, that term is called an eterm.  For example,
   f(3,4), a, P(0), and Q.

   Each eterm has a unique ID which is used as an index into
   the cell table, for example, when a new eterm is obtained by
   evaluating a subterm to a domain element, we have to quickly
   check if this new eterm can be evaluated.  We do this by
   calculating its ID and looking up in Cells[ID].value.
   And when we have a new assignment, say f(3,4)=2, we find the
   list of occurrences of f(3,4) by looking in Cells[ID].occurrences.
*/

/*************
 *
 *   overall_init()
 *
 *   This is the initialization that has to be done only once
 *   for a given set of clauses.  It is independent of the
 *   domain size.
 *
 *************/

static
void overall_init(Clist clauses)
{
  Clist_pos p;
  int max, i;
  Symbol_data s;

  /* In ground clauses, VARIABLEs represent domain elements,
     so from here on, print variables as integers. */

  set_variable_style(INTEGER_STYLE);

  /* These flags are for ground clause (mclause) literals. */

  Relation_flag = claim_term_flag();
  Negation_flag = claim_term_flag();

  /* Cache some symbol numbers. */

  Eq_sn  = str_to_sn("=", 2);
  Or_sn  = str_to_sn("|", 2);
  Not_sn = str_to_sn("~", 1);

  /* Set up Symbols list. */

  init_built_in_symbols();  /* =/2 (and maybe others) */

  /* Collect data for each symbol. */

  Max_domain_element_in_input = -1;
  for (p = clauses->first; p != NULL; p = p->next) {
    Literal lit;
    for (lit = p->c->first_literal; lit != NULL; lit = lit->next) {
      int x = collect_mace_syms(lit->atom, RELATION);
      Max_domain_element_in_input = MAX(Max_domain_element_in_input, x);
    }
  }

  printf("\nMaximum domain element in input is %d.\n",
	 Max_domain_element_in_input);

  /* Set up map from ordinary symnums to mace symnums. */

  max = 0;
  i = 0;

  for (s = Symbols; s != NULL; s = s->next) {
    s->mace_sn = i++;
    max = (s->sn > max ? s->sn : max);
  }

  Sn_map_size = max+1;

  Sn_to_mace_sn = malloc(Sn_map_size * sizeof(void *));

  for (i = 0; i < Sn_map_size; i++)
    Sn_to_mace_sn[i] = NULL;

  for (s = Symbols; s != NULL; s = s->next) {
    Sn_to_mace_sn[s->sn] = s;
  }
}  /* overall_init */

/*************
 *
 *   init_for_domain_size()
 *
 *   Given the list of (general) clauses, set up the various data
 *   structures that will be needed for a given domain size.
 *
 *************/

static
void init_for_domain_size(Clist clauses)
{
  int i, j, nextbase, id;
  Symbol_data s;

  /* Give each symbol its "base" value, which is used to index cells.  */

  nextbase = 0;
  for (s = Symbols; s != NULL; s = s->next) {
    s->base = nextbase;
    nextbase += int_power(Domain_size, s->arity);
  }

  /* Set up the array of domain terms.  All ground terms refer to these. */

  Domain = malloc(Domain_size * sizeof(void *));
  for (i = 0; i < Domain_size; i++)
    Domain[i] = get_variable_term(i);
  
  /* Set up the table of cells. */

  Number_of_cells = nextbase;
  Cells           = malloc(Number_of_cells * sizeof(struct cell));
  Ordered_cells   = malloc(Number_of_cells * sizeof(void *));

  for (id = 0; id < Number_of_cells; id++) {
    struct cell *c = Cells + id;
    int n;
    c->id = id;
    c->occurrences = NULL;
    c->value = NULL;
    c->symbol = find_symbol_node(id);
    c->eterm = decode_eterm_id(id);
    c->max_index = max_index(id, c->symbol);
    n = id_to_domain_size(id);
    c->possible = malloc(n * sizeof(void *));
    for (j = 0; j < n; j++)
      c->possible[j] = Domain[j];  /* really just a flag */
  }

  order_cells();

  if (flag(NEGPROP))
    init_negprop_index();

} /* init_for_domain_size */

/*************
 *
 *   built_in_assignments()
 *
 *************/

static
void built_in_assignments(void)
{
  Symbol_data s;
  for (s = Symbols; s != NULL; s = s->next) {
    if (s->attribute == EQUALITY_SYMBOL) {
      int i, j;
      for (i = 0; i < Domain_size; i++)
	for (j = 0; j < Domain_size; j++)
          Cells[X2(s->base,i,j)].value = (Domain[i==j ? 1 : 0]);
    }
  }
}  /* built_in_assignments */

/*************
 *
 *   check_that_ground_clauses_are_true()
 *
 *************/

static
void check_that_ground_clauses_are_true(void)
{
  Glist g;
  for (g = Ground_clauses; g != NULL; g = g->next) {
    Mclause c = g->u.v;
    if (!c->subsumed) {
      printf("ERROR, model reported, but clause not true! ");
      p_mclause(c);
      fprintf(stderr, "ERROR, model reported, but clause not true!\n");
    }
  }
}  /* check_that_ground_clauses_are_true */

/*************
 *
 *   we_have_a_model()
 *
 *************/

static
void we_have_a_model(void)
{
  static int next_message = 1;
  Current_models++;
  All_models++;
  check_that_ground_clauses_are_true();  /* in case of a soundness bug */
  if (flag(PRINT_MODELS_PORTABLE))
    print_model_portable(stdout, TRUE);
  else if (flag(PRINT_MODELS))
    p_model(TRUE);

  else if (next_message == All_models) {
    printf("\nModel %d has been found.\n", All_models);
    next_message *= 10;
  }
  fflush(stdout);
  if (All_models >= parm(MAX_MODELS))
    mace4_exit(MAX_MODELS_EXIT);
}  /* we_have_a_model */

/*************
 *
 *   check_time()
 *
 *************/

static
void check_time(void)
{
  if (parm(MAX_SECONDS) < INT_MAX && user_time() > parm(MAX_SECONDS) * 1000)
    mace4_exit(MAX_SECONDS_EXIT);
}  /* check_time */

/*************
 *
 *   p_possible_values()
 *
 *************/

void p_possible_values(void)
{
  int i;
  for (i = 0; i < Number_of_cells; i++) {
    if (Cells[i].symbol->attribute == ORDINARY_SYMBOL) {
      int j;
      printf("Cell %d: ", i);
      for (j = 0; j < id_to_domain_size(i); j++) {
	if (Cells[i].possible[j] != NULL)
	  printf(" %d", j);
      }
      printf("\n");
    }
  }
}  /* p_possible_values */

/*************
 *
 *   search()
 *
 *   Max_constrained is the maximum constrained domain element
 *   (or -1 is none is constrained).  Greater domain elements
 *   can all be considered symmetric.  An element can become
 *   constrained in two ways:  (1) it is an index of some selected
 *   cell, or (2) it is the value assigned to some selected cell.
 *   (Propagation does not constrain elements.  This might need
 *   careful justification.)
 *
 *   To apply the least number heuristic, we consider values
 *   0 ... MIN(max_constrained+1, Domain_size-1).
 *
 *   To make this effective, we should keep max_constrained as low as
 *   possible by selecting cells with maximum index <= max_constrained.
 *
 *************/

static
void search(int max_constrained, int depth)
{
  int id;
  check_time();
  id = select_cell(max_constrained);
  if (id == -1)
    we_have_a_model();
  else {
    int i, last;  /* we'll do 0 .. last */
    int x = Cells[id].max_index;
    max_constrained = MAX(max_constrained, x);
    Selections++;

    if (flag(TRACE)) {
      printf("select: ");
      p_model(FALSE);
      /* p_possible_values(); */
    }
	
    if (Cells[id].symbol->type == RELATION)
      last = 1;
    else if (flag(LNH))
      last = MIN(max_constrained+1, Domain_size-1);
    else
      last = Domain_size-1;

    for (i = 0; i <= last; i++) {
      Estack stk;

      if (flag(TRACE)) {
	printf("assign: ");
	fwrite_term(stdout, Cells[id].eterm);
	printf("=%d (%d) depth=%d\n", i, last, depth);
      }
	
      stk = assign_and_propagate(id, Domain[i]);

      if (stk == NULL)
	Backups++;  /* this is only a statistic */
      else {
	search(MAX(max_constrained, i), depth+1);
	restore_from_stack(stk);
      }
    }
  }
}  /* search */

/*************
 *
 *   mace4() -- look for a model of a specific size
 *
 *************/

static
void mace4(Clist clauses, int order)
{
  Clist_pos p;
  Glist g;
  int i;
  Mstate initial_state = get_mstate();

  if (Max_domain_element_in_input >= order)
    fatal_error("mace4: domain_element too big");

  Domain_size = order;

  init_for_domain_size(clauses);

  built_in_assignments();  /* Fill out equality table (and maybe others). */

  /* Instantiate clauses over the domain.  This also 
     (1) makes any domain element constants into real domain elements,
     (2) applies OR, NOT, and EQ simplification, and
     (3) does unit propagation (which pushes events onto initial_state->stack).
     Do the units first, then the 2-clauses, then the rest. */

  for (p = clauses->first; initial_state->ok && p != NULL; p = p->next)
    if (number_of_literals(p->c) < 2)
      generate_ground_clauses(p->c, initial_state);

  for (p = clauses->first; initial_state->ok && p != NULL; p = p->next)
    if (number_of_literals(p->c) == 2)
      generate_ground_clauses(p->c, initial_state);

  for (p = clauses->first; initial_state->ok && p != NULL; p = p->next)
    if (number_of_literals(p->c) > 2)
      generate_ground_clauses(p->c, initial_state);

  /* The preceding calls push propagation events onto initial_state->stack.
     We won't have to undo those initial events during the search,
     but we can undo them after the search.
  */

  printf("\nInitial partial model:\n");
  p_model(FALSE);
  fflush(stdout);

  /* Here we go! */

  if (initial_state->ok)
    search(Max_domain_element_in_input, 0);

  /* OK, we're done.  (We may have exited from within search.) */

  /* Free all of the memory associated with the current domain size. */

  restore_from_stack(initial_state->stack);
  free_mstate(initial_state);

  if (flag(NEGPROP))
    free_negprop_index();

  free(Ordered_cells);
  Ordered_cells = NULL;

  for (i = 0; i < Number_of_cells; i++) {
    zap_mterm(Cells[i].eterm);
    free(Cells[i].possible);
  }
  free(Cells);
  Cells = NULL;

  for (i = 0; i < Domain_size; i++)
    zap_term(Domain[i]);
  free(Domain);
  Domain = NULL;

  for (g = Ground_clauses; g != NULL; g = g->next)
    zap_mclause(g->u.v);
  zap_glist(Ground_clauses);
  Ground_clauses = NULL;

}  /* mace4 */

/*************
 *
 *    void print_banner(argc, argv)
 *
 *************/

static
void print_banner(int argc, char **argv)
{
  int i;
  int pid = 0;
    
  printf("----- %s %s, %s -----\n", MACE_NAME, MACE_VERSION, VERSION_DATE);
# if 0
  printf("The process was started by %s on %s,\n%s",
	 username(), hostname(), get_time());
  pid = my_process_id();
#endif

  printf("The command was \"");
  for(i = 0; i < argc; i++)
    printf("%s%s", argv[i], (i < argc-1 ? " " : ""));
  printf("\".");
  if (pid != 0)
    printf("  The process ID is %d.\n\n", pid);
  else
    printf("\n\n");
    
}  /* print_banner */

/*************
 *
 *   read_the_input_file()
 *
 *************/

static
Clist read_the_input_file(char *filename)
{
  FILE *fp;
  Clist clauses;
  Term t;

  fp = fopen(filename, "r");
  if (fp == NULL)
    fatal_error("file can't be opened for reading");

  t = read_commands(fp, stdout, TRUE);  /* fatal_error() on error */

  if (t == NULL || !is_symbol(SYMNUM(t), "end_of_list", 0))
    fatal_error("commands must be termintated with end_of_list");
  else
    zap_term(t);

  /* Read list of clauses. */

  if (flag(PROLOG_STYLE_VARIABLES))
    set_variable_style(PROLOG_STYLE);
  
  clauses = read_clause_list(fp, stderr, "input_clauses", FALSE);

  fclose(fp);

  return clauses;
}  /* read_the_input_file */

/*************
 *
 *   mace4_exit()
 *
 *************/

void mace4_exit(int exit_code)
{
  char *m = (All_models == 1 ? "model" : "models");

  p_mem();

  if (exit_code == MAX_MODELS_EXIT || exit_code == MAX_SECONDS_EXIT)
    p_stats();

  switch(exit_code) {
  case MAX_MODELS_EXIT:
    fprintf(stdout, "\n----- MAX_MODELS EXIT (%u %s) -----\n", All_models, m);
    fprintf(stderr, "\n----- MAX_MODELS EXIT (%u %s) -----\n", All_models, m);
    break;
  case ALL_MODELS_EXIT:
    fprintf(stdout, "\n----- ALL_MODELS EXIT (%u %s) -----\n", All_models, m);
    fprintf(stderr, "\n----- ALL_MODELS EXIT (%u %s) -----\n", All_models, m);
    break;
  case EXHAUSTED_EXIT:
    fprintf(stdout, "\n----- exhausted exit -----\n");
    fprintf(stderr, "\n----- exhausted exit -----\n");
    break;
  case MAX_SECONDS_EXIT:
    fprintf(stdout, "\n----- max_seconds exit -----\n");
    fprintf(stderr, "\n----- max_seconds exit -----\n");
    break;
  default:
    fprintf(stdout, "\n----- unknown exit code: %d -----\n", exit_code);
    fprintf(stderr, "\n----- unknown exit code: %d -----\n", exit_code);
    break;
  }
  exit(exit_code);
}  /* mace4_exit */

/*************
 *
 *   main()
 *
 *************/

int main(int argc, char **argv)
{
  Clist clauses;
  int domain_size_from_arg = -1;
  int i, end;

  print_banner(argc, argv);

  if (argc < 2) {
    fprintf(stderr, "need 1 or 2 args: filename [domain-size]\n");
    exit(1);
  }
  else if (argc == 3) {
    /* Domain_size argument overrides domain_size flag in input file. */
    if (!str_to_int(argv[2], &domain_size_from_arg) ||
	domain_size_from_arg < 2)
      fatal_error("second argument (domain_size) must be integer >= 2");
  }
  
  set_max_megs(192);

  declare_standard_parse_types();
  translate_neg_equalities(TRUE);

  init_options();

  init_parm(DOMAIN_SIZE,       "domain_size",       2,       2, 100);
  init_parm(ITERATE_UP_TO,     "iterate_up_to",     0,       0, 100);
  init_parm(MAX_MODELS,        "max_models",        1,       1, INT_MAX);
  init_parm(MAX_SECONDS,       "max_seconds",       INT_MAX, 1, INT_MAX);
  init_parm(SELECTION_ORDER,   "selection_order",   2, 0, 2); /* see select.c */
  init_parm(SELECTION_MEASURE, "selection_measure", 4, 0, 4); /* see select.c */
         
  init_flag(PRINT_MODELS,            "print_models",            TRUE);
  init_flag(PRINT_MODELS_PORTABLE,   "print_models_portable",   FALSE);
  init_flag(LNH,                     "lnh",                     TRUE);
  init_flag(TRACE,                   "trace",                   FALSE);
  init_flag(NEGPROP,                 "negprop",                 TRUE);
  init_flag(NEG_ASSIGN,            "neg_assign",            TRUE);
  init_flag(NEG_ASSIGN_NEAR,            "neg_assign_near",            TRUE);
  init_flag(NEG_ELIM,              "neg_elim",              TRUE);
  init_flag(NEG_ELIM_NEAR,              "neg_elim_near",              TRUE);
  init_flag(PROLOG_STYLE_VARIABLES,  "prolog_style_variables",  FALSE);

  Mace4_clock = clock_init("Mace4");

  clauses = read_the_input_file(argv[1]);
  fwrite_clause_list(stdout, clauses, FALSE);

  overall_init(clauses);

  if (domain_size_from_arg >= 2)
    assign_parm(DOMAIN_SIZE, domain_size_from_arg);

  end = (parm(ITERATE_UP_TO) == 0 ? parm(DOMAIN_SIZE) : parm(ITERATE_UP_TO));

  for (i = parm(DOMAIN_SIZE); i <= end ; i++) {
    fprintf(stdout,"\n======= Starting search for domain size %d. =======\n",i);
    fprintf(stderr,"\n======= Starting search for domain size %d. =======\n",i);
    clock_start(Mace4_clock);
    mace4(clauses, i);
    clock_stop(Mace4_clock);
    p_stats();
    reset_current_stats();
    clock_reset(Mace4_clock);
  }

  clist_zap(clauses);

  /* Note that max_models and max_seconds exits occur within mace4(). */

  mace4_exit(All_models == 0 ? EXHAUSTED_EXIT : ALL_MODELS_EXIT);

  exit(0);  /* to please the compiler */

}  /* main */

