#include "clause.h"

/* Private definitions and types */

#define CSTACK_ARRAY_SIZE    100  /* see use below */

/*
 * memory management
 */

static Literal literal_avail;
static unsigned literal_gets, literal_frees, literal_avails;

static Clause clause_avail;
static unsigned clause_gets, clause_frees, clause_avails;

/* end of counters for each type */

/*************
 *
 *    literal get_literal()
 *
 *************/

static
Literal get_literal(void)
{
  Literal p;

  literal_gets++;
  if (literal_avail == NULL)
    p = tp_alloc(sizeof(struct literal));
  else {
    literal_avails--;
    p = literal_avail;
    literal_avail = literal_avail->next;
  }
  /* initialization */	
  	
  p->next = NULL;
  p->atom = NULL;

  /* end of initialization */	
  return(p);
}  /* get_literal */

/*************
 *
 *    free_literal()
 *
 *************/

static
void free_literal(Literal p)
{
  literal_frees++;
  literal_avails++;
  p->next = literal_avail;
  literal_avail = p;
}  /* free_literal */

/*************
 *
 *    Clause get_clause()
 *
 *************/

/* DOCUMENTATION
This routine allocates and returns an empty clause.
*/

/* PUBLIC */
Clause get_clause(void)
{
  Clause p;

  clause_gets++;
  if (clause_avail == NULL)
    p = tp_alloc(sizeof(struct clause));
  else {
    clause_avails--;
    p = clause_avail;
    clause_avail = (Clause) clause_avail->first_literal;
  }
  /* initialization */	
  	
  p->id = 0;
  p->first_literal = NULL;
  p->justification = NULL;
  p->containers = NULL;

  /* end of initialization */	
  return(p);
}  /* get_clause */

/*************
 *
 *    free_clause()
 *
 *************/

static
void free_clause(Clause p)
{
  clause_frees++;
  clause_avails++;
  p->first_literal = (Literal) clause_avail;
  clause_avail = p;
}  /* free_clause */

/* end of get and free routines for each type */

/*************
 *
 *   fprint_clause_mem()
 *
 *************/

/* DOCUMENTATION
This routine prints (to FILE *fp) memory usage statistics for data types
associated clauses.
The Boolean argument heading tells whether to print a heading on the table.
*/

/* PUBLIC */
void fprint_clause_mem(FILE *fp, BOOL heading)
{
  int n;
  if (heading)
    fprintf(fp, "  type (bytes each)        gets      frees     in use      avail      bytes\n");

  n = sizeof(struct literal);
  fprintf(fp, "literal (%4d)      %11u%11u%11u%11u%9.1f K\n",
	  n, literal_gets, literal_frees,
	  literal_gets - literal_frees, literal_avails,
	  (((literal_gets - literal_frees) + literal_avails) * n) / 1024.);

  n = sizeof(struct clause);
  fprintf(fp, "clause (%4d)       %11u%11u%11u%11u%9.1f K\n",
	  n, clause_gets, clause_frees,
	  clause_gets - clause_frees, clause_avails,
	  (((clause_gets - clause_frees) + clause_avails) * n) / 1024.);

  /* end of printing for each type */
  
}  /* fprint_clause_mem */

/*************
 *
 *   p_clause_mem()
 *
 *************/

/* DOCUMENTATION
This routine prints (to stdout) memory usage statistics for data types
associated with clauses.
*/

/* PUBLIC */
void p_clause_mem()
{
  fprint_clause_mem(stdout, 1);
}  /* p_clause_mem */

/*
 *  end of memory management
 */

/*************
 *
 *    zap_clause(c)
 *
 *************/

/* DOCUMENTATION
This routine frees a clause (but not any justification list or
attributes).  The caller should make sure that nothing (e.g., indexes)
refer to the clause or any of its subterms.
*/

/* PUBLIC */
void zap_clause(Clause c)
{
  Literal l1, l2;

  l1 = c->first_literal;
  while(l1 != NULL) {
    zap_term(l1->atom);
    l2 = l1;
    l1 = l1->next;
    free_literal(l2);
  }
  free_clause(c);
}  /* zap_clause */

/*************
 *
 *   fprint_clause()
 *
 *************/

/* DOCUMENTATION
This routine prints a clause to a file.
*/

/* PUBLIC */
void fprint_clause(FILE *fp, Clause c)
{
  Literal lit;

  if (c->id > 0)
    fprintf(fp, "%d: ", c->id);

  if (c->first_literal == NULL)
    fprintf(fp, "%s", FALSE_SYM);
  else {
    for (lit = c->first_literal; lit != NULL; lit = lit->next) {
      if (!lit->sign)
	fprintf(fp, "%s", NOT_SYM);
      fprint_term(fp, lit->atom);
      if (lit->next != NULL)
	fprintf(fp, " %s ", OR_SYM);
    }
  }
  fprintf(fp, ".\n");
  fflush(fp);
}  /* fprint_clause */

/*************
 *
 *   p_clause()
 *
 *************/

/* DOCUMENTATION
This routine prints a clause to stdout.
*/

/* PUBLIC */
void p_clause(Clause c)
{
  fprint_clause(stdout, c);
}  /* p_clause */

/*************
 *
 *   new_literal()
 *
 *************/

/* DOCUMENTATION
This routine takes a sign (Boolean) and a Term atom, and returns
a literal.  The atom is not copied.
*/

/* PUBLIC */
Literal new_literal(int sign, Term atom)
{
  Literal lit = get_literal();
  lit->sign = sign;
  lit->atom = atom;
  return lit;
}  /* new_literal */

/*************
 *
 *   append_literal()
 *
 *************/

/* DOCUMENTATION
This routine appends a literal to a clause.
*/

/* PUBLIC */
void append_literal(Clause c, Literal lit)
{
  Literal l = c->first_literal;

  if (l == NULL) {
    c->first_literal = lit;
  }
  else {
    while (l->next != NULL)
      l = l->next;
    l->next = lit;
  }
  lit->next = NULL;
}  /* append_literal */

/*************
 *
 *   clause_length()
 *
 *************/

/* DOCUMENTATION
This routine returns the length of a clause, which is the sum of
the lengths of its atoms.
*/

/* PUBLIC */
int clause_length(Clause c)
{
  int n = 0;
  Literal lit;
  for (lit = c->first_literal; lit != NULL; lit = lit->next)
    n += symbol_count(lit->atom);
  return n;
}  /* clause_length */

/*************
 *
 *   term_to_literals()
 *
 *************/

static
Literal term_to_literals(Term t, Literal lits)
{
  Literal l;

  if (COMPLEX(t) && is_symbol(SYMNUM(t), OR_SYM, 2)) {
    /* Traverse term right-to-left and add to the
     * front of the clause, so order is preserved.
     */
    l = term_to_literals(t->args[1], lits);
    l = term_to_literals(t->args[0], l);
  }
  else {
    l = get_literal();
    l->next = lits;
    l->sign = !(COMPLEX(t) && is_symbol(SYMNUM(t), NOT_SYM, 1));
    if (l->sign)
      l->atom = copy_term(t);
    else
      l->atom = copy_term(t->args[0]);
  }
  return(l);
}  /* term_to_literals */

/*************
 *
 *   term_to_clause()
 *
 *************/

/* DOCUMENTATION
This routine takes a Term t (presumably a disjunction with binary
symbol OR_SYM), and constructs a Clause.  The Clause is entirely new.
<P>
The main use of this routine is intended to be as follows: a
Term representing a clause is parsed (using mixfix notation)
from the input, then here it is copied translated into
a Clause data structure.
*/

/* PUBLIC */
Clause term_to_clause(Term t)
{
  Clause c = get_clause();
  c->first_literal = term_to_literals(t, NULL);
  return(c);
}  /* term_to_clause */

/*************
 *
 *   literals_to_term()
 *
 *************/

static
Term literals_to_term(Literal l)
{
  Term t;
  if (l->sign)
    t = copy_term(l->atom);
  else {
    t = get_rigid_term(NOT_SYM, 1);
    t->args[0] = copy_term(l->atom);
  }
  if (l->next) {
    Term d = get_rigid_term(OR_SYM, 2);
    d->args[0] = t;
    d->args[1] = literals_to_term(l->next);
    return d;
  }
  else
    return t;
}  /* literals_to_term */

/*************
 *
 *   clause_to_term()
 *
 *************/

/* DOCUMENTATION
This routine takes a Clause and returns an entirely new Term
which represents the clause.  The disjunction symbol for the
term is binary OR_SYM, and the negation symbols is NOT_SYM.
*/

/* PUBLIC */
Term clause_to_term(Clause c)
{
  if (c->first_literal != NULL)
    return literals_to_term(c->first_literal);
  else {
    return get_rigid_term("$F", 0);
  }
}  /* clause_to_term */

/*************
 *
 *   literals_to_term_x()
 *
 *************/

static
Term literals_to_term_x(Literal l)
{
  Term t;
  if (l->sign)
    t = l->atom;
  else {
    t = get_rigid_term(NOT_SYM, 1);
    t->args[0] = l->atom;
  }
  if (l->next) {
    Term d = get_rigid_term(OR_SYM, 2);
    d->args[0] = t;
    d->args[1] = literals_to_term_x(l->next);
    t = d;
  }
  free_literal(l);
  return t;
}  /* literals_to_term_x */

/*************
 *
 *   clause_to_term_x()
 *
 *************/

/* DOCUMENTATION
This routine takes a Clause and returns its Term form.
The disjunction symbol for the
term is binary OR_SYM, and the negation symbols is NOT_SYM.
<P>
This version destroys the clause.
*/

/* PUBLIC */
Term clause_to_term_x(Clause c)
{
  Term t;
  if (c->first_literal != NULL)
    t = literals_to_term_x(c->first_literal);
  else
    t = get_rigid_term("$F", 0);
  free_clause(c);  /* literals are already freed */
  return t;
}  /* clause_to_term_x */

/*************
 *
 *   trm_set_vars_recurse()
 *
 *************/

static
Term trm_set_vars_recurse(Term t, char **varnames, int max_vars, BOOL *rcp)
{
  if (CONSTANT(t)) {
    char *name = sn_to_str(SYMNUM(t));
    if (variable_name(name)) {
      int i = 0;
      while (i < max_vars && varnames[i] != NULL && varnames[i] != name)
	i++;
      if (i == max_vars) 
	*rcp = FALSE;
      else {
	if (varnames[i] == NULL)
	  varnames[i] = name;
	free_term(t);
	t = get_variable_term(i);
      }
    }
  }
  else {
    int i;
    for (i = 0; i < t->arity && *rcp; i++)
      t->args[i] = trm_set_vars_recurse(t->args[i], varnames, max_vars, rcp);
  }
  return t;
}  /* trm_set_vars_recurse */

/*************
 *
 *   clause_set_variables()
 *
 *************/

/* DOCUMENTATION
This routine traverses a clause and changes the constants
that should be variables, into variables.  On input, the clause
should have no variables.  The new variables are numbered
0, 1, 2 ... according the the first occurrence, reading from the
left.
<P>
If there are not more than max_vars distinct variables,
1 is returned;  if there are more than max_vars variables,
the transformation is not completed, and 0 is returned.
<P>
The intended is use is for input clauses that
are built without regard to variable/constant distinction.
*/

/* PUBLIC */
BOOL clause_set_variables(Clause c, int max_vars)
{
  char *a[CSTACK_ARRAY_SIZE], **vmap;
  int i;
  Literal lit;
  BOOL ok = TRUE;

  if (max_vars > CSTACK_ARRAY_SIZE)
    vmap = my_malloc((max_vars * sizeof(char *)));
  else
    vmap = a;

  for (i = 0; i < max_vars; i++)
    vmap[i] = NULL;

  for (lit = c->first_literal; lit != NULL && ok; lit = lit->next) {
    Term a = lit->atom;
    for (i = 0; i < a->arity && ok; i++) 
      a->args[i] = trm_set_vars_recurse(a->args[i], vmap, max_vars, &ok);
  }

  if (max_vars > CSTACK_ARRAY_SIZE)
    my_free(vmap);
  return ok;
}  /* clause_set_variables */

/*************
 *
 *   term_renumber_variables()
 *
 *************/

static
BOOL term_renumber_variables(Term t, int vmap[], int max_vars)
{
  if (VARIABLE(t)) {
    int i = 0;
    while (i < max_vars && vmap[i] != -1 && vmap[i] != VARNUM(t))
      i++;
    if (i == max_vars) 
      return FALSE;
    else {
      if (vmap[i] == -1)
	vmap[i] = VARNUM(t);
      set_variable(t, i);
      return TRUE;
    }
  }
  else {
    int i;
    BOOL ok = TRUE;
    for (i = 0; i < t->arity && ok; i++)
      ok = term_renumber_variables(t->args[i], vmap, max_vars);
    return ok;
  }
}  /* term_renumber_variables */

/*************
 *
 *   renumber_variables()
 *
 *************/

/* DOCUMENTATION
This routine renumbers the variables of a clause.  The variables are
renumbered 0, 1, 2 ... according the the first occurrence, reading
from the left.
<P>
If there are not more than max_vars distinct variables,
1 is returned;  if there are more than max_vars variables,
the transformation is not completed, and 0 is returned.
<P>
The intended is use is for inferred clauses that
may contain variable indexes greater than max_vars.
*/

/* PUBLIC */
BOOL renumber_variables(Clause c, int max_vars)
{
  int a[CSTACK_ARRAY_SIZE], *vmap;
  int i;
  Literal lit;
  BOOL ok = TRUE;

  if (max_vars > CSTACK_ARRAY_SIZE)
    vmap = my_malloc((max_vars * sizeof(int)));
  else
    vmap = a;

  for (i = 0; i < max_vars; i++)
    a[i] = -1;

  for (lit = c->first_literal; lit != NULL && ok; lit = lit->next) {
    for (i = 0; i < lit->atom->arity && ok; i++) 
      ok = term_renumber_variables(lit->atom->args[i], vmap, max_vars);
  }

  if (max_vars > CSTACK_ARRAY_SIZE)
    my_free(vmap);
  return ok;
}  /* renumber_variables */

/*************
 *
 *   positive_literals()
 *
 *************/

/* DOCUMENTATION
This function returns the number of positive literals in a clause.
*/

/* PUBLIC */
int positive_literals(Clause c)
{
  Literal lit;
  int n = 0;
  for (lit = c->first_literal; lit != NULL; lit = lit->next)
    if (lit->sign)
      n++;
  return n;
}  /* positive_literals */

/*************
 *
 *   negative_literals()
 *
 *************/

/* DOCUMENTATION
This function returns the number of negative literals in a clause.
*/

/* PUBLIC */
int negative_literals(Clause c)
{
  Literal lit;
  int n = 0;
  for (lit = c->first_literal; lit != NULL; lit = lit->next)
    if (!lit->sign)
      n++;
  return n;
}  /* negative_literals */

/*************
 *
 *   positive_clause()
 *
 *************/

/* DOCUMENTATION
This function checks if all of the literals of a clause are positive.
*/

/* PUBLIC */
BOOL positive_clause(Clause c)
{
  return negative_literals(c) == 0;
}  /* positive_clause */

/*************
 *
 *   any_clause()
 *
 *************/

/* DOCUMENTATION
This function is always TRUE.  (It it intended to be used as an argument.)
*/

/* PUBLIC */
BOOL any_clause(Clause c)
{
  return TRUE;
}  /* positive_clause */

/*************
 *
 *   negative_clause()
 *
 *************/

/* DOCUMENTATION
This function checks if all of the literals of a clause are negative.
*/

/* PUBLIC */
BOOL negative_clause(Clause c)
{
  return positive_literals(c) == 0;
}  /* negative_clause */

/*************
 *
 *   mixed_clause()
 *
 *************/

/* DOCUMENTATION
This function checks if a clause has at least one positive and
at least one negative literal.
*/

/* PUBLIC */
BOOL mixed_clause(Clause c)
{
  return (positive_literals(c) >= 1 && negative_literals(c) >= 1);
}  /* mixed_clause */

/*************
 *
 *   number_of_literals()
 *
 *************/

/* DOCUMENTATION
This function returns the number of literals in a clause.
*/

/* PUBLIC */
int number_of_literals(Clause c)
{
  return positive_literals(c) + negative_literals(c);
}  /* number_of_literals */

/*************
 *
 *   unit_clause()
 *
 *************/

/* DOCUMENTATION
This function checks if a clause has exactly one literal.
*/

/* PUBLIC */
BOOL unit_clause(Clause c)
{
  return c->first_literal != NULL && c->first_literal->next == NULL;
}  /* unit_clause */

/*************
 *
 *   horn_clause()
 *
 *************/

/* DOCUMENTATION
This function checks if a clause has at most one positive literal.   
*/

/* PUBLIC */
BOOL horn_clause(Clause c)
{
  return positive_literals(c) <= 1;
}  /* horn_clause */

/*************
 *
 *   definite_clause()
 *
 *************/

/* DOCUMENTATION
This Boolean function checks if a clause has exactly one positive literal.   
*/

/* PUBLIC */
BOOL definite_clause(Clause c)
{
  return positive_literals(c) == 1;
}  /* definite_clause */

/*************
 *
 *   greatest_variable_in_clause(c)
 *
 *************/

/* DOCUMENTATION
This routine returns the greatest variable index in a clause.
If the clause is ground, -1 is returned.
*/

/* PUBLIC */
int greatest_variable_in_clause(Clause c)
{
  Literal lit;
  int max, v;
    
  for (max = -1, lit = c->first_literal; lit != NULL; lit = lit->next) {
    v = greatest_variable(lit->atom);
    max = (v > max ? v : max);
  }
  return(max);
}  /* greatest_variable_in_clause */

/*************
 *
 *   copy_clause()
 *
 *************/

/* DOCUMENTATION
This routine builds and returns a copy of a clause.
*/

/* PUBLIC */
Clause copy_clause(Clause c)
{
  Literal l;
  Clause c2 = get_clause();

  for (l = c->first_literal; l != NULL; l = l->next) {
    Literal l2 = get_literal();
    l2->sign = l->sign;
    l2->atom = copy_term(l->atom);
    append_literal(c2, l2);
  }
  return c2;
}  /* copy_clause */

/*************
 *
 *   upward_clause_links()
 *
 *************/

/* DOCUMENTATION
In the given Clause c, make the "container" field of each subterm
point to c.
*/

/* PUBLIC */
void upward_clause_links(Clause c)
{
  Literal lit;
  for (lit = c->first_literal; lit != NULL; lit = lit->next)
    upward_term_links(lit->atom, c);
}  /* upward_clause_links */

/*************
 *
 *   literal_number()
 *
 *************/

/* DOCUMENTATION
Given a clause and a literal, return the position of the literal
(counting from 1) in the clause.  The check is by pointer only.
If the literal does not occur in the clause, 0 is returned.
*/

/* PUBLIC */
int literal_number(Clause c, Literal l)
{
  int n = 1;
  Literal lit = c->first_literal;

  while (lit != NULL && lit != l) {
    n++;
    lit = lit->next;
  }
  return (lit == NULL ? 0 : n);
}  /* literal_number */

/*************
 *
 *   atom_number()
 *
 *************/

/* DOCUMENTATION
Given a clause and an atom, return the position of the atom
(counting from 1) in the clause.  The check is by pointer only.
If the atom does not occur in the clause, 0 is returned.
*/

/* PUBLIC */
int atom_number(Clause c, Term atom)
{
  int n = 1;
  Literal lit = c->first_literal;

  while (lit != NULL && lit->atom != atom) {
    n++;
    lit = lit->next;
  }
  return (lit == NULL ? 0 : n);
}  /* atom_number */

/*************
 *
 *   ith_literal()
 *
 *************/

/* DOCUMENTATION
Return the i-th literal of a clause, counting from 1.
Return NULL if i is out of range.
*/

/* PUBLIC */
Literal ith_literal(Clause c, int i)
{
  Literal lit = c->first_literal;
  int n = 1;
  while (lit != NULL && n < i) {
    lit = lit->next;
    n++;
  }
  return lit;
}  /* ith_literal */
