#include "clauseid.h"

/* Private definitions and types */

#define CLAUSE_ID_TAB_SIZE  1000

Glist     *Clause_id_tab  = NULL;
unsigned  Clause_id_count = 0;

/*************
 *
 *   next_clause_id()
 *
 *************/

static
int next_clause_id()
{
  return ++Clause_id_count;
}  /* next_clause_id */

/*************
 *
 *    init_clause_id_tab()
 *
 *    Allocate a clause-id table.
 *
 *************/

static
void init_clause_id_tab(void)
{
  int i;
  Clause_id_tab = tp_alloc(CLAUSE_ID_TAB_SIZE * sizeof(Glist *));
  for (i = 0; i < CLAUSE_ID_TAB_SIZE; i++)
    Clause_id_tab[i] = NULL;
}  /* init_clause_id_tab */

/*************
 *
 *   assign_clause_id(c)
 *
 *************/

/* DOCUMENTATION
This routine assigns a unique identifier to the id field of a clause.
It also inserts the clause into a hash table so that given an id
number, the corresponding clause can be retrieved quickly (see
find_clause_by_id()).
*/

/* PUBLIC */
void assign_clause_id(Clause c)
{
  int i;

  if (c->id > 0) {
    p_clause(c);
    fatal_error("assign_clause_id, clause already has ID.");
  }
  c->id = next_clause_id();
  if (Clause_id_tab == NULL)
    init_clause_id_tab();
  i = c->id % CLAUSE_ID_TAB_SIZE;
  Clause_id_tab[i] = insert_clause_into_glist(Clause_id_tab[i], c, TRUE);
}  /* assign_clause_id */

/*************
 *
 *     unassign_clause_id(c)
 *
 *************/

/* DOCUMENTATION
This routine removes a clause from the ID hash table.
*/

/* PUBLIC */
void unassign_clause_id(Clause c)
{
  int i;
  Glist p1, p2;

  i = c->id % CLAUSE_ID_TAB_SIZE;
  for (p2 = NULL, p1 = Clause_id_tab[i];
       p1 && ((Clause) p1->u.v)->id < c->id;
       p2 = p1, p1 = p1->next);
  if (p1 == NULL || ((Clause) p1->u.v)->id != c->id) {
    p_clause(c);
    fatal_error("unassign_clause_id, cannot find clause.");
  }
  if (p2)
    p2->next = p1->next;
  else
    Clause_id_tab[i] = p1->next;
  free_glist(p1);
}  /* unassign_clause_id */

/*************
 *
 *   has_clause_id()
 *
 *************/

/* DOCUMENTATION
This function checks if a clause has been assigned an ID.
*/

/* PUBLIC */
BOOL has_clause_id(Clause c)
{
  return c->id > 0;
}  /* has_clause_id */

/*************
 *
 *     find_clause_by_id(id)
 *
 *     Given a clause ID, retrieve the clause (or NULL).
 *
 *************/

/* DOCUMENTATION
This routine retrieves the clause with the given ID number
(or NULL, if there is no such clause).
*/

/* PUBLIC */
Clause find_clause_by_id(int id)
{
  int i;
  Glist p1;

  i = id % CLAUSE_ID_TAB_SIZE;
  for (p1 = Clause_id_tab[i];
       p1 && ((Clause) p1->u.v)->id < id;
       p1 = p1->next);
  if (p1 != NULL && ((Clause) p1->u.v)->id == id)
    return(p1->u.v);
  else
    return(NULL);
}  /* find_clause_by_id */

/*************
 *
 *     fprint_clause_id_tab(fp)
 *
 *************/

/* DOCUMENTATION
This routine prints (to FILE *fp) all the clauses in the ID hash table.
*/

/* PUBLIC */
void fprint_clause_id_tab(FILE *fp)
{
  int i;
  Glist p;

  fprintf(fp, "\nID clause table:\n");
  for (i=0; i<CLAUSE_ID_TAB_SIZE; i++)
    for (p = Clause_id_tab[i]; p; p = p->next)
      fprint_clause(fp, p->u.v);
  fflush(fp);
}  /* fprint_clause_id_tab */

/*************
 *
 *     p_clause_id_tab(tab)
 *
 *************/

/* DOCUMENTATION
This routine prints (to stdout) all the clauses in the ID hash table.
*/

/* PUBLIC */
void p_clause_id_tab()
{
  fprint_clause_id_tab(stdout);
}  /* p_clause_id_tab */

/*************
 *
 *   insert_clause_into_glist()
 *
 *************/

/* DOCUMENTATION
This routine inserts a clause into a sorted (by ID) Glist of clauses.
Boolean paramemeter "increasing" tells whether the list is increasing
or decreasing.
The updated Glist is returned.
If the clause is already there, nothing happens.
*/

/* PUBLIC */
Glist insert_clause_into_glist(Glist p, Clause c, BOOL increasing)
{
  Glist prev, curr, new;
  prev = NULL;
  curr = p;
  while (curr != NULL && (increasing ? ((Clause) curr->u.v)->id < c->id
	                             : ((Clause) curr->u.v)->id > c->id)) {
    prev = curr;
    curr = curr->next;
  }
  if (curr == NULL || ((Clause) curr->u.v)->id != c->id) {
    new = get_glist();
    new->u.v = c;
    new->next = curr;
    if (prev != NULL)
      prev->next = new;
    else
      p = new;
  }
  return p;
}  /* insert_clause_into_glist */
