#include "commands.h"

/* Private definitions and types */

/*************
 *
 *   bell()
 *
 *************/

static
void bell(FILE *fp)
{
  fprintf(fp, "%c", '\007');
}  /* bell */

/*************
 *
 *   read_commands()
 *
 *************/

/* DOCUMENTATION
<P>
This is for input that contains optional commands;
if commands are present, then "end_of_list." is optional at the
end of the commands.
The commands are read from file fin.  Here are examples
of the currently supported commands.
<UL>
<LI>  set(verbose).
<LI>  clear(verbose).
<LI>  assign(max_mem, 2000).
<LI>  assoc_comm(+).
<LI>  commutative(+).
<LI>  symmetric(+).
<LI>  op(400, infix, ^).  op(400, infix, [+,*,v,^]).
<LI>  lex([e, a, b, _*_, _', h(_,_)]).
</UL>
<P>
A term is returned:
<UL>
<LI>  NULL: commands were read up to EOF.
<LI>  the constant end_of_list: commands were read up to the term "end_of_list".
<LI>  any other term: commands were read up to this unrecognized term; you
      will probably wish to continue your input processing with this term.
</UL>
<P>
If any error occurs, a message goes to file fout and to stderr,
and a fatal_error() occurs.
*/

/* PUBLIC */
Term read_commands(FILE *fin, FILE *fout, BOOL echo)
{
  Term t = read_term(fin, fout);
  BOOL go = (t != NULL);

  while (go) {
    if (is_symbol(SYMNUM(t), "set", 1) || is_symbol(SYMNUM(t), "clear", 1)) {
      int flag = str_to_flag_id(sn_to_str(SYMNUM(t->args[0])));
      if (flag == -1) {
	bell(stderr);
	fwrite_term_nl(fout, t);
	fwrite_term_nl(stderr, t);
	fatal_error("flag not recognized");
      }
      else {
	if (is_symbol(SYMNUM(t), "set", 1))
	  set_flag(flag);
	else
	  clear_flag(flag);
      }
    }
    else if (is_symbol(SYMNUM(t), "assign", 2)) {
      int val;
      BOOL ok = str_to_int(sn_to_str(SYMNUM(t->args[1])), &val);
      if (!ok) {
	bell(stderr);
	fwrite_term_nl(fout, t);
	fwrite_term_nl(stderr, t);
	fatal_error("bad parm value");
      }
      else {
	int parm = str_to_parm_id(sn_to_str(SYMNUM(t->args[0])));
	if (parm == -1) {
	  bell(stderr);
	  fwrite_term_nl(fout, t);
	  fwrite_term_nl(stderr, t);
	  fatal_error("parm not recognized");
	}
	else {
	  assign_parm(parm, val);
	}
      }
    }
    else if (is_symbol(SYMNUM(t), "assoc_comm", 1) ||
             is_symbol(SYMNUM(t), "commutative", 1) ||
	     is_symbol(SYMNUM(t), "symmetric", 1)) {
      Term f = t->args[0];
      if (!CONSTANT(f)) {
	bell(stderr);
	fwrite_term_nl(fout, t);
	fwrite_term_nl(stderr, t);
	fatal_error("argument must be symbol only");
      }
      else {
	if (is_symbol(SYMNUM(t), "assoc_comm", 1))
	  set_assoc_comm(sn_to_str(SYMNUM(f)));
	else
	  set_commutative(sn_to_str(SYMNUM(f)));  /* for symmetric as well */
      }
    }
    else if (is_symbol(SYMNUM(t), "op", 3)) {
      /* e.g., op(300, infix, +); */
      Term prec_term = t->args[0];
      Term type_term = t->args[1];
      Term symb_term = t->args[2];
      int prec;
      BOOL ok = str_to_int(sn_to_str(SYMNUM(prec_term)), &prec);
      if (!ok || prec < MIN_PRECEDENCE || prec > MAX_PRECEDENCE) {
	bell(stderr);
	fwrite_term_nl(fout, t);
	fwrite_term_nl(stderr, t);
	fatal_error("bad precedence in op command");
      }
      else {
	Parsetype pt = NOTHING_SPECIAL;
	if (is_symbol(SYMNUM(type_term), "infix", 0))
	  pt = INFIX;
	else if (is_symbol(SYMNUM(type_term), "infix_left", 0))
	  pt = INFIX_LEFT;
	else if (is_symbol(SYMNUM(type_term), "infix_right", 0))
	  pt = INFIX_RIGHT;
	else if (is_symbol(SYMNUM(type_term), "prefix", 0))
	  pt = PREFIX;
	else if (is_symbol(SYMNUM(type_term), "prefix_paren", 0))
	  pt = PREFIX_PAREN;
	else if (is_symbol(SYMNUM(type_term), "postfix", 0))
	  pt = POSTFIX;
	else if (is_symbol(SYMNUM(type_term), "postfix_paren", 0))
	  pt = POSTFIX_PAREN;
	else {
	  bell(stderr);
	  fwrite_term_nl(fout, t);
	  fwrite_term_nl(stderr, t);
	  fatal_error("bad parse-type in op command");
	}
	set_parse_type(sn_to_str(SYMNUM(symb_term)), prec, pt);
      }
    }
    else if (is_symbol(SYMNUM(t), "lex", 1)) {
      Term l = t->args[0];
      if (!proper_listterm(l)) {
	bell(stderr);
	fwrite_term_nl(fout, t);
	fwrite_term_nl(stderr, t);
	fatal_error("lex command must contain a proper list, e.g., [a,b,c]");
      }
      else {
	int n = 0;
	while (cons_term(l)) {
	  set_lex_val(SYMNUM(l->args[0]), ++n);
	  l = l->args[1];
	}
      }
    }
    else {
      go = FALSE;  /* return this unrecognized term (could be end_of_list) */
    }
    if (go) {
      if (echo)
	fwrite_term_nl(fout, t);
      zap_term(t);
      t = read_term(fin, fout);
    }
  }
  return t;
}  /* read_commands */

