#include "glist.h"

/*
 * memory management
 */

static Glist glist_avail;
static unsigned glist_gets, glist_frees, glist_avails;

static G2list g2list_avail;
static unsigned g2list_gets, g2list_frees, g2list_avails;

/* end of counters for each type */

/*************
 *
 *    Glist get_glist()
 *
 *************/

/* DOCUMENTATION
This routine allocates and returns a Glist node.
*/

/* PUBLIC */
Glist get_glist(void)
{
  Glist p;
    
  glist_gets++;
  if (glist_avail == NULL)
    p = tp_alloc(sizeof(struct glist));
  else {
    glist_avails--;
    p = glist_avail;
    glist_avail = glist_avail->next;
  }
  p->u.v = NULL;
  p->next = NULL;
  return(p);
}  /* get_glist */

/*************
 *
 *    free_glist()
 *
 *************/

/* DOCUMENTATION
This routine frees a Glist node.  If you wish to free a whole
Glist list, call zap_glist();
*/

/* PUBLIC */
void free_glist(Glist p)
{
  glist_frees++;
  glist_avails++;
  p->next = glist_avail;
  glist_avail = p;
}  /* free_glist */

/*************
 *
 *    G2list get_g2list()
 *
 *************/

/* DOCUMENTATION
This routine allocates and returns a G2list node.
*/

/* PUBLIC */
G2list get_g2list(void)
{
  G2list p;
    
  g2list_gets++;
  if (g2list_avail == NULL)
    p = tp_alloc(sizeof(struct g2list));
  else {
    g2list_avails--;
    p = g2list_avail;
    g2list_avail = g2list_avail->next;
  }
  p->u1.v = NULL;
  p->u2.v = NULL;
  p->next = NULL;
  return(p);
}  /* get_g2list */

/*************
 *
 *    free_g2list()
 *
 *************/

/* DOCUMENTATION
This routine frees a G2list node.  If you wish to free a whole
G2list 2list, call zap_g2list();
*/

/* PUBLIC */
void free_g2list(G2list p)
{
  g2list_frees++;
  g2list_avails++;
  p->next = g2list_avail;
  g2list_avail = p;
}  /* free_g2list */

/* end of get and free routines for each type */

/*************
 *
 *   fprint_glist_mem()
 *
 *************/

/* DOCUMENTATION
This routine prints (to FILE *fp) memory usage statistics for Glists.
The Boolean argument heading tells whether to print a heading on the table.
*/

/* PUBLIC */
void fprint_glist_mem(FILE *fp, BOOL heading)
{
  int n;
  if (heading)
    fprintf(fp, "  type (bytes each)        gets      frees     in use      avail      bytes\n");
  
  n = sizeof(struct glist);
  fprintf(fp, "glist (%4d)        %11u%11u%11u%11u%9.1f K\n",
	  n, glist_gets, glist_frees, glist_gets - glist_frees,
	  glist_avails,
	  (((glist_gets - glist_frees) + glist_avails) * n) / 1024.);

  fprintf(fp, "g2list (%4d)       %11u%11u%11u%11u%9.1f K\n",
	  n, g2list_gets, g2list_frees, g2list_gets - g2list_frees,
	  g2list_avails,
	  (((g2list_gets - g2list_frees) + g2list_avails) * n) / 1024.);

  /* end of printing for each type */
  
}  /* fprint_glist_mem */

/*************
 *
 *   p_glist_mem()
 *
 *************/

/* DOCUMENTATION
This routine prints (to stdout) memory usage statistics for Glists.
*/

/* PUBLIC */
void p_glist_mem(void)
{
  fprint_glist_mem(stdout, 1);
}  /* p_glist_mem */

/*
 *  end of memory management
 */

/*************
 *
 *   rev_app_glist(p1, p2)
 *
 *************/

static
Glist rev_app_glist(Glist p1, Glist p2)
{
  Glist p3;

  if (p1 == NULL)
    return(p2);
  else {
    p3 = p1->next;
    p1->next = p2;
    return(rev_app_glist(p3, p1));
  }
}  /* rev_app_glist */

/*************
 *
 *     reverse_glist(p1)
 *
 *************/

/* DOCUMENTATION
This routine reverses a Glist.  The list is reversed in-place,
so you should not refer to the argument after calling this routine.
A good way to use it is like this:
<PRE>
p = reverse_glist(p);
</PRE>
*/

/* PUBLIC */
Glist reverse_glist(Glist p)
{
  return rev_app_glist(p, NULL);
}  /* reverse_glist */

/*************
 *
 *   copy_glist(p)
 *
 *************/

/* DOCUMENTATION
This routine copies a Glist (the whole list) and returns the copy.
*/

/* PUBLIC */
Glist copy_glist(Glist p)
{
  Glist start, p1, p2;

  start = NULL;
  p2 = NULL;
  for ( ; p; p = p->next) {
    p1 = get_glist();
    p1->u.v = p->u.v;
    if (p2)
      p2->next = p1;
    else
      start = p1;
    p2 = p1;
  }
  return(start);
}  /* copy_glist */

/*************
 *
 *   zap_glist(p)
 *
 *************/

/* DOCUMENTATION
This routine frees a Glist (the whole list).  If the list contains
any pointers, the things to which they point are not freed.
*/

/* PUBLIC */
void zap_glist(Glist p)
{
  Glist curr, prev;

  curr = p;
  while (curr != NULL) {
    prev = curr;
    curr = curr->next;
    free_glist(prev);
  }
}  /* zap_glist */

/*************
 *
 *   glist_count()
 *
 *************/

/* DOCUMENTATION
This routine returns the length of a Glist.
*/

/* PUBLIC */
int glist_count(Glist p)
{
  int n;
  for (n = 0; p; p = p->next, n++);
  return(n);
}  /* glist_count */

/*************
 *
 *   fprint_ilist()
 *
 *************/

/* DOCUMENTATION
This routine assumes that the Glist p contains integers.
The list of integers is printed to FILE *fp like this: (4 5 1 3),
without a newline.
*/

/* PUBLIC */
void fprint_ilist(FILE *fp, Glist p)
{
  fprintf(fp, "(");
  for (; p != NULL; p = p->next) {
    fprintf(fp, "%d", p->u.i);
    if (p->next != NULL)
      fprintf(fp, " ");
  }
  fprintf(fp, ")");
  fflush(fp);
}  /* fprint_ilist */

/*************
 *
 *   p_ilist()
 *
 *************/

/* DOCUMENTATION
This routine assumes that the Glist p contains integers.
The list of integers is printed to stdout like this: (4 5 1 3),
with a '\n' at the end.
*/

/* PUBLIC */
void p_ilist(Glist p)
{
  fprint_ilist(stdout, p);
  printf("\n");
  fflush(stdout);
}  /* p_ilist */

/*************
 *
 *   ilist_append()
 *
 *************/

/* DOCUMENTATION
This routine appends an integer to a Glist.  The updated Glist
is returned.
*/

/* PUBLIC */
Glist ilist_append(Glist lst, int i)
{
  if (lst == NULL) {
    Glist g = get_glist();
    g->u.i = i;
    return g;
  }
  else {
    lst->next = ilist_append(lst->next, i);
    return lst;
  }
}  /* ilist_append */

/*************
 *
 *   ilist_prepend()
 *
 *************/

/* DOCUMENTATION
This routine inserts an integer as the first member of a Glist.
The updated Glist is returned.
*/

/* PUBLIC */
Glist ilist_prepend(Glist lst, int i)
{
  Glist g = get_glist();
  g->u.i = i;
  g->next = lst;
  return g;
}  /* ilist_prepend */

/*************
 *
 *   ilist_member()
 *
 *************/

/* DOCUMENTATION
This function checks if an integer is a member of a Glist.
(If a node in the Glist contains a pointer instead of an integer,
the result is undefined.)
*/

/* PUBLIC */
BOOL ilist_member(Glist lst, int i)
{
  if (lst == NULL)
    return FALSE;
  else if (lst->u.i == i)
    return TRUE;
  else
    return ilist_member(lst->next, i);
}  /* ilist_member */

/*************
 *
 *   plist_append()
 *
 *************/

/* DOCUMENTATION
This routine appends a pointer to a Glist.  The updated Glist is returned.
(The whole Glist is traversed.)
*/

/* PUBLIC */
Glist plist_append(Glist lst, void *p)
{
  if (lst == NULL) {
    Glist g = get_glist();
    g->u.v = p;
    return g;
  }
  else {
    lst->next = plist_append(lst->next, p);
    return lst;
  }
}  /* plist_append */

/*************
 *
 *   plist_prepend()
 *
 *************/

/* DOCUMENTATION
This routine inserts a pointer as the first member of a Glist.
The updated Glist is returned.
*/

/* PUBLIC */
Glist plist_prepend(Glist lst, void *p)
{
  Glist g = get_glist();
  g->u.v = p;
  g->next = lst;
  return g;
}  /* plist_prepend */

/*************
 *
 *   plist_member()
 *
 *************/

/* DOCUMENTATION
This routine checks if a pointer is a member of a Glist.
If any node of the Glist contains an integer instead of a pointer,
the result is undefined.
*/

/* PUBLIC */
BOOL plist_member(Glist lst, void *p)
{
  if (lst == NULL)
    return FALSE;
  else if (lst->u.v == p)
    return TRUE;
  else
    return plist_member(lst->next, p);
}  /* plist_member */

/*************
 *
 *   glist_pop()
 *
 *************/

/* DOCUMENTATION
This routine takes a nonempty Glist, removes and frees the first node
(ignoring the contents), and returns the remainder of the list.
*/

/* PUBLIC */
Glist glist_pop(Glist p)
{
  Glist q = p;
  p = p->next;
  free_glist(q);
  return p;
}  /* glist_pop */

/*************
 *
 *   glist_cat()
 *
 *************/

/* DOCUMENTATION
Concatenate two Glists and return the result.  The result is constructed
from the arguments, so not refer to either of the arguments after the call.
*/

/* PUBLIC */
Glist glist_cat(Glist p1, Glist p2)
{
  if (p1 == NULL)
    return p2;
  else if (p2 == NULL)
    return p1;
  else {
    Glist p = p1;
    while (p->next != NULL)
      p = p->next;
    p->next = p2;
    return p1;
  }
}  /* glist_cat */

/***************************************************************** g2list */

/*************
 *
 *   zap_g2list(p)
 *
 *************/

/* DOCUMENTATION
This routine frees a G2list (the whole 2list).  If the 2list contains
any pointers, the things to which they point are not freed.
*/

/* PUBLIC */
void zap_g2list(G2list p)
{
  G2list curr, prev;

  curr = p;
  while (curr != NULL) {
    prev = curr;
    curr = curr->next;
    free_g2list(prev);
  }
}  /* zap_g2list */

/*************
 *
 *   i2list_prepend()
 *
 *************/

/* DOCUMENTATION
This routine inserts an integer as the first member of a G2list.
The updated G2list is returned.
*/

/* PUBLIC */
G2list i2list_prepend(G2list lst, int i1, int i2)
{
  G2list g = get_g2list();
  g->u1.i = i1;
  g->u2.i = i2;
  g->next = lst;
  return g;
}  /* i2list_prepend */

/*************
 *
 *   g2list_pop()
 *
 *************/

/* DOCUMENTATION
This routine takes a nonempty G2list, removes and frees the first node
(ignoring the contents), and returns the remainder of the 2list.
*/

/* PUBLIC */
G2list g2list_pop(G2list p)
{
  G2list q = p;
  p = p->next;
  free_g2list(q);
  return p;
}  /* g2list_pop */

