#include "interp.h"

/* Private definitions and types */

struct interp {
  Term  t;        /* the term representation (from which this was built) */
  int   size;     /* domain size */
  /* Array of pointers to tables, one for each constant, function,
   * or predicate symbol (indexed by symbol number).  The arity and
   * print symbol can be obtained from the symbol number.
   */
  int **tables;
  int table_size;
  Interp next;  /* for avail list */
};

/*
 * memory management
 */

static Interp interp_avail;
static unsigned interp_gets, interp_frees, interp_avails;

/* end of counters for each type */

/*************
 *
 *    Interp get_interp()
 *
 *************/

static
Interp get_interp(void)
{
  Interp p;

  interp_gets++;
  if (interp_avail == NULL)
    p = tp_alloc(sizeof(struct interp));
  else {
    interp_avails--;
    p = interp_avail;
    interp_avail = interp_avail->next;
  }
  /* initialization */	
  	
  p->next = NULL;
  p->t = NULL;
  p->size = 0;
  p->table_size = 0;
  p->tables = NULL;

  /* end of initialization */	
  return(p);
}  /* get_interp */

/*************
 *
 *    free_interp()
 *
 *************/

static
void free_interp(Interp p)
{
  interp_frees++;
  interp_avails++;
  p->next = interp_avail;
  interp_avail = p;
}  /* free_interp */

/* end of get and free routines for each type */

/*************
 *
 *   fprint_interp_mem()
 *
 *************/

/* DOCUMENTATION
This routine prints (to FILE *fp) memory usage statistics for data types
associated with interpretations.
The Boolean argument heading tells whether to print a heading on the table.
*/

/* PUBLIC */
void fprint_interp_mem(FILE *fp, BOOL heading)
{
  int n;
  if (heading)
    fprintf(fp, "  type (bytes each)        gets      frees     in use      avail      bytes\n");

  n = sizeof(struct interp);
  fprintf(fp, "interp (%4d)       %11u%11u%11u%11u%9.1f K\n",
	  n, interp_gets, interp_frees,
	  interp_gets - interp_frees, interp_avails,
	  (((interp_gets - interp_frees) + interp_avails) * n) / 1024.);

  /* end of printing for each type */
  
}  /* fprint_interp_mem */

/*************
 *
 *   p_interp_mem()
 *
 *************/

/* DOCUMENTATION
This routine prints (to stdout) memory usage statistics for data types
associated with interpretations.
*/

/* PUBLIC */
void p_interp_mem()
{
  fprint_interp_mem(stdout, 1);
}  /* p_interp_mem */

/*
 *  end of memory management
 */

/*************
 *
 *   compile_interp()
 *
 *************/

/* DOCUMENTATION
This routine takes a term representing an interpretation and
builds a data structure that allows fast evaluation of clauses
and formulas w.r.t. the interpretation.  Here is an example of
the term form of an interpretation.
<PRE>
   interpretation(3, [
       function(e,      [0]),
       function(*(_,_), [0,1,2,1,2,0,2,0,1]),
       relation(p,      [1])
       relation(q(_),   [1,0,1])
       ])
</PRE>
Arity of symbols must be <= 4 (or a fatal error occurs).
The given Term t is not changed.
*/

/* PUBLIC */
Interp compile_interp(Term t)
{
  Interp p;
  int number_of_ops, domain_size, arity;
  int i, j, n, symnum, val, max, rc;
  int function = 0;
  int *table; 
  Term operations, lst;

  if (t->arity != 2) 
    fatal_error("compile_interp, bad arity.");

  rc = str_to_int(sn_to_str(SYMNUM(t->args[0])), &domain_size);

  if (!rc || domain_size < 1)
    fatal_error("compile_interp, domain size out of range.");

  operations = t->args[1];

  number_of_ops = listterm_length(operations);
  if (number_of_ops == 0)
    fatal_error("compile_interp, no operations.");

  /* Get the largest symnum, so we can get an table array big enough. */

  max = 0;
  for (i = 1; i <= number_of_ops; i++) {
    Term f = listterm_i(operations, i);
    if (f->arity != 2 || VARIABLE(f->args[0]))
      fatal_error("compile_interp, bad operation.");
    symnum = SYMNUM(f->args[0]);
    max = symnum > max ? symnum : max;
  }

  p = get_interp();
  p->t = copy_term(t);
  p->size = domain_size;
  p->table_size = max+1;

  p->tables = my_malloc(p->table_size * sizeof(int *));
  for (i = 0; i < p->table_size; i++)
    p->tables[i] = NULL;

  for (i = 1; i <= number_of_ops; i++) {
    Term f = listterm_i(operations, i);
    if (is_symbol(SYMNUM(f), "function", 2))
      function = 1;
    else if (is_symbol(SYMNUM(f), "relation", 2) || 
	     is_symbol(SYMNUM(f), "predicate", 2))
      function = 0;
    else
      fatal_error("compile_interp, bad function/relation");

    symnum = SYMNUM(f->args[0]);
    arity = f->args[0]->arity;
    if (arity > 4)
      fatal_error("compile_interp, arity > 4.");

    /* n = domain_size^arity */

    for (j = 0, n = 1; j < arity; j++)
      n = n * domain_size;

    lst = f->args[1];
    if (listterm_length(lst) != n)
      fatal_error("compile_interp, bad list.");

    p->tables[symnum] = my_malloc(n * sizeof(int));
    table = p->tables[symnum];

    for (j = 0; j < n; j++, lst = lst->args[1]) {
      rc = str_to_int(sn_to_str(SYMNUM(lst->args[0])), &val);
      if (!rc)
	fatal_error("compile_interp, bad domain elemnt.");
      else if (function && (val < 0 || val > domain_size-1))
	fatal_error("compile_interp, function element out of range.");
      else if (!function && (val < 0 || val > 1))
	fatal_error("compile_interp, relation element out of range.");
      else
	table[j] = val;
    }
  }
  return p;
}  /* compile_interp */

/*************
 *
 *   transpose_binary()
 *
 *************/

/* DOCUMENTATION
This routine takes a term representing an interpretation and
(destructively) transposes all of the binary functions and relations.
It is assumed that the interpretation is well-formed.  You can check
well-formedness first by calling compile_interp().
*/

/* PUBLIC */
void transpose_binary(Term t)
{
  int number_of_ops, n;
  int i, rc;
  Term operations;

  rc = str_to_int(sn_to_str(SYMNUM(t->args[0])), &n);
  operations = t->args[1];
  number_of_ops = listterm_length(operations);

  for (i = 1; i <= number_of_ops; i++) {
    Term f = listterm_i(operations, i);  /* e.g., function(j(_,_), [0,1,1,0]) */
    if (f->args[0]->arity == 2) {
      int j, k;
      Term lst = f->args[1];             /* e.g., [0,1,1,0] */
      for (j = 0; j < n; j++) {
	for (k = j+1; k < n; k++) {
	  Term t1 = listterm_i(lst, j*n+k+1);
	  Term t2 = listterm_i(lst, k*n+j+1);
	  int tmp = t1->private_symbol;
	  t1->private_symbol = t2->private_symbol;
	  t2->private_symbol = tmp;
	}
      }
    }
  }
}  /* transpose_binary */

/*************
 *
 *   zap_interp()
 *
 *************/

/* DOCUMENTATION
Free a compiled interpretation.
*/

/* PUBLIC */
void zap_interp(Interp p)
{
  int i;

  for (i = 0; i < p->table_size; i++)
    if (p->tables[i] != NULL)
      my_free(p->tables[i]);

  my_free(p->tables);

  zap_term(p->t);
  free_interp(p);
}  /* zap_interp */

/*************
 *
 *   fprint_interp()
 *
 *************/

/* DOCUMENTATION
This routine prints (to FILE *fp) a compiled interpretation.
This is mostly for debugging---the form is not pretty, and it
is not easily parsable.
*/

/* PUBLIC */
void fprint_interp(FILE *fp, Interp p)
{
  int i, j, n, arity, *table;

  fprintf(fp, "\nInterpretation, domain_size=%d.\n", p->size);

  for (i = 0; i < p->table_size; i++) {
    table = p->tables[i];
    if (table != NULL) {
      arity = sn_to_arity(i);
      fprintf(fp, "%s/%d : [", sn_to_str(i), arity);
      for (j = 0, n = 1; j < arity; j++, n = n * p->size);
      for (j = 0; j < n-1; j++)
	fprintf(fp, "%d,", table[j]);
      fprintf(fp, "%d]\n", table[n-1]);
    }
  }
}  /* fprint_interp */

/*************
 *
 *   p_interp()
 *
 *************/

/* DOCUMENTATION
This routine prints (to stdout) an interpretation.
This is mostly for debugging---the form is not pretty, and it
is not easily parsable.
*/

/* PUBLIC */
void p_interp(Interp p)
{
  fprint_interp(stdout, p);
}  /* p_interp */

/*************
 *
 *   eval_term_ground()
 *
 *   This version is for clauses, where variables are represented as VARIABLEs.
 *   This works for arity <= 3.  It works for terms and atoms.
 *
 *************/

int Eval_count;

static
int eval_term_ground(Term t, Interp p, int *vals)
{
  if (VARIABLE(t)) {
    return vals[VARNUM(t)];
  }
  else {
    int *table;
    int v0, v1, v2, v3;
    int n = p->size;
    int sn = SYMNUM(t);

    if (sn >= p->table_size || p->tables[sn] == NULL) {
      printf("ready to abort, bad term: "); p_term(t);
      fatal_error("eval_term_ground, symbol not in interpretation");
    }

    table = p->tables[sn];

    switch (t->arity) {
    case 0:
      return table[0];
      break;
    case 1:
      v0 = eval_term_ground(t->args[0], p, vals);
      return table[v0];
      break;
    case 2:
      v0 = eval_term_ground(t->args[0], p, vals);
      v1 = eval_term_ground(t->args[1], p, vals);
      return table[v0*n + v1];
      break;
    case 3:
      v0 = eval_term_ground(t->args[0], p, vals);
      v1 = eval_term_ground(t->args[1], p, vals);
      v2 = eval_term_ground(t->args[2], p, vals);
      return table[v0*n*n + v1*n + v2];
      break;
    case 4:
      v0 = eval_term_ground(t->args[0], p, vals);
      v1 = eval_term_ground(t->args[1], p, vals);
      v2 = eval_term_ground(t->args[2], p, vals);
      v3 = eval_term_ground(t->args[3], p, vals);
      return table[v0*n*n*n + v1*n*n + v2*n + v3];
      break;
    default:
      fatal_error("eval_term_ground, arity too big");
      return FALSE;  /* to quiet compiler */
    }
  }
}  /* eval_term_ground */

/*************
 *
 *   eval_clause_ground()
 *
 *   Given a ground clause and an interpretation,
 *   return "at least one literal is true in the interpretation".
 *
 *************/

static
BOOL eval_clause_ground(Clause c, Interp p, int *vals)
{
  Literal lit;
  BOOL atom_val, true_literal;

  Eval_count++;
  true_literal = FALSE;
  for (lit = c->first_literal; lit && !true_literal; lit = lit->next) {

    if (is_eq_symbol(SYMNUM(lit->atom)))
      atom_val = (eval_term_ground(lit->atom->args[0], p, vals) ==
		  eval_term_ground(lit->atom->args[1], p, vals));
    else
      atom_val = eval_term_ground(lit->atom, p, vals);

    true_literal = (lit->sign ? atom_val : !atom_val);
  }
  return true_literal;
}  /* eval_clause_ground */

/*************
 *
 *   all_recurse()
 *
 *************/

static
BOOL all_recurse(Clause c, Interp p, int *vals, int nextvar, int nvars)
{
  if (nextvar == nvars) {
    return eval_clause_ground(c, p, vals);
  }
  else {
    int i, rc;
    for (i = 0; i < p->size; i++) {
      vals[nextvar] = i;
      rc = all_recurse(c, p, vals, nextvar+1, nvars);
      if (!rc)
	return FALSE;
    }
    return TRUE;
  }
}  /* all_recurse */

/*************
 *
 *   eval_clause()
 *
 *************/

#define MAX_VARS_EVAL 100

/* DOCUMENTATION
This routine evaluates a clause in an interpretation.
If all instances (over the domain of the interpretation) of the clause
are true in the interpretaion, TRUE is returned.  If any instance
is false, FALSE is returned.
<P>
Note that if the interpretation has d elements and the clause has
v variables, it takes d^v evaluations to verify the clause.
<P>
A fatal error occurs if any constant, function or predicate symbol
in the clause (other than "=", which is always built in)
is absent from the interpetation.
*/

/* PUBLIC */
BOOL eval_clause(Clause c, Interp p)
{
  int a[MAX_VARS_EVAL], *vals;
  int nvars;
  BOOL rc;

  nvars = greatest_variable_in_clause(c) + 1;
  if (nvars > MAX_VARS_EVAL)
    vals = my_malloc((nvars * sizeof(int)));
  else
    vals = a;

  rc = all_recurse(c, p, vals, 0, nvars);

  if (nvars > MAX_VARS_EVAL)
    my_free(vals);

#if 0
  if (rc)
    printf("Clause is true in this interpretation.\n");
  else {
    int i;
    printf("False for values: ");
    for (i = 0; i < nvars; i++)
      printf("v%d=%d ", i, vals[i]);
    printf("\n");
  }
#endif
  return rc;
}  /* eval_clause */

/*************
 *
 *   eval_fterm_ground()
 *
 *   This version is for formulas, where variables are
 *   represented as CONSTANTs.
 *   This works for arity <= 4.  It works for terms and atoms.
 *
 *************/

static
int eval_fterm_ground(Term t, Interp p, int *vals)
{
  if (VARIABLE(t))
    fatal_error("eval_fterm_ground, VARIABLE encountered.");

  if (vals[SYMNUM(t)] != -1)
    return vals[SYMNUM(t)];

  else {
    int *table;
    int v0, v1, v2, v3;
    int n = p->size;
    int sn = SYMNUM(t);

    if (sn >= p->table_size || p->tables[sn] == NULL) {
      printf("ready to abort, bad term: "); p_term(t);
      fatal_error("eval_fterm_ground, symbol not in interpretation");
    }

    table = p->tables[sn];

    switch (t->arity) {
    case 0:
      return table[0];
      break;
    case 1:
      v0 = eval_fterm_ground(t->args[0], p, vals);
      return table[v0];
      break;
    case 2:
      v0 = eval_fterm_ground(t->args[0], p, vals);
      v1 = eval_fterm_ground(t->args[1], p, vals);
      return table[v0*n + v1];
      break;
    case 3:
      v0 = eval_fterm_ground(t->args[0], p, vals);
      v1 = eval_fterm_ground(t->args[1], p, vals);
      v2 = eval_fterm_ground(t->args[2], p, vals);
      return table[v0*n*n + v1*n + v2];
      break;
    case 4:
      v0 = eval_fterm_ground(t->args[0], p, vals);
      v1 = eval_fterm_ground(t->args[1], p, vals);
      v2 = eval_fterm_ground(t->args[2], p, vals);
      v3 = eval_fterm_ground(t->args[3], p, vals);
      return table[v0*n*n*n + v1*n*n + v2*n + v3];
      break;
    default:
      fatal_error("eval_fterm_ground, arity too big");
      return 0;  /* to quiet compiler */
    }
  }
}  /* eval_fterm_ground */

/*************
 *
 *   eval_form()
 *
 *************/

static
BOOL eval_form(Formula f, Interp p, int vals[])
{
  if (f->type == ATOM_FORM) {
    if (is_eq_symbol(SYMNUM(f->atom)))
      return (eval_fterm_ground(f->atom->args[0], p, vals) ==
	      eval_fterm_ground(f->atom->args[1], p, vals));
    else
      return eval_fterm_ground(f->atom, p, vals);
  }
  else if (f->type == ALL_FORM) {
    /* ok if true for every element of domain */
    int i;
    BOOL ok = TRUE;
    int sn = str_to_sn(f->qvar, 0);
    if (vals[sn] != -1)
      fatal_error("eval_form, variable conflict.");
    for (i = 0; i < p->size && ok; i++) {
      vals[sn] = i;
      if (!eval_form(f->kids[0], p, vals))
	ok = FALSE;
    }
    vals[sn] = -1;
    return ok;
  }
  else if (f->type == EXISTS_FORM) {
    /* ok if true for any element of domain */
    int i;
    BOOL ok = FALSE;
    int sn = str_to_sn(f->qvar, 0);
    if (vals[sn] != -1)
      fatal_error("eval_form, variable conflict.");
    for (i = 0; i < p->size && !ok; i++) {
      vals[sn] = i;
      if (eval_form(f->kids[0], p, vals))
	ok = TRUE;
    }
    vals[sn] = -1;
    return ok;
  }
  else if (f->type == AND_FORM) {
    int i;
    BOOL ok = TRUE;
    for (i = 0; i < f->arity && ok; i++)
      if (!eval_form(f->kids[i], p, vals))
	ok = FALSE;
    return ok;
  }
  else if (f->type == OR_FORM) {
    int i;
    BOOL ok = FALSE;
    for (i = 0; i < f->arity && !ok; i++)
      if (eval_form(f->kids[i], p, vals))
	ok = TRUE;
    return ok;
  }
  else if (f->type == NOT_FORM) {
    return !eval_form(f->kids[0], p, vals);
  }
  else if (f->type == IFF_FORM) {
    return (eval_form(f->kids[0], p, vals) == eval_form(f->kids[1], p, vals));
  }
  else if (f->type == IMP_FORM) {
    return (!eval_form(f->kids[0], p, vals) || eval_form(f->kids[1], p, vals));
  }
  else if (f->type == IMPBY_FORM) {
    return (eval_form(f->kids[0], p, vals) || !eval_form(f->kids[1], p, vals));
  }
  else {
    fatal_error("eval_form, bad formula.");
    return 0;  /* to please the compiler */
  }
}  /* eval_form */

/*************
 *
 *   eval_formula()
 *
 *************/

/* DOCUMENTATION
This routine evaluates a formula in an interpretation.
There is no restriction on the structure of the formula.
However,
<I>quantified variables must be named in such a way that</I>
when a quantifier binds a variable, say (all x F), then
no quantifier in F can rebind x.  The routine eliminate_rebinding()
can be called to transform a formula, if necessary, so that it
satisfies the rule.
<P>
A fatal error occurs if any constant, function or predicate symbol
in the formula (other than "=", which is always built in)
is absent from the interpetation.
*/

/* PUBLIC */
BOOL eval_formula(Formula f, Interp p)
{
  int a[MAX_VARS_EVAL], *vals;
  int nsyms, i;
  BOOL rc;

  nsyms = greatest_symnum_in_formula(f) + 1;
  if (nsyms > MAX_VARS_EVAL)
    vals = my_malloc((nsyms * sizeof(int)));
  else
    vals = a;

  for (i = 0; i < nsyms; i++)
    vals[i] = -1;

  rc = eval_form(f, p, vals);

  if (nsyms > MAX_VARS_EVAL)
    my_free(vals);

#if 0
  if (rc)
    printf("Formula is TRUE in this interpretation.\n");
  else {
    printf("Formula is FALSE in this interpretation.\n");
  }
#endif
  return rc;
}  /* eval_formula */

/*************
 *
 *   fprint_interp_from_term()
 *
 *************/

/* DOCUMENTATION
This routine pretty prints (to FILE *fp) the term form of
an interpretation.  The result is mixfix-parsable.
<P>
For now, this works only for unsorted interpretations.
*/

/* PUBLIC */
void fprint_interp_from_term(FILE *fp, Term t)
{
  Term l1;
  int domain_size;
  int rc;
  
  rc = str_to_int(sn_to_str(SYMNUM(t->args[0])), &domain_size);

  fprintf(fp,"\n%s( %s, [\n",sn_to_str(SYMNUM(t)),sn_to_str(SYMNUM(t->args[0])));

  for (l1 = t->args[1]; !nil_term(l1); l1 = l1->args[1]) {
    Term f = l1->args[0];
    Term l2 = f->args[1];
    int arity = f->args[0]->arity;
    fprintf(fp, "\n\t%s( ", sn_to_str(SYMNUM(f)));
    fprint_term(fp, f->args[0]);
    fprintf(fp, ", ");
    if (arity == 0 || arity == 1) {
      fwrite_term(fp, l2);
      fprintf(fp, ")");
    }
    else if (arity == 2) {
      int i, j;
      fprintf(fp, "[\n\t\t");
      for (i = 0; i < domain_size; i++) {
	for (j = 0; j < domain_size; j++) {
	  fprint_term(fp, l2->args[0]);
	  l2 = l2->args[1];
	  if (!nil_term(l2))
	    fprintf(fp, ", ");
	  else
	    fprintf(fp, " ])");
	}
	if (!nil_term(l2))
	  fprintf(fp, "\n\t\t");
      }
    }  /* arity 2 */
    else if (arity == 3) {
      int i, j, k;
      fprintf(fp, "[\n\t\t");
      for (i = 0; i < domain_size; i++) {
	for (j = 0; j < domain_size; j++) {
	  for (k = 0; k < domain_size; k++) {
	    fprint_term(fp, l2->args[0]);
	    l2 = l2->args[1];
	    if (!nil_term(l2))
	      fprintf(fp, ", ");
	    else
	      fprintf(fp, " ])");
	  }
	  if (!nil_term(l2))
	    fprintf(fp, "\n\t\t");
	}
	if (!nil_term(l2))
	  fprintf(fp, "\n\t\t");
      }
    }  /* arity 3 */
    else if (arity == 4) {
      int i, j, k, l;
      fprintf(fp, "[\n\t\t");
      for (i = 0; i < domain_size; i++) {
	for (j = 0; j < domain_size; j++) {
	  for (k = 0; k < domain_size; k++) {
	    for (l = 0; l < domain_size; l++) {
	      fprint_term(fp, l2->args[0]);
	      l2 = l2->args[1];
	      if (!nil_term(l2))
		fprintf(fp, ", ");
	      else
		fprintf(fp, " ])");
	    }
	    if (!nil_term(l2))
	      fprintf(fp, "\n\t\t");
	  }
	  if (!nil_term(l2))
	    fprintf(fp, "\n\t\t");
	}
	if (!nil_term(l2))
	  fprintf(fp, "\n\t\t");
      }
    }  /* arity 4 */
    else {
      fwrite_term(fp, l2);
      fprintf(fp, ")");
    }
    if (!nil_term(l1->args[1]))
      fprintf(fp, ",\n");
    else
      fprintf(fp, " ]).\n");
  }  /* for each function or relation */
}  /* fprint_interp_from_term */

/*************
 *
 *   fprint_interp_from_term2()
 *
 *************/

/* DOCUMENTATION
This routine pretty prints (to FILE *fp) the term form of
an interpretation.  The result is mixfix-parsable.
<P>
For now, this works only for unsorted interpretations.
*/

/* PUBLIC */
void fprint_interp_from_term2(FILE *fp, Term t)
{
  Term l1;
  int domain_size;
  int rc;
  
  rc = str_to_int(sn_to_str(SYMNUM(t->args[0])), &domain_size);

  fprintf(fp,"%s(%s, [",sn_to_str(SYMNUM(t)),sn_to_str(SYMNUM(t->args[0])));

  for (l1 = t->args[1]; !nil_term(l1); l1 = l1->args[1]) {
    Term f = l1->args[0];
    Term l2 = f->args[1];
    int arity = f->args[0]->arity;
    fprintf(fp, " %s(", sn_to_str(SYMNUM(f)));
    fprint_term(fp, f->args[0]);
    fprintf(fp, ", ");
    if (arity == 0 || arity == 1) {
      fwrite_term(fp, l2);
      fprintf(fp, ")");
    }
    else if (arity == 2) {
      int i, j;
      fprintf(fp, "[ ");
      for (i = 0; i < domain_size; i++) {
	for (j = 0; j < domain_size; j++) {
	  fprint_term(fp, l2->args[0]);
	  l2 = l2->args[1];
	  if (!nil_term(l2))
	    fprintf(fp, ",");
	  else
	    fprintf(fp, " ])");
	}
	if (!nil_term(l2))
	  fprintf(fp, " ");
      }
    }  /* arity 2 */
    else if (arity == 3) {
      int i, j, k;
      fprintf(fp, "[ ");
      for (i = 0; i < domain_size; i++) {
	for (j = 0; j < domain_size; j++) {
	  for (k = 0; k < domain_size; k++) {
	    fprint_term(fp, l2->args[0]);
	    l2 = l2->args[1];
	    if (!nil_term(l2))
	      fprintf(fp, ",");
	    else
	      fprintf(fp, " ])");
	  }
	  if (!nil_term(l2))
	    fprintf(fp, " ");
	}
	if (!nil_term(l2))
	  fprintf(fp, " ");
      }
    }  /* arity 3 */
    else if (arity == 4) {
      int i, j, k, l;
      fprintf(fp, "[ ");
      for (i = 0; i < domain_size; i++) {
	for (j = 0; j < domain_size; j++) {
	  for (k = 0; k < domain_size; k++) {
	    for (l = 0; l < domain_size; l++) {
	      fprint_term(fp, l2->args[0]);
	      l2 = l2->args[1];
	      if (!nil_term(l2))
		fprintf(fp, ",");
	      else
		fprintf(fp, " ])");
	    }
	    if (!nil_term(l2))
	      fprintf(fp, " ");
	  }
	  if (!nil_term(l2))
	    fprintf(fp, " ");
	}
	if (!nil_term(l2))
	  fprintf(fp, " ");
      }
    }  /* arity 4 */
    else {
      fwrite_term(fp, l2);
      fprintf(fp, ")");
    }
    if (!nil_term(l1->args[1]))
      fprintf(fp, ",");
    else
      fprintf(fp, " ]).\n");
  }  /* for each function or relation */
}  /* fprint_interp_from_term2 */

/*************
 *
 *   fprint_interp_tabular()
 *
 *************/

/* DOCUMENTATION
This routine pretty prints (to FILE *fp)
an interpretation in tabular (non-mixfix-parsable) form.
<P>
For now, this works only for unsorted interpretations.
*/

/* PUBLIC */
void fprint_interp_tabular(FILE *fp, Term t)
{
  Term l1;
  int n;
  int rc, v, i, j;
  
  rc = str_to_int(sn_to_str(SYMNUM(t->args[0])), &n);

  for (l1 = t->args[1]; !nil_term(l1); l1 = l1->args[1]) {
    Term f = l1->args[0];
    Term l2 = f->args[1];
    int arity = f->args[0]->arity;
    fprintf(fp, "\n %s : ", sn_to_str(SYMNUM(f->args[0])));

    if (arity == 0) {
      fprintf(fp, sn_to_str(SYMNUM(l2->args[0])));
      fprintf(fp, "\n");
    }
    else if (arity == 1) {
      fprintf(fp, "\n       ");
      for (i = 0; i < n; i++)
	fprintf(fp, "%2d", i);
      fprintf(fp, "\n    ---");
      for (i = 0; i < n; i++)
	fprintf(fp, "--");
      fprintf(fp, "\n       ");
      for (i = 0; i < n; i++) {
	rc = str_to_int(sn_to_str(SYMNUM(l2->args[0])), &v);
	fprintf(fp, "%2d", v);
	l2 = l2->args[1];
      }
      fprintf(fp, "\n");
    }  /* arity 1 */
    else if (arity == 2) {
      fprintf(fp, "\n      |");
      for (i = 0; i < n; i++)
	fprintf(fp, "%2d", i);
      fprintf(fp, "\n    --+");
      for (i = 0; i < n; i++)
	fprintf(fp, "--");
      for (i = 0; i < n; i++) {
	fprintf(fp, "\n    %d |", i);
	for (j = 0; j < n; j++) {
	  rc = str_to_int(sn_to_str(SYMNUM(l2->args[0])), &v);
	  fprintf(fp, "%2d", v);
	  l2 = l2->args[1];
	}
      }
      fprintf(fp, "\n");
    }  /* arity 2 */
    else {
      fprintf(fp, "big_arity");
    }
  }  /* for each function or relation */
}  /* fprint_interp_tabular */

