#include "ioutil.h"

/* Private definitions and types */

/*************
 *
 *   fwrite_formula()
 *
 *************/

/* DOCUMENTATION
This routine prints a formula, followed by ".\n" to a file.
This version does not print extra parentheses
(it first translates the formula to a term, then
prints the term, then frees the term).
To print the formula directly, with extra parentheses,
call fprint_formula() instead.
*/

/* PUBLIC */
void fwrite_formula(FILE *fp, Formula f)
{
  Term t = formula_to_term(f);
  if (t == NULL)
    fatal_error("fwrite_formula, formula_to_term returns NULL.");
  fwrite_term_nl(fp, t);
  fflush(fp);
  zap_term(t);
}  /* fwrite_formula */

/*************
 *
 *   read_clause()
 *
 *************/

/* DOCUMENTATION
This routine reads a clause from FILE *fin.
If there are no more clauses in *fin, NULL is returned.
<P>
If any error occurs, a message is sent to FILE *fout and a fatal_error occurs.
<P>
Variables are "set", upward links ar made from all subterms to the
clause, and ac_canonical() is called on each atom.
*/

/* PUBLIC */
Clause read_clause(FILE *fin, FILE *fout)
{
  Term t;

  t = read_term(fin, fout);
  if (t == NULL)
    return NULL;
  else {
    Clause c = term_to_clause(t);
    Literal lit;
    BOOL ok = clause_set_variables(c, MAX_VARS);;
    if (!ok) {
      fprintf(fout, "read_clause, too many variables: ");
      fwrite_term_nl(fout, t);
      fatal_error("read_clause, too many variables");
    }
    zap_term(t);
    for (lit = c->first_literal; lit != NULL; lit = lit->next)
      ac_canonical(lit->atom, -1);
    upward_clause_links(c);
    return c;
  }
}  /* read_clause */

/*************
 *
 *   end_of_list_clause()
 *
 *************/

static
BOOL end_of_list_clause(Clause c)
{
  if (c == NULL)
    return FALSE;
  else if (number_of_literals(c) != 1)
    return FALSE;
  else {
    Term a = c->first_literal->atom;
    if (!CONSTANT(a))
      return FALSE;
    else
      return is_symbol(SYMNUM(a), "end_of_list", 0);
  }
}  /* end_of_list_clause */

/*************
 *
 *   read_clause_list()
 *
 *************/

/* DOCUMENTATION
This routine reads a list of clauses from FILE *fin.
If you with the list to have a name, send a string;
othersize send NULL.  (You can name the list later with name_clist().)
If there are no more clauses in *fin, an empty Clist is returned.
<P>
If any error occurs, a message is sent to FILE *fout and a fatal_error occurs.
*/

/* PUBLIC */
Clist read_clause_list(FILE *fin, FILE *fout, char *name, BOOL assign_id)
{
  Clist lst = clist_init(name);
  Clause c;

  c = read_clause(fin, fout);
  while (c != NULL && !end_of_list_clause(c)) {
    if (assign_id)
      assign_clause_id(c);
    clist_append(c, lst);
    c = read_clause(fin, fout);
  }
  if (c != NULL)
    zap_clause(c);  /* end_of_list_clause */
  return lst;
}  /* read_clause_list */

/*************
 *
 *   fwrite_clause()
 *
 *************/

/* DOCUMENTATION
This routine prints (to FILE *fp) a clause in mixfix form,
followed by ".\n".
*/

/* PUBLIC */
void fwrite_clause(FILE *fp, Clause c, BOOL print_id)
{
  Term t = clause_to_term(c);

  if (t == NULL)
    fatal_error("fwrite_clause, clause_to_term returns NULL.");
  
  if (print_id) {
    fprintf(fp, "%d. ", c->id);
    fprint_just(fp, c->justification);
    fprintf(fp, " ");
  }
  fwrite_term_nl(fp, t);
  fflush(fp);
  zap_term(t);
}  /* fwrite_clause */

/*************
 *
 *   fwrite_clause_list()
 *
 *************/

/* DOCUMENTATION
This routine prints (to FILE *fp) a list of clauses in mixfix form.
Example:
<PRE>
list(sos).
a = b.
end_of_list.
</PRE>
If the name of the list is "", it is written as list(anonymous).
*/

/* PUBLIC */
void fwrite_clause_list(FILE *fp, Clist lst, BOOL print_ids)
{
  Clist_pos p;
    
  if (lst->name == NULL || str_ident(lst->name, ""))
    fprintf(fp, "\nlist(anonymous).\n");
  else
    fprintf(fp, "\nlist(%s).\n", lst->name);
	  
  for (p = lst->first; p != NULL; p = p->next)
    fwrite_clause(fp, p->c, print_ids);
  fprintf(fp, "end_of_list.\n");
  fflush(fp);
}  /* fwrite_clause_list */

/*************
 *
 *   read_formula()
 *
 *************/

/* DOCUMENTATION
This routine reads a formula from FILE *fin.
If there are no more formulas in *fin, NULL is returned.
<P>
If any error occurs, a message is sent to FILE *fout and a fatal_error occurs.
*/

/* PUBLIC */
Formula read_formula(FILE *fin, FILE *fout)
{
  Term t;

  t = read_term(fin, fout);
  if (t == NULL)
    return NULL;
  else {
    Formula c = term_to_formula(t);
    zap_term(t);
    return c;
  }
}  /* read_formula */

/*************
 *
 *   end_of_list_formula()
 *
 *************/

static
BOOL end_of_list_formula(Formula f)
{
  if (f == NULL)
    return FALSE;
  else if (f->type != ATOM_FORM)
    return FALSE;
  else {
    Term a = f->atom;
    if (!CONSTANT(a))
      return FALSE;
    else
      return is_symbol(SYMNUM(a), "end_of_list", 0);
  }
}  /* end_of_list_formula */

/*************
 *
 *   read_formula_list()
 *
 *************/

/* DOCUMENTATION
This routine reads a list of formulas from FILE *fin.
If there are no more formulas in *fin, NULL is returned.
<P>
If any error occurs, a message is sent to FILE *fout and a fatal_error occurs.
*/

/* PUBLIC */
Glist read_formula_list(FILE *fin, FILE *fout)
{
  Glist p = NULL;
  Formula f;

  f = read_formula(fin, fout);
  while (f != NULL && !end_of_list_formula(f)) {
    p = plist_append(p, f);
    f = read_formula(fin, fout);
  }
  if (f != NULL)
    zap_formula(f);
  return p;
}  /* read_formula_list */

/*************
 *
 *   fwrite_formula_list()
 *
 *************/

/* DOCUMENTATION
This routine prints (to FILE *fp) a list of formulas in mixfix form.
Example:
<PRE>
list(sos).
a = b.
end_of_list.
</PRE>
If the name of the list is "", it is written as list(anonymous).
*/

/* PUBLIC */
void fwrite_formula_list(FILE *fp, Glist lst, char *name)
{
  Glist p;


  if (name == NULL || str_ident(name, ""))
    fprintf(fp, "\nformula_list(anonymous).\n");
  else
    fprintf(fp, "\nformula_list(%s).\n", name);

  for (p = lst; p != NULL; p = p->next)
    fwrite_formula(fp, p->u.v);
  fprintf(fp, "end_of_list.\n");
  fflush(fp);
}  /* fwrite_formula_list */

/*************
 *
 *   zap_formula_list()
 *
 *************/

/* DOCUMENTATION
Free a Glist of formulas.
*/

/* PUBLIC */
void zap_formula_list(Glist lst)
{
  Glist p = lst;
  while (p != NULL) {
    Glist p2 = p;
    p = p->next;
    zap_formula(p2->u.v);
    free_glist(p2);
  }
}  /* zap_formula_list */

/*************
 *
 *   end_of_list_term()
 *
 *************/

/* DOCUMENTATION
Check if a term is the constant "end_of_list".
*/

/* PUBLIC */
BOOL end_of_list_term(Term t)
{
  if (t == NULL)
    return FALSE;
  else if (!CONSTANT(t))
    return FALSE;
  else
    return is_symbol(SYMNUM(t), "end_of_list", 0);
}  /* end_of_list_term */

/*************
 *
 *   read_term_list()
 *
 *************/

/* DOCUMENTATION
This routine reads a list of terms from FILE *fin.
If there are no more terms in *fin, NULL is returned.
<P>
None of the returned terms or their subterms will have type VARIABLE,
even if it looks like a variable (e.g., x).
<P>
If any error occurs, a message is sent to FILE *fout and a fatal_error occurs.
*/

/* PUBLIC */
Glist read_term_list(FILE *fin, FILE *fout)
{
  Glist p = NULL;
  Term t;

  t = read_term(fin, fout);
  while (t != NULL && !end_of_list_term(t)) {
    p = plist_append(p, t);
    t = read_term(fin, fout);
  }
  if (t != NULL)
    zap_term(t);
  return p;
}  /* read_term_list */

/*************
 *
 *   fwrite_term_list()
 *
 *************/

/* DOCUMENTATION
This routine prints (to FILE *fp) a list of terms in mixfix form.
Example:
<PRE>
list(sos).
a = b.
end_of_list.
</PRE>
If the name of the list is "" or NULL, it is written as list(anonymous).
*/

/* PUBLIC */
void fwrite_term_list(FILE *fp, Glist lst, char *name)
{
  Glist p;

  if (name == NULL || str_ident(name, ""))
    fprintf(fp, "\nterm_list(anonymous).\n");
  else
    fprintf(fp, "\nterm_list(%s).\n", name);

  for (p = lst; p != NULL; p = p->next) {
    fwrite_term(fp, p->u.v);
    fprintf(fp, ".\n");
  }
  fprintf(fp, "end_of_list.\n");
  fflush(fp);
}  /* fwrite_term_list */

/*************
 *
 *   zap_term_list()
 *
 *************/

/* DOCUMENTATION
Free a Glist of terms.
*/

/* PUBLIC */
void zap_term_list(Glist lst)
{
  Glist p = lst;
  while (p != NULL) {
    Glist p2 = p;
    p = p->next;
    zap_term(p2->u.v);
    free_glist(p2);
  }
}  /* zap_term_list */

