#include "just.h"

/* Private definitions and types */

/* Justification Lists.
 *
 * Hyper, UR, binary (clash rules) --
 *   (rule (nucid nuclit1 sat1id sat1lit nuclit2 sat2id sat2lit ...))
 *
 * Para (with AC, position vectors are w.r.t flattened AC terms.)
 *   (rule (fromid fromvec intoid intovec))
 *
 * Factor
 *   (factor (parentid removed-literal))
 *
 * Demod
 *   (..... demod (id1 id2 ... idn))
 *
 * CD
 *   (CD (major-id minor-id))
 */  

/*************
 *
 *   fprint_just()
 *
 *************/

/* DOCUMENTATION
This routine prints (to FILE *fp) a clause justification that has
been encoded into a Glist.
No whitespace is printed before or after.
*/

/* PUBLIC */
void fprint_just(FILE *fp, Glist just)
{
  Glist g = just;
  fprintf(fp, "[");
  while (g != NULL) {
    char *s;
    Just rule = g->u.i;
    if (rule==BINARY_RES_JUST || rule==HYPER_RES_JUST || rule==UR_RES_JUST) {
      Glist p;
      switch (rule) {
      case BINARY_RES_JUST: s = "res"; break;
      case HYPER_RES_JUST:  s = "hyper"; break;
      case UR_RES_JUST:     s = "ur"; break;
      default:              s = "?";
      }
      fprintf(fp, "%s (", s);
      for (p = g->next->u.v; p != NULL; p = p->next) {
	fprintf(fp, "%d", p->u.i);
	fprintf(fp, "%s", p->next == NULL ? ")" : " ");
      }
    }
    else if (rule == FACTOR_JUST) {
      Glist p = g->next->u.v;
      fprintf(fp, "factor (%d %d)", p->u.i, p->next->u.i); 
    }
    else if (rule == CD_JUST) {
      Glist p = g->next->u.v;
      fprintf(fp, "cd (%d %d)", p->u.i, p->next->u.i); 
    }
    else if (rule == FLIP_JUST) {
      fprintf(fp, " flip %d", g->next->u.i); 
    }
    else if (rule == BACK_DEMOD_JUST) {
      fprintf(fp, "back_demod %d", g->next->u.i); 
    }
    else if (rule == DEMOD_JUST) {
      Glist p;
      fprintf(fp, " demod (");
      for (p = g->next->u.v; p != NULL; p = p->next) {
	fprintf(fp, "%d", p->u.i);
	fprintf(fp, "%s", p->next == NULL ? "" : " ");
      }
      fprintf(fp, ")");
    }
    else if (rule == PARA_JUST || rule == PARA_FX_JUST ||
	     rule == PARA_IX_JUST || rule == PARA_FX_IX_JUST) {
      Glist p        = g->next->u.v;
      int from_id    = p->u.i;
      Glist from_vec = p->next->u.v;
      int into_id    = p->next->next->u.i;
      Glist into_vec = p->next->next->next->u.v;
      char *srule = NULL;

      switch (rule) {
      case PARA_JUST:       srule = "para";       break;
      case PARA_FX_JUST:    srule = "para_fx";    break;
      case PARA_IX_JUST:    srule = "para_ix";    break;
      case PARA_FX_IX_JUST: srule = "para_fx_ix"; break;
      default: s = "?";
      }

      fprintf(fp, "%s (%d ", srule, from_id);
      fprint_ilist(fp, from_vec);
      fprintf(fp, " %d ", into_id);
      fprint_ilist(fp, into_vec);
      fprintf(fp, ")");
    }
    else {
      fprintf(fp, "unknown rule");
    }
    g = g->next->next;
  }
  fprintf(fp, "]");
}  /* fprint_just */

/*************
 *
 *   factor_just()
 *
 *************/

/* DOCUMENTATION
This routine builds and returns a justification list for a factorization.
*/

/* PUBLIC */
Glist factor_just(Clause c, int literal_num)
{
  /* (FACTOR_JUST (clause_id literal_num)) */
  return plist_append(ilist_append(NULL, FACTOR_JUST),
		      ilist_append(ilist_append(NULL, c->id),
				   literal_num));
}  /* factor_just */

/*************
 *
 *   cd_just()
 *
 *************/

/* DOCUMENTATION
This routine builds and returns a justification list for a factorization.
*/

/* PUBLIC */
Glist cd_just(Clause major, Clause minor)
{
  /* (FACTOR_JUST (clause_id literal_num)) */
  return plist_append(ilist_append(NULL, CD_JUST),
		      ilist_append(ilist_append(NULL, major->id),
				   minor->id));
}  /* cd_just */

/*************
 *
 *   back_demod_just()
 *
 *************/

/* DOCUMENTATION
This routine builds and returns a justification list for back_demod.
*/

/* PUBLIC */
Glist back_demod_just(Clause c)
{
  /* (BACK_DEMOD_JUST parent_id) */
  return ilist_append(ilist_append(NULL, BACK_DEMOD_JUST), c->id);
}  /* back_demod_just */

/*************
 *
 *   binary_res_just()
 *
 *************/

/* DOCUMENTATION
This routine builds and returns a justification list for binary resolution.
(Binary res justifications may also be constructed in resolve(), along
with hyper and UR.)
*/

/* PUBLIC */
Glist binary_res_just(Clause c1, int n1, Clause c2, int n2)
{
  /* (BINARY_RES_JUST (id1 lit1 id2 lit2) */
  return plist_append(ilist_append(NULL, BINARY_RES_JUST),
		      ilist_append(
                        ilist_append(
                          ilist_append(
                            ilist_append(NULL,c1->id),n1),c2->id),n2));

}  /* binary_res_just */

/*************
 *
 *   para_just()
 *
 *************/

/* DOCUMENTATION
This routine builds and returns a justification list for a paramodulation
inference.  The position vectors are copied and reversed.
*/

/* PUBLIC */
Glist para_just(int rule,
		Clause from, Glist from_vec,
		Clause into, Glist into_vec)
{
  Glist fv = reverse_glist(copy_glist(from_vec));
  Glist iv = reverse_glist(copy_glist(into_vec));

  /* (rule from_id from_vec into_id, into_vec) */

  return plist_append(ilist_append(NULL,rule),
	   plist_append(ilist_append(plist_append(ilist_append(NULL,
             from->id),fv),into->id),iv));
}  /* para_just */

/*************
 *
 *   zap_just()
 *
 *************/

/* DOCUMENTATION
This routine frees a justification list, including any sublists.
*/

/* PUBLIC */
void zap_just(Glist just)
{
  /* First, zap all of the sublists, then zap the main list. */
  Glist g = just;
  while (g != NULL) {
    int rule = g->u.i;
    if (rule==BINARY_RES_JUST || rule==HYPER_RES_JUST ||
	rule==UR_RES_JUST || rule == FACTOR_JUST ||
	rule==CD_JUST) {
      zap_glist(g->next->u.v);
    }
    else if (rule == DEMOD_JUST) {
      zap_glist(g->next->u.v);
    }
    else if (rule == FLIP_JUST) {
      ;  /* Do nothing. */
    }
    else if (rule == BACK_DEMOD_JUST) {
      ;  /* Do nothing. */
    }
    else if (rule == PARA_JUST || rule == PARA_FX_JUST ||
	     rule == PARA_IX_JUST || rule == PARA_FX_IX_JUST) {
      Glist p        = g->next->u.v;
      Glist from_vec = p->next->u.v;
      Glist into_vec = p->next->next->next->u.v;
      zap_glist(from_vec);
      zap_glist(into_vec);
      zap_glist(p);
    }
    else {
      fatal_error("zap_just, unknown rule");
    }
    g = g->next->next;
  }
  zap_glist(just);
}  /* zap_just */

/*************
 *
 *   get_clanc()
 *
 *************/

static
Glist get_clanc(int id, Glist anc)
{
  Clause c = find_clause_by_id(id);
  if (c == NULL) {
    printf("get_clanc, clause with id=%d not found.\n", id);
    fatal_error("get_clanc, clause not found.");
  }

  if (!plist_member(anc, c)) {
    Glist g = c->justification;
    anc = insert_clause_into_glist(anc, c, TRUE);

    while (g != NULL) {
      Just rule = g->u.i;
      if (rule==BINARY_RES_JUST || rule==HYPER_RES_JUST || rule==UR_RES_JUST) {
	/* [rule (nucid nuclit sat1id sat1lit nuclit2 sat2id sat2lit ...)] */
	Glist p = g->next->u.v;
	int nuc_id = p->u.i;
	anc = get_clanc(nuc_id, anc);
	p = p->next;
	while (p != NULL) {
	  int sat_id = p->next->u.i;
	  anc = get_clanc(sat_id, anc);
	  p = p->next->next->next;
	}
      }
      else if (rule == FACTOR_JUST) {
	Glist p = g->next->u.v;
	int parent_id = p->u.i;
	anc = get_clanc(parent_id, anc);
      }
      else if (rule == CD_JUST) {
	Glist p = g->next->u.v;
	int major_id = p->u.i;
	int minor_id = p->next->u.i;
	anc = get_clanc(major_id, anc);
	anc = get_clanc(minor_id, anc);
      }
      else if (rule == BACK_DEMOD_JUST) {
	int parent_id = g->next->u.i;
	anc = get_clanc(parent_id, anc);
      }
      else if (rule == DEMOD_JUST) {
	Glist p;
	for (p = g->next->u.v; p != NULL; p = p->next)
	  anc = get_clanc(p->u.i, anc);
      }
      else if (rule == FLIP_JUST) {
	;  /* Do nothing. */
      }
      else if (rule == PARA_JUST || rule == PARA_FX_JUST ||
	       rule == PARA_IX_JUST || rule == PARA_FX_IX_JUST) {
	Glist p      = g->next->u.v;
	int from_id  = p->u.i;
	int into_id  = p->next->next->u.i;
	anc = get_clanc(from_id, anc);
	anc = get_clanc(into_id, anc);
      }
      else {
	fatal_error("get_clanc, unknown rule.");
      }
      g = g->next->next;
    }
  }
  return anc;
}  /* get_clanc */

/*************
 *
 *   get_clause_ancestors()
 *
 *************/

/* DOCUMENTATION
This routine returns the Glist of clauses that are ancestors of Clause c,
including clause c.  The result is sorted (increasing) by ID.
*/

/* PUBLIC */
Glist get_clause_ancestors(Clause c)
{
  Glist ancestors = get_clanc(c->id, NULL);
  return ancestors;
}  /* get_clause_ancestors */

/*************
 *
 *   demod_just()
 *
 *************/

/* DOCUMENTATION
This routine builds and returns a justification list for demodulation.
*/

/* PUBLIC */
Glist demod_just(Glist g)
{
  return plist_append(ilist_append(NULL, DEMOD_JUST), g);
}  /* demod_just */

