#include "lindex.h"

/* Private definitions and types */

/*
 * memory management
 */

static Lindex lindex_avail;
static unsigned lindex_gets, lindex_frees, lindex_avails;

/* end of counters for each type */

/*************
 *
 *    Lindex get_lindex()
 *
 *************/

static
Lindex get_lindex(void)
{
  Lindex p;

  lindex_gets++;
  if (lindex_avail == NULL)
    p = tp_alloc(sizeof(struct lindex));
  else {
    lindex_avails--;
    p = lindex_avail;
    lindex_avail = lindex_avail->next;
  }
  /* initialization */	
  	
  p->pos = NULL;
  p->neg = NULL;
  p->next = NULL;

  /* end of initialization */	
  return(p);
}  /* get_lindex */

/*************
 *
 *    free_lindex()
 *
 *************/

static
void free_lindex(Lindex p)
{
  lindex_frees++;
  lindex_avails++;
  p->next = lindex_avail;
  lindex_avail = p;
}  /* free_lindex */

/* end of get and free routines for each type */

/*************
 *
 *   fprint_lindex_mem()
 *
 *************/

/* DOCUMENTATION
This routine prints (to FILE *fp) memory usage statistics for data types
associated with Lindex.
The Boolean argument heading tells whether to print a heading on the table.
*/

/* PUBLIC */
void fprint_lindex_mem(FILE *fp, BOOL heading)
{
  int n;
  if (heading)
    fprintf(fp, "  type (bytes each)        gets      frees     in use      avail      bytes\n");

  n = sizeof(struct lindex);
  fprintf(fp, "lindex (%4d)       %11u%11u%11u%11u%9.1f K\n",
	  n, lindex_gets, lindex_frees,
	  lindex_gets - lindex_frees, lindex_avails,
	  (((lindex_gets - lindex_frees) + lindex_avails) * n) / 1024.);

  /* end of printing for each type */
  
}  /* fprint_lindex_mem */

/*************
 *
 *   p_lindex_mem()
 *
 *************/

/* DOCUMENTATION
This routine prints (to stdout) memory usage statistics for data types
associated with Lindex.
*/

/* PUBLIC */
void p_lindex_mem()
{
  fprint_lindex_mem(stdout, 1);
}  /* p_lindex_mem */

/*
 *  end of memory management
 */

/*************
 *
 *   lindex_init()
 *
 *************/

/* DOCUMENTATION
This routine allocates and returns a literal index (Lindex), which is
a pair of Mindexes, one for positive literals, and one for negative literals.
The first three parameters are for the positive literal Mindex,
and the second three are for the negative.
<P>
<UL>
<LI><TT>Mindextype: {LINEAR, FPA, DISCRIM, DISCRIM_BIND}</TT>
<LI><TT>Uniftype:   {ORDINARY_UNIF, BACKTRACK_UNIF}</TT>
<LI><TT>fpa_depth: </TT>depth of FPA indexing
    (ignored for other index types).
</UL>
See the routine mindex_init() for further information.
*/

/* PUBLIC */
Lindex lindex_init(Mindextype pos_mtype, Uniftype pos_utype, int pos_fpa_depth,
		   Mindextype neg_mtype, Uniftype neg_utype, int neg_fpa_depth)
{
  Lindex ldx = get_lindex();

  ldx->pos = mindex_init(pos_mtype, pos_utype, pos_fpa_depth);
  ldx->neg = mindex_init(neg_mtype, neg_utype, neg_fpa_depth);

  return ldx;
}  /* lindex_init */

/*************
 *
 *   lindex_destroy()
 *
 *************/

/* DOCUMENTATION
This frees all the memory associated with a Lindex.  Do not refer
to the Lindex after calling this routine.
*/

/* PUBLIC */
void lindex_destroy(Lindex ldx)
{
  mindex_destroy(ldx->pos);
  mindex_destroy(ldx->neg);
  free_lindex(ldx);
}  /* lindex_destroy */

/*************
 *
 *   lindex_insert()
 *
 *************/

/* DOCUMENTATION
This routine inserts the literals of a clause into a Lindex.
*/

/* PUBLIC */
void lindex_insert(Lindex ldx, Clause c)
{
  Literal lit;
  for (lit = c->first_literal; lit != NULL; lit = lit->next) {
    if (lit->sign)
      mindex_insert(ldx->pos, lit->atom);
    else
      mindex_insert(ldx->neg, lit->atom);
  }
}  /* lindex_insert */

/*************
 *
 *   lindex_delete()
 *
 *************/

/* DOCUMENTATION
This routine deletes the literals of a clause from a Lindex.
*/

/* PUBLIC */
void lindex_delete(Lindex ldx, Clause c)
{
  Literal lit;
  for (lit = c->first_literal; lit != NULL; lit = lit->next) {
    if (lit->sign)
      mindex_delete(ldx->pos, lit->atom);
    else
      mindex_delete(ldx->neg, lit->atom);
  }
}  /* lindex_delete */

/*************
 *
 *   lindex_empty()
 *
 *************/

/* DOCUMENTATION
This Boolean routine checks if an Lindex is empty, that is, has no
atoms.  It must exist (be non-NULL).
*/

/* PUBLIC */
BOOL lindex_empty(Lindex idx)
{
  return mindex_empty(idx->pos) && mindex_empty(idx->neg);
}  /* lindex_empty */

/*************
 *
 *   lindex_backtrack()
 *
 *************/

/* DOCUMENTATION
This Boolean function checks if either of the Mindex components (pos, neg)
uses backtrack unification.
*/

/* PUBLIC */
BOOL lindex_backtrack(Lindex idx)
{
  return (idx->pos->unif_type == BACKTRACK_UNIF ||
	  idx->neg->unif_type == BACKTRACK_UNIF);
}  /* lindex_backtrack */
