#include "mace.h"

/* TO DO
 *  1. quasigroup, bijection.
 *  DONE 2. domain elements in input.
 *  DONE 3. distinct_constants option.
 *  4. break up complicated ground clauses.
 *  5. break up complicated nonground clauses.
 *  DONE (for unsorted) 6. crash if domain el too big for domain.
 *  DONE (for unsorted 7. put domain elements in term form of interp.
 *  8. Sorted problems (some of it is already in place).
 *  9. memory_out, time_out for preprocessing.
 */

/* #define PRINT_ADDED_CLAUSES */

/* Private definitions and types */

#define MAX_MACE_LITERALS 50     /* per propositional clause */
#define MAX_MACE_VARS     50     /* per propositional clause */
#define MAX_MACE_ARITY    4      /* per propositional atom */

typedef struct macesym * Macesym;
typedef struct sort * Sort;

/* A mace_state struct maintains information from one call to the next. */

struct mace_state {
  DP_state    dp;     /* for Davis-Putnam engine */
  Macesym     m;      /* symbol data */
  Sort        s;      /* sort data */
  BOOL        elements_in_input;
  Mace_state  next;   /* memory management */
};

/* A macesym struct holds data for each rigid symbol in the flat clauses. */

struct macesym {
  Macesym  next;
  int      base;                  /* prop. atom numbers start here */
  int      top;                   /* prop. atom numbers end at top-1 */
  int      arity;                 /* ordinary */
  int      mace_arity;            /* (function ? arity+1 : arity) */
  struct {
    int  sort;                    /* index of sort */
    int  n;                       /* size of sort */
    int  mult;                    /* for constructing prop. variables */
  }   args[MAX_MACE_ARITY];
  int      symnum;                /* ordinary symbol ID */
  int      mace_symnum;           /* symbol ID for mace */
  int      assigned_value;        /* for distinct constants option */
  int      function;              /* properties of symbol */
  int      quasigroup;
  int      bijection;
  int      equality;
  int      order;
};

/* A sort struct holds the name and size of each sort. */

struct sort {  /* stored as an array, indexed by sort ID */
  int  symnum;   /* symbol number of name of sort */
  int  n;        /* size of sort */
};

/* The following global variables are the translation to the propositional
 * problem.  They are not needed after the mace_first_model() call.  (The
 * data needed for mace_next_model() are kept in a mace_state struct.)
 */

static int      Eq_symnum; /* symnum for "=" 2. */
static int      Lt_symnum; /* symnum for "<" 2. */
static Glist Functions; /* symnums of pred syms that are really funcs */
static Macesym  Symbols;   /* list of syms for "foreach" access */
static Macesym  *Symarray; /* array of ptrs to syms for access by symnum */
static int      Max_reserved_element = -1;  /* for distinct_constants opt */
static int      *Sn_to_el; /* mapping for fixed domain elements */

static DP_state Dp;       /* Davis-Putnam state */
static int      *Units;   /* Array of propositional units for preprocessing. */
static int      Clauses_generated;  /* propositional */
static int      Clauses_inserted;   /* sent to DP_state */

static struct sort *Sorts;
static int Num_sorts;

/* The following macros take sequences of integers and give unique integers
 * that are the propositional variables.  The input, for example b,i,j,
 * represents application of symbol with "base" b to elements of the
 * domain.
 */

#define ATOM0(m) (m->base)
#define ATOM1(m,i) (i + m->base)
#define ATOM2(m,i,j) ((i)*m->args[0].mult + j + m->base)
#define ATOM3(m,i,j,k) ((i)*m->args[0].mult + (j)*m->args[1].mult + k + m->base)
#define ATOM4(m,i,j,k,l) ((i)*m->args[0].mult + (j)*m->args[1].mult + (k)*m->args[2].mult + l + m->base)

/*
 * memory management
 */

static Macesym macesym_avail;
static unsigned macesym_gets, macesym_frees, macesym_avails;

static Mace_state mace_state_avail;
static unsigned mace_state_gets, mace_state_frees, mace_state_avails;

static Mace_state mace_state_avail;
static unsigned mace_state_gets, mace_state_frees, mace_state_avails;

/* end of counters for each type */

/*************
 *
 *    Macesym get_macesym()
 *
 *************/

static
Macesym get_macesym(void)
{
  Macesym p;

  macesym_gets++;
  if (macesym_avail == NULL)
    p = tp_alloc(sizeof(struct macesym));
  else {
    macesym_avails--;
    p = macesym_avail;
    macesym_avail = macesym_avail->next;
  }
  /* initialization */	
  	
  p->next = NULL;
  p->base = 0;
  p->top = 0;
  p->arity = 0;
  p->mace_arity = 0;
  p->symnum = 0;
  p->assigned_value = 0;
  p->function = 0;
  p->quasigroup = 0;
  p->bijection = 0;
  p->equality = 0;
  p->order = 0;

  /* end of initialization */	
  return(p);
}  /* get_macesym */

/*************
 *
 *    free_macesym()
 *
 *************/

static
void free_macesym(Macesym p)
{
  macesym_frees++;
  macesym_avails++;
  p->next = macesym_avail;
  macesym_avail = p;
}  /* free_macesym */

/*************
 *
 *    Mace_state get_mace_state()
 *
 *************/

static
Mace_state get_mace_state(void)
{
  Mace_state p;

  mace_state_gets++;
  if (mace_state_avail == NULL)
    p = tp_alloc(sizeof(struct mace_state));
  else {
    mace_state_avails--;
    p = mace_state_avail;
    mace_state_avail = mace_state_avail->next;
  }
  /* initialization */	

  p->dp = NULL;
  p->m = NULL;
  p->s = NULL;
  p->elements_in_input = 0;
  p->next = NULL;

  /* end of initialization */	
  return(p);
}  /* get_mace_state */

/*************
 *
 *    free_mace_state()
 *
 *************/

static
void free_mace_state(Mace_state p)
{
  mace_state_frees++;
  mace_state_avails++;
  p->next = mace_state_avail;
  mace_state_avail = p;
}  /* free_mace_state */

/* end of get and free routines for each type */

/*************
 *
 *   fprint_mace_mem()
 *
 *************/

/* DOCUMENTATION
This routine prints (to FILE *fp) memory usage statistics for data types
associated with mace.
The Boolean argument heading tells whether to print a heading on the table.
*/

/* PUBLIC */
void fprint_mace_mem(FILE *fp, BOOL heading)
{
  int n;
  if (heading)
    fprintf(fp, "  type (bytes each)        gets      frees     in use      avail      bytes\n");

  n = sizeof(struct macesym);
  fprintf(fp, "macesym (%4d)      %11u%11u%11u%11u%9.1f K\n",
	  n, macesym_gets, macesym_frees,
	  macesym_gets - macesym_frees, macesym_avails,
	  (((macesym_gets - macesym_frees) + macesym_avails) * n) / 1024.);

  n = sizeof(struct mace_state);
  fprintf(fp, "mace_state (%4d)   %11u%11u%11u%11u%9.1f K\n",
	  n, mace_state_gets, mace_state_frees,
	  mace_state_gets - mace_state_frees, mace_state_avails,
	  (((mace_state_gets - mace_state_frees) + mace_state_avails) * n) / 1024.);

  /* end of printing for each type */
  
}  /* fprint_mace_mem */

/*************
 *
 *   p_mace_mem()
 *
 *************/

/* DOCUMENTATION
This routine prints (to stdout) memory usage statistics for data types
associated with mace.
*/

/* PUBLIC */
void p_mace_mem()
{
  fprint_mace_mem(stdout, 1);
}  /* p_mace_mem */

/*
 *  end of memory management
 */

/*************
 *
 *   sn_to_element_insert()
 *
 *   Add symnum->element to the mapping from symnums to elements.
 *
 *************/

static
void sn_to_element_insert(int symnum, int element)
{
  if (Sn_to_el == NULL) {
    int i;
    int n = greatest_symnum() + 1;
    Sn_to_el = my_malloc(n * sizeof(int));
    for (i = 0; i < n; i++)
      Sn_to_el[i] = -1;
  }
  Sn_to_el[symnum] = element;
}  /* sn_to_element_insert */

/*************
 *
 *   sn_to_element()
 *
 *   To what domain element does the symnum map?
 *   If nothing, then -1 is returned.
 *
 *************/

static
int sn_to_element(int symnum)
{
  if (Sn_to_el == NULL)
    return -1;
  else
    return Sn_to_el[symnum];
}  /* sn_to_element */

/*************
 *
 *   domain_element()
 *
 *   Does the given term represent a domain element?
 *
 *************/

static
BOOL domain_element(Term t)
{
  return (CONSTANT(t) && sn_to_element(SYMNUM(t)) != -1);
}  /* domain_element */

#ifdef PRINT_ADDED_CLAUSES

/*************
 *
 *   decode_int()
 *
 *   Given a propositional variable (an integer), find the atom that
 *   it represents.  This is the inverse of the ATOM macros.
 *   This is used mostly for printing ground clauses in a readable way.
 *
 *************/

static
int decode_int(int atom, char **func, int *arity, int *args)
{
  int i, c, found;
  Macesym m;

  for (i = 0, found = 0; !found; i++) {
    m = Symarray[i];
    found = (m != NULL && atom >= m->base && atom < m->top);
  }

  c = atom - m->base;
  *func = sn_to_str(m->symnum);
  *arity = m->arity;

  switch (*arity) {
  case 1:
    args[0] = c;
    break;
  case 2:
    args[1] = c % m->args[1].n; c = c / m->args[1].n;
    args[0] = c;
    break;
  case 3:
    args[2] = c % m->args[2].n; c = c / m->args[2].n;
    args[1] = c % m->args[1].n; c = c / m->args[1].n;
    args[0] = c;
    break;
  case 4:
    args[3] = c % m->args[3].n; c = c / m->args[3].n;
    args[2] = c % m->args[2].n; c = c / m->args[2].n;
    args[1] = c % m->args[1].n; c = c / m->args[1].n;
    args[0] = c;
    break;
  }
  return(1);
}  /* decode_int */

/*************
 *
 *   decode_and_p_clause()
 *
 *   Decode and print a propositional clause.
 *
 *************/

static void decode_and_p_clause(int c[], int n)
{
  char *name, *sign;
  int arity, args[MAX_MACE_ARITY], i, j;

  for (i = 0; i < n; i++) {
    sign = (c[i] < 0 ? "~ " : "");
    decode_int(abs(c[i]), &name, &arity, args);
    printf("  %s%s(", sign, name);
    for (j = 0; j < arity; j++) {
      printf("%d", args[j]);
      if (j < arity-1)
	printf(",");
    }
    printf(")");
    printf(" [%d] ", c[i]);
  }
  printf("\n");
}  /* decode_and_p_clause */
#endif

/*************
 *
 *   sort_clause() -- propositional clause, by atom.
 *
 *************/

static void sort_clause(int c[], int n)
{
  int i, j, temp;

  for (i = 0; i < n-1; i++)
    for (j = i+1; j < n; j++)
      if (abs(c[j]) < abs(c[i])) {
	temp = c[i]; c[i] = c[j]; c[j] = temp;
      }

}  /* sort_clause */

/*************
 *
 *   unit_subsumed()
 *
 *   Is a given propositional clause subsumed by anything in the Units array?
 *
 *************/

static int unit_subsumed(int c[], int n)
{
  int i, a, sign;
  for (i = 0; i < n; i++) {
    a = abs(c[i]);
    sign = (c[i] < 0) ? -1 : 1;
    if (Units[a] == sign)
      return(1);
  }
  return(0);
}  /* unit_subsumed */

/*************
 *
 *   tautology() -- propositional; assume sorted by atom.
 *
 *************/

static int tautology(int c[], int n)
{
  int i;
  for (i = 0; i < n-1; i++)
    if (c[i] == -c[i+1])
      return(1);
  return(0);
}  /* tautology */

/*************
 *
 *   merge_lits() -- propositional; assume sorted by atom.
 *
 *************/

static int merge_lits(int c[], int n)
{
  int i, j, m;
  m = n;
  for (i = 0; i < m-1; i++)
    if (c[i] == c[i+1]) {
      for (j = i+1; j < m-1; j++)
	c[j] = c[j+1];
      m--;
      i--;
    }
  return(m);
}  /* merge_lits */

/*************
 *
 *   unit_delete()
 *
 *   Given a propositional clause, do unit resolution with Units array.
 *
 *************/

int unit_delete(int c[], int n)
{
  int i, j, l, a, v, m;
  m = n;
  for (i = 0; i < m; i++) {
    l = c[i]; a = abs(l); v = Units[a];
    if ((v == -1 && l > 0) || (v == 1 && l < 0)) {
      for (j = i; j < m-1; j++)
	c[j] = c[j+1];
      m--;
      i--;
    }
  }

  return(m);
}  /* unit_delete */

/*************
 *
 *   add_clause()
 *
 *   Given a propositional clause, do some preprocessing, and if it
 *   passes the tests, send it to insert_dp_clause(), which is part
 *   of the interface to the DP procedure.
 *
 *************/

static void add_clause(int c[], int n)
{
  Clauses_generated++;
#ifdef PRINT_ADDED_CLAUSES
  decode_and_p_clause(c,n);
#endif
  n = unit_delete(c, n);
  if (n == 0) {
    fatal_error("add_clause: propositionally unsatisfiable");
  }
  sort_clause(c, n);
  if (!unit_subsumed(c, n) && !tautology(c, n)) {
    int rc;
    n = merge_lits(c, n);
    if (n == 1)
      Units[abs(c[0])] = c[0] < 0 ? -1 : 1;
#ifdef PRINT_ADDED_CLAUSES
    decode_and_p_clause(c,n);
#endif
    rc = dp_append_clause(Dp, c, n);
    if (rc == MEMORY_OUT_RETURN)
      fatal_error("add_clause, dp_append_clause MEMORY_OUT_RETURN");
    Clauses_inserted++;
  }
}  /* add_clause */

/*************
 *
 *   add_2clause()
 *
 *   Given 2 literals, make a 2-clause and send it to add_clause.
 *
 *************/

static void add_2clause(int a, int b)
{
  int c[2];
  c[0] = a; c[1] = b;
  add_clause(c, 2);
}  /* add_2clause */

/*************
 *
 *   add_unit()
 *
 *   Build a literal and send it to add_clause.
 *
 *************/

static void add_unit(int sign, Macesym s, int a0, int a1, int a2, int a3)
{
  int c[1];

  if (a0 >= s->args[0].n || a1 >= s->args[1].n || a2 >= s->args[2].n || a3 >= s->args[3].n)
    fatal_error("add_unit: element out of domain");
  if (abs(sign) != 1)
    fatal_error("add_unit: sign must be 1 or -1");

#if 0
  printf("add_unit: %s%s %d %d %d %d.\n", 
	 sign==1?"":"-", Symbols[functor].name, a0, a1, a2, a3);
#endif
    
  switch(s->arity) {
  case 0: c[0] = sign * ATOM0(s); break;
  case 1: c[0] = sign * ATOM1(s, a0); break;
  case 2: c[0] = sign * ATOM2(s, a0, a1); break;
  case 3: c[0] = sign * ATOM3(s, a0, a1, a2); break;
  case 4: c[0] = sign * ATOM4(s, a0, a1, a2, a3); break;
  }
  add_clause(c, 1);
}  /* add_unit */

/*
 * The following set of routines assert various properties of functions.
 * Recall that an n-ary function is handled as an (n+1)-ary relation,
 * so we have to say that the function is well-defined and closed.
 * Similar routines can be used to assert that functions are 1-1, onto,
 * cancellative, etc.
 */

/*************
 *
 *   well_defined_0() - constant cannot have 2 values.
 *
 *************/

static void well_defined_0(Macesym s)
{
  int y, z;
  int n0 = s->args[0].n;
    
  /* -By | -Bz, for y < z */
  for (y = 0; y < n0; y++)
    for (z = y+1; z < n0; z++)
      add_2clause( - ATOM1(s,y), - ATOM1(s,z));
}  /* well_defined_0 */

/*************
 *
 *   well_defined_1()
 *
 *************/

static void well_defined_1(Macesym s)
{
  int x, y, z;
  int n0 = s->args[0].n;
  int n1 = s->args[1].n;
    
  /* -Bxy | -Bxz, for y < z */
  for (x = 0; x < n0; x++)
    for (y = 0; y < n1; y++)
      for (z = y+1; z < n1; z++)
	add_2clause( - ATOM2(s,x,y), - ATOM2(s,x,z));
}  /* well_defined_1 */

/*************
 *
 *   well_defined_2()
 *
 *************/

static void well_defined_2(Macesym s)
{
  int u, x, y, z;
  int n0 = s->args[0].n;
  int n1 = s->args[1].n;
  int n2 = s->args[2].n;

  /* -Buxy | -Buxz, for y < z */
  for (u = 0; u < n0; u++)
    for (x = 0; x < n1; x++)
      for (y = 0; y < n2; y++)
	for (z = y+1; z < n2; z++)
	  add_2clause( - ATOM3(s,u,x,y), - ATOM3(s,u,x,z));
}  /* well_defined_2 */

/*************
 *
 *   well_defined_3()
 *
 *************/

static void well_defined_3(Macesym s)
{
  int v, u, x, y, z;
  int n0 = s->args[0].n;
  int n1 = s->args[1].n;
  int n2 = s->args[2].n;
  int n3 = s->args[3].n;

  /* -Bvuxy | -Bvuxz, for y < z */
  for (v = 0; v < n0; v++)
    for (u = 0; u < n1; u++)
      for (x = 0; x < n2; x++)
	for (y = 0; y < n3; y++)
	  for (z = y+1; z < n3; z++)
	    add_2clause( - ATOM4(s,v,u,x,y), - ATOM4(s,v,u,x,z));
}  /* well_defined_3 */

/*************
 *
 *   closed_0() - value of constant is one of the domain elements.
 *
 *************/

static void closed_0(Macesym s)
{
  int x, c[MAX_MACE_LITERALS];
  int n0 = s->args[0].n;

  /* B0 | B1 | .. | Bn-1 */
  for (x = 0; x < n0; x++)
    c[x] = ATOM1(s,x);
  add_clause(c, n0);
}  /* closed_0 */

/*************
 *
 *   closed_1()
 *
 *************/

static void closed_1(Macesym s)
{
  int x, y, c[MAX_MACE_LITERALS];
  int n0 = s->args[0].n;
  int n1 = s->args[1].n;

  /* Bx0 | Bx1 | .. | Bxn-1, 0<=x<=n-1 */
  for (x = 0; x < n0; x++) {
    for (y = 0; y < n1; y++)
      c[y] = ATOM2(s,x,y);
    add_clause(c, n1);
  }
}  /* closed_1 */

/*************
 *
 *   closed_2()
 *
 *************/

static void closed_2(Macesym s)
{
  int x, y, z, c[MAX_MACE_LITERALS];
  int n0 = s->args[0].n;
  int n1 = s->args[1].n;
  int n2 = s->args[2].n;

  /* Bxy0 | Bxy1 | .. | Bxyn-1, 0<=x<=n-1, 0<=y<=n-1 */
  for (x = 0; x < n0; x++)
    for (y = 0; y < n1; y++) {
      for (z = 0; z < n2; z++)
	c[z] = ATOM3(s,x,y,z);
      add_clause(c, n2);
    }
}  /* closed_2 */

/*************
 *
 *   closed_3()
 *
 *************/

static void closed_3(Macesym s)
{
  int x, y, z, u, c[MAX_MACE_LITERALS];
  int n0 = s->args[0].n;
  int n1 = s->args[1].n;
  int n2 = s->args[2].n;
  int n3 = s->args[3].n;

  /* Bxyz0 | Bxyz1 | .. | Bxyzn-1, 0<=x<=n-1, 0<=y<=n-1, 0<=z<=n-1 */
  for (x = 0; x < n0; x++)
    for (y = 0; y < n1; y++)
      for (z = 0; z < n2; z++) {
	for (u = 0; u < n3; u++)
	  c[u] = ATOM4(s,x,y,z,u);
	add_clause(c, n3);
      }
}  /* closed_3 */

/*************
 *
 *   well_defined()
 *
 *************/

void well_defined(Macesym s)
{
  switch(s->arity) {
  case 1: well_defined_0(s); break;
  case 2: well_defined_1(s); break;
  case 3: well_defined_2(s); break;
  case 4: well_defined_3(s); break;
  default: fatal_error("well_defined, bad arity");
  }
}  /* well_defined */

/*************
 *
 *   closed()
 *
 *************/

void closed(Macesym s)
{
  switch(s->arity) {
  case 1: closed_0(s); break;
  case 2: closed_1(s); break;
  case 3: closed_2(s); break;
  case 4: closed_3(s); break;
  default: fatal_error("closed, bad arity");
  }
}  /* closed */

/*************
 *
 *   one_one()
 *
 *************/

static void one_one(Macesym s)
{
  int x, y, z;
  int n0 = s->args[0].n;
  int n1 = s->args[1].n;

  /* -Bxy | -Bzy, x < z */
  for (y = 0; y < n1; y++)
    for (x = 0; x < n0; x++)
      for (z = x+1; z < n0; z++)
	add_2clause( - ATOM2(s,x,y), - ATOM2(s,z,y));

}  /* one_one */

/*************
 *
 *   onto()
 *
 *************/

static void onto(Macesym s)
{
  int x, y, c[MAX_MACE_LITERALS];
  int n0 = s->args[0].n;
  int n1 = s->args[1].n;

  /* B0x | B1x | .. | Bn-1x, 0<=x<=n-1 */

  for (x = 0; x < n1; x++) {
    for (y = 0; y < n0; y++)
      c[y] = ATOM2(s,y,x);
    add_clause(c, n1);
  }
}  /* onto */

/*************
 *
 *   left_cancel()
 *
 *************/

static void left_cancel(Macesym s)
{
  int u, x, y, z;
  int n0 = s->args[0].n;
  int n1 = s->args[1].n;
  int n2 = s->args[2].n;

  /* -Buyx | -Buzx, for y < z */

  for (u = 0; u < n0; u++)
    for (x = 0; x < n2; x++)
      for (y = 0; y < n1; y++)
	for (z = y+1; z < n1; z++)
	  add_2clause( - ATOM3(s,u,y,x), - ATOM3(s,u,z,x));

}  /* left_cancel */

/*************
 *
 *   right_cancel()
 *
 *************/

static void right_cancel(Macesym s)
{
  int u, x, y, z;
  int n0 = s->args[0].n;
  int n1 = s->args[1].n;
  int n2 = s->args[2].n;

  /* -Byux | -Bzux, for y < z */

  for (u = 0; u < n1; u++)
    for (x = 0; x < n2; x++)
      for (y = 0; y < n0; y++)
	for (z = y+1; z < n0; z++)
	  add_2clause( - ATOM3(s,y,u,x), - ATOM3(s,z,u,x));

}  /* right_cancel */

/*************
 *
 *   left_onto()
 *
 *************/

static void left_onto(Macesym s)
{
  int x, y, z, c[MAX_MACE_LITERALS];
  int n0 = s->args[0].n;
  int n1 = s->args[1].n;
  int n2 = s->args[2].n;

  /* Bx0y | Bx1y | .. | Bxn-1y, 0<=x<=n-1, 0<=y<=n-1 */

  for (x = 0; x < n0; x++)
    for (y = 0; y < n2; y++) {
      for (z = 0; z < n1; z++)
	c[z] = ATOM3(s,x,z,y);
      add_clause(c, n1);
    }
}  /* left_onto */

/*************
 *
 *   right_onto()
 *
 *************/

static void right_onto(Macesym s)
{
  int x, y, z, c[MAX_MACE_LITERALS];
  int n0 = s->args[0].n;
  int n1 = s->args[1].n;
  int n2 = s->args[2].n;

  /* B0xy | B1xy | .. | Bn-1xy, 0<=x<=n-1, 0<=y<=n-1 */

  for (x = 0; x < n1; x++)
    for (y = 0; y < n2; y++) {
      for (z = 0; z < n0; z++)
	c[z] = ATOM3(s,z,x,y);
      add_clause(c, n0);
    }
}  /* right_onto */

/*************
 *
 *   setup_symbol_array()
 *
 *************/

static
void setup_symbol_array(Macesym syms)
{
  Macesym m;
  int max = 0;
  int i;

  /* Find the largest symnum. */
  for (m = syms, i = 0; m != NULL; m = m->next, i++) {
    max = (m->symnum > max ? m->symnum : max);
  }
  /* Allocate an array for access by symnum. */
  Symarray = my_malloc((max+1) * sizeof(Macesym));
  for (i = 0; i < max+1; i++)
    Symarray[i] = NULL;

  for (m = syms; m != NULL; m = m->next)
    Symarray[m->symnum] = m;
}  /* setup_symbol_array */

/*************
 *
 *   setup_sort_array()
 *
 *************/

static
void setup_sort_array(int domain_size)
{
  int n = 1;
  /* Only one sort for now. */

  Sorts = my_malloc(n * sizeof(struct sort));
  Sorts[0].n = domain_size;
  Sorts[0].symnum = str_to_sn("univ", 0);
  Num_sorts = 1;
}  /* setup_sort_array */

/*************
 *
 *   fix_bases_and_mults_unsorted()
 *
 *************/

static
int fix_bases_and_mults_unsorted(Macesym syms, int domain_size)
{
  int nextbase = 1;
  Macesym m;
  for (m = syms; m != NULL; m = m->next) {
    int i;
    int mult;
    m->base = nextbase;   /* start of atoms for this symbol */

    /* Fill in the domain sizes.  All have size n. */
    for (i = 0; i < m->arity; i++) {
      m->args[i].n = domain_size;
      m->args[i].sort = 0;
    }

    /* Fill in the multipliers for constructing prop atoms.
     * (This part is more general than it needs to be.)
     */
    mult = 1;
    for (i = m->arity-1; i >= 0; i--) {
      m->args[i].mult = mult;
      mult *= m->args[i].n;
    }
    nextbase += mult;
    m->top = nextbase;
  }
  return nextbase-1;
}  /* fix_bases_and_mults_unsorted */

/*************
 *
 *   macesym_member()
 *
 *************/

static
Macesym macesym_member(int sn, Macesym m)
{
  if (m == NULL)
    return NULL;
  else if (m->symnum == sn)
    return m;
  else
    return macesym_member(sn, m->next);
}  /* macesym_member */

/*************
 *
 *   setup_symbol_data()
 *
 *************/

static
Macesym setup_symbol_data(Clist lst, int domain_size)
{
  Macesym m = NULL;
  Macesym last;
  Clist_pos p;

  /* Include "=" first. */

  m = get_macesym();
  m->symnum = Eq_symnum;
  m->arity = 2;
  m->equality = 1;
  last = m;

  for (p = lst->first; p != NULL; p = p->next) {
    Clause c = p->c;
    Literal lit;
    for (lit = c->first_literal; lit != NULL; lit = lit->next) {
      Term a = lit->atom;
      int sn = SYMNUM(a);
      Macesym m2 = macesym_member(sn, m);
      if (m2 == NULL) {
	m2 = get_macesym();
	m2->symnum = sn;
	m2->function = ilist_member(Functions, sn);
	m2->arity = sn_to_arity(sn);
	if (m2->function && m2->arity > MAX_MACE_ARITY)
	  fatal_error("arity of function symbols is limited to 3");
	else if (!m2->function && m2->arity > MAX_MACE_ARITY)
	  fatal_error("arity of relation symbols is limited to 4");
	m2->equality =  !m2->function && sn == Eq_symnum;
	m2->order    =  !m2->function && sn == Lt_symnum;
	last->next = m2;
	last = m2;
      }
    }
  }
  return m;
}  /* setup_symbol_data */

/*************
 *
 *   process_variables()
 *
 *   Check that all occurrences of each variable are in a position of the
 *   same sort, collect the domain sizes of the variables into an array,
 *   and return the biggest variable.
 *
 *************/

static int process_variables(Clause c, int domains[])
{
  int i, maxvar;
  Macesym m;
  Literal lit;

  for (i = 0; i < MAX_MACE_VARS; i++)
    domains[i] = -1;
  maxvar = -1;

  for (lit = c->first_literal; lit != NULL; lit = lit->next) {
    Term atom = lit->atom;
    int b = biggest_variable(atom);
    if (b >= MAX_MACE_VARS)
      fatal_error("process_variables, variable too big.");
    maxvar = (b > maxvar ? b : maxvar);

    m = Symarray[SYMNUM(atom)];

    for (i = 0; i < atom->arity; i++) {
      Term a = atom->args[i];
      int v;
      if (domain_element(a))
	;  /* ok */
      else if (VARIABLE(a)) {
	v = VARNUM(a);
	if (domains[v] == -1)
	  domains[v] = m->args[i].sort;
	else if (domains[v] != m->args[i].sort)
	  fatal_error("process_variables, badly sorted clause");
      }
      else
	fatal_error("process_variables, not variable or domain element");
    }
  }  /* foreach literal */

  /* Now replace sort indexes with sort sizes. */

  for (i = 0; i <= maxvar; i++)
    domains[i] = Sorts[domains[i]].n;

  return maxvar;
}  /* process_variables */

/*************
 *
 *   instances_recurse()
 *
 *************/

static
void instances_recurse(Clause c, int *vals, int *domains,
		       int nextvar, int nvars)
{
  if (nextvar == nvars) {
    Literal lit;
    int prop_clause[MAX_MACE_LITERALS];
    int litnum = 0;
    for (lit = c->first_literal; lit != NULL; lit = lit->next, litnum++) {
      int a[MAX_MACE_ARITY];
      int sign = (lit->sign ? 1 : -1);
      Term atom = lit->atom;
      Macesym m = Symarray[SYMNUM(atom)];
      int i;
      for (i = 0; i < atom->arity; i++) {
	Term arg = atom->args[i];
	if (VARIABLE(arg))
	  a[i] = vals[VARNUM(arg)];
	else
	  a[i] = sn_to_element(SYMNUM(arg)); /* fixed domain element  */
      }

      switch (atom->arity) {
      case 0: prop_clause[litnum] = sign * ATOM0(m); break;
      case 1: prop_clause[litnum] = sign * ATOM1(m,a[0]); break;
      case 2: prop_clause[litnum] = sign * ATOM2(m,a[0],a[1]); break;
      case 3: prop_clause[litnum] = sign * ATOM3(m,a[0],a[1],a[2]); break;
      case 4: prop_clause[litnum] = sign * ATOM4(m,a[0],a[1],a[2],a[3]); break;
      }
    }
    add_clause(prop_clause, litnum);
  }
  else if (domains[nextvar] == -1)
    /* in case the current variable does not appear in the clause */
    instances_recurse(c, vals, domains, nextvar+1, nvars);
  else {
    int i;
    for (i = 0; i < domains[nextvar]; i++) {
      vals[nextvar] = i;
      instances_recurse(c, vals, domains, nextvar+1, nvars);
    }
  }
}  /* instances_recurse */

/*************
 *
 *   generate_prop_clauses()
 *
 *   Given a flattened general clause, generate the corresponding
 *   ground clauses over the domain and send them to add_clause.
 *
 *   Note that the limit on the arity is wired into the code.
 *
 *************/

static void generate_prop_clauses(Clause c)
{
  int biggest_var, vals[MAX_MACE_VARS], domains[MAX_MACE_VARS];
	
  biggest_var = process_variables(c, domains);

  /* Now domains[] has the domain sizes for all the variables. */
  
  instances_recurse(c, vals, domains, 0, biggest_var+1);

}  /* generate_prop_clauses */

/*************
 *
 *   yankable()
 *
 *   Should the term be yanked out for flattening?
 *
 *************/

static
BOOL yankable(Mace_state ms, Term t)
{
  if (domain_element(t))
    return FALSE;
  else
    return !VARIABLE(t);
}  /* yankable */

/*************
 *
 *   first_nonvar_arg()
 *
 *************/

static
Term first_nonvar_arg(Mace_state ms, Term t)
{
  Term nv = NULL;
  int i;
  for (i = 0; i < t->arity && nv == NULL; i++) {
    if (yankable(ms, t->args[i]))
      nv = t->args[i];
  }
  return nv;
}  /* first_nonvar_arg */

/*************
 *
 *   first_nonvar_subterm()
 *
 *   For equalities (pos or neg) don't consider the left argument.
 *
 *************/

static
Term first_nonvar_subterm(Mace_state ms, Clause c)
{
  Literal l;
  Term t = NULL;

  for (l = c->first_literal; l != NULL && t == NULL; l = l->next) {
    if (SYMNUM(l->atom) == Eq_symnum) {
      t = first_nonvar_arg(ms, l->atom->args[0]);
      if (t == NULL && yankable(ms, l->atom->args[1]))
	t = l->atom->args[1];
    }
    else
      t = first_nonvar_arg(ms, l->atom);
  }
  return t;
}  /* first_nonvar_subterm */

/*************
 *
 *   flip_backward_eqs()
 *
 *   Flip any equality literals (pos or neg) with var on left, nonvar on right.
 *
 *************/

static
void flip_backward_eqs(Clause c)
{
  Literal lit;
  for (lit = c->first_literal; lit != NULL; lit = lit->next) {
    Term a = lit->atom;
    if (SYMNUM(a) == Eq_symnum &&
	VARIABLE(a->args[0]) &&
	!VARIABLE(a->args[1])) {
      Term t = a->args[0];
      a->args[0] = a->args[1];
      a->args[1] = t;
    }
  }
}  /* flip_backward_eqs */

/*************
 *
 *   flatten_clause()
 *
 *   Return a flattened copy.
 *
 *************/

static
Clause flatten_clause(Mace_state ms, Clause c)
{
  Term t;
  Clause d = copy_clause(c);
  int next_var = greatest_variable_in_clause(d) + 1;
  Literal lit;

  t = first_nonvar_subterm(ms, d);
  while (t != NULL) {
    Literal l1, l2;
    Term atom;
    t = copy_term(t);
    atom = build_binary_term(Eq_symnum, t, get_variable_term(next_var));
    l1 = new_literal(0, atom);

    for (l2 = d->first_literal; l2 != NULL; l2 = l2->next)
      l2->atom = subst_term(l2->atom, t, atom->args[1]);

    append_literal(d, l1);
    next_var++;
    t = first_nonvar_subterm(ms, d);
  }

  if (!renumber_variables(d, MAX_MACE_VARS))
    fatal_error("flatten_clause, too many variables");

  /* Replace, e.g., f(x,y)=z with f(x,y,z), noting that f is a function. */

  for (lit = d->first_literal; lit != NULL; lit = lit->next) {
    Term a = lit->atom;
    if (SYMNUM(a) == Eq_symnum &&
	!VARIABLE(a->args[0]) &&
	!domain_element(a->args[0])) {
      Term t = a->args[0];
      Term newatom = get_rigid_term(sn_to_str(SYMNUM(t)), t->arity+1);
      int i;
      Functions = ilist_append(Functions, SYMNUM(newatom));
      for (i = 0; i < t->arity; i++) {
	newatom->args[i] = copy_term(t->args[i]);
      }
      newatom->args[i] = copy_term(lit->atom->args[1]);
      zap_term(lit->atom);
      lit->atom = newatom;
    }
  }

  return d;
}  /* flatten_clause */

/*************
 *
 *   flip_some_equations()
 *
 *   For each clause c, if it is a positive unit equality, with neither
 *   argument a variable or domain element, insert a flipped copy after c.
 *
 *************/

static
void flip_some_equations(Mace_state ms, Clist lst)
{
  Clist_pos p;
  for (p = lst->first; p != NULL; p = p->next) {
    Clause c = p->c;
    if (number_of_literals(c) == 1) {
      Literal lit = c->first_literal;
      if (lit->sign &&
	  SYMNUM(lit->atom) == Eq_symnum &&
	  !VARIABLE(lit->atom->args[0]) &&
	  !VARIABLE(lit->atom->args[1]) &&
	  !domain_element(lit->atom->args[0]) &&
	  !domain_element(lit->atom->args[1])) {
	Clause d = copy_clause(c);
	Term a = d->first_literal->atom;
	Term t = a->args[0];
	a->args[0] = a->args[1];
	a->args[1] = t;
	clist_insert_after(d, p);
	p = p->next;  /* now p is at d */
      }
    }
  }
}  /* flip_some_equations */

/*************
 *
 *   process_domain_elements_in_term()
 *
 *************/

static
void process_domain_elements_in_term(Mace_state ms, Term t, FILE *fp)
{
  if (CONSTANT(t)) {
    int e;
    BOOL rc = str_to_int(sn_to_str(SYMNUM(t)), &e);
    if (rc && e >= 0 && e < ms->s[0].n) {
      if (sn_to_element(SYMNUM(t)) == -1) {
	sn_to_element_insert(SYMNUM(t), e);
	fprintf(fp, "Constant %s is interpreted as a domain element.\n",
		sn_to_str(SYMNUM(t)));
	if (e > Max_reserved_element)
	  Max_reserved_element = e;
      }
    }
  }
  else if (COMPLEX(t)) {
    int i;
    for (i = 0; i < t->arity; i++)
      process_domain_elements_in_term(ms, t->args[i], fp);
  }
}  /* process_domain_elements_in_term */

/*************
 *
 *   assert_propositions()
 *
 *   This routine translates the first-order clauses to propositional
 *   clauses, and sends the propositional clauses to Davis-Putnam.
 *
 *************/

static
void assert_propositions(Mace_state ms,
			 Clist lst,
			 int domain_size,
			 BOOL distinct_constants,
			 BOOL flip_equations,
			 FILE *fp)
{
  Macesym m;
  Clist flats;
  Clist_pos p;
  int greatest_atom, i;

  /* First set some global variables. */
  Dp = ms->dp;  /* DP_state */
  Eq_symnum = str_to_sn("=", 2);

  setup_sort_array(domain_size);
  ms->s = Sorts;

  /* If there can be domain element in the clauses, identify them. */

  if (ms->elements_in_input) {
    for (p = lst->first; p != NULL; p = p->next) {
      Literal lit;
      for (lit = p->c->first_literal; lit != NULL; lit = lit->next) {
	int i;
	for (i = 0; i < lit->atom->arity; i++)
	  process_domain_elements_in_term(ms, lit->atom->args[i], fp);
      }
    }
  }

  for (p = lst->first; p != NULL; p = p->next)
    flip_backward_eqs(p->c);

  if (flip_equations)
    flip_some_equations(ms, lst);

  flats = clist_init("flat_clauses");
  for (p = lst->first; p != NULL; p = p->next) {
    Clause c = flatten_clause(ms, p->c);
    clist_append(c, flats);
  }
  fwrite_clause_list(fp, flats, FALSE);

  ms->m = setup_symbol_data(flats, domain_size);
  Symbols = ms->m;

  setup_symbol_array(ms->m);
  greatest_atom = fix_bases_and_mults_unsorted(ms->m, domain_size);

#if 0
  for (m = ms->m; m != NULL; m = m->next) {
    fprintf(fp,"%s, function=%d, equality=%d, base=%d.\n",
	    sn_to_str(m->symnum),
	    m->function,
	    m->equality,
	    m->base);
  }
#endif

  /* Allocate Units array for unit subsumption, unit deletion. */

  Units = my_malloc((greatest_atom+1) * sizeof(int));
  for (i = 1; i < greatest_atom+1; i++)
    Units[i] = 0;

  /* Assignments for distinct_constants option. */

  if (distinct_constants) {
    for (m = ms->m; m != NULL; m = m->next) {
      if (m->function && m->arity == 1) {
	if (Max_reserved_element+1 == domain_size)
	  fprintf(fp,"Cannot make constant %s distinct, domain is too small.\n",
		  sn_to_str(m->symnum));
	else {
	  Max_reserved_element++;
	  fprintf(fp,"Adding unit (distinct_constants) assigning %s %d.\n",
		 sn_to_str(m->symnum), Max_reserved_element);
	  add_unit(1, m, Max_reserved_element, -1, -1, -1);
	}
      }
    }
  }

  /* Assert units for equality and maybe other special relations. */

  for (m = ms->m; m != NULL; m = m->next) {
    if (m->equality) {
      int i, j, n;
      n = m->args[0].n;
      fprintf(fp, "Adding units for equality relation %s.\n",
	      sn_to_str(m->symnum));
      for (i = 0; i < n; i++)
	for (j = 0; j < n; j++)
	  add_unit((i == j ? 1 : -1), m, i, j, -1, -1);
    }
  }

  /* Process clauses with <= 2 literals. */

  for (p = flats->first; p != NULL; p = p->next) {
    Clause c = p->c;
    if (number_of_literals(c) <= 2) {
      fprintf(fp, "Processing clause: ");
      fwrite_clause(fp, c, FALSE);
      generate_prop_clauses(c);
      /* exit_if_over_time_limit(); */
    }
  }

  /* Functions well-defined and closed, and also special properties. */

  for (m = ms->m; m != NULL; m = m->next) {
    if (m->function) {
      int real_arity = m->arity-1;
      fprintf(fp, "function %s, arity %d, well-defined and closed.\n",
	      sn_to_str(m->symnum), real_arity);
      well_defined(m);
      closed(m);
      if (m->bijection && real_arity == 1) {
	fprintf(fp, "function %s, bijection.\n", sn_to_str(m->symnum));
	one_one(m);
	onto(m);
      }
      else if (m->quasigroup && real_arity == 2) {
	fprintf(fp, "function %s, quasigroup.\n", sn_to_str(m->symnum));
	left_cancel(m);
	right_cancel(m);
	left_onto(m);
	right_onto(m);
      }
    }
    else {  /* relation */
      fprintf(fp, "relation %s, arity %d.\n", sn_to_str(m->symnum), m->arity);
    }
  }

  /* Process clauses with > 2 literals. */

  for (p = flats->first; p != NULL; p = p->next) {
    Clause c = p->c;
    if (number_of_literals(c) > 2) {
      fprintf(fp, "Processing clause: ");
      fwrite_clause(fp, c, FALSE);
      generate_prop_clauses(c);
      /* exit_if_over_time_limit(); */
    }
  }

  /* Free all of the temp memory we malloced.
   * (We need to keep Symbols and Sorts in case we construct models.)
   */

  clist_zap(flats);
  zap_glist(Functions);
  my_free(Symarray);
  Symarray = NULL;
  my_free(Units);
  Units = NULL;
  if (Sn_to_el != NULL) {
    my_free(Sn_to_el);
    Sn_to_el = NULL;
  }

  fflush(fp);
}  /* assert_propositions */

/*************
 *
 *   function_value_0()
 *
 *************/

int function_value_0(DP_state dp, Macesym s)
{
  int b, k;
  int n0 = s->args[0].n;

  b = ATOM1(s, 0);
  for (k = 0; k < n0 && dp_atom_value(dp, b+k) != 1; k++);
  return (k == n0 ? -1 : k);
}  /* function_value_0 */

/*************
 *
 *   function_value_1()
 *
 *************/

int function_value_1(DP_state dp, Macesym s, int i)
{
  int b, k;
  int n1 = s->args[1].n;

  b = ATOM2(s, i, 0);
  for (k = 0; k < n1 && dp_atom_value(dp, b+k) != 1; k++);
  return(k == n1 ? -1 : k);
}  /* function_value_1 */

/*************
 *
 *   function_value_2()
 *
 *************/

int function_value_2(DP_state dp, Macesym s, int i, int j)
{
  int b, k;
  int n2 = s->args[2].n;

  b = ATOM3(s, i, j, 0);
  for (k = 0; k < n2 && dp_atom_value(dp, b+k) != 1; k++);
  return(k == n2 ? -1 : k);
}  /* function_value_2 */

/*************
 *
 *   function_value_3()
 *
 *************/

int function_value_3(DP_state dp, Macesym s, int i, int j, int k)
{
  int b, val;
  int n3 = s->args[3].n;

  b = ATOM4(s, i, j, k, 0);
  for (val = 0; val < n3 && dp_atom_value(dp, b+val) != 1; val++);
  return(val == n3 ? -1 : val);
}  /* function_value_3 */

/*************
 *
 *   print_table_f0()
 *
 *************/

void print_table_f0(FILE *fp, Macesym s)
{
  fprintf(fp, "\n %s: %d\n", sn_to_str(s->symnum), function_value_0(Dp, s));
}  /* print_table_f0 */

/*************
 *
 *   print_table_f1()
 *
 *************/

void print_table_f1(FILE *fp, Macesym s)
{
  int i;
  char *s1, *s2;
  int n0 = s->args[0].n;

  if (n0 <= 10) {
    s1 = "%2d";
    s2 = "--";
  }
  else {
    s1 = "%3d";
    s2 = "---";
  }
    
  fprintf(fp, "\n %s :\n", sn_to_str(s->symnum));
  fprintf(fp, "       ");
  for (i = 0; i < n0; i++)
    fprintf(fp, s1, i);
  fprintf(fp, "\n    ---");
  for (i = 0; i < n0; i++)
    fprintf(fp, s2);
  fprintf(fp, "\n       ");
  for (i = 0; i < n0; i++)
    fprintf(fp, s1, function_value_1(Dp, s, i));
  fprintf(fp, "\n");
}  /* print_table_f1 */

/*************
 *
 *   print_table_f2()
 *
 *************/

void print_table_f2(FILE *fp, Macesym s)
{
  int i, j, v;
  char *s1, *s2, *s3;
  int n0 = s->args[0].n;
  int n1 = s->args[1].n;

  if (n1 <= 10) {
    s1 = "%2d";
    s2 = "--";
    s3 = " -";
  }
  else {
    s1 = "%3d";
    s2 = "---";
    s3 = "  -";
  }
    
  fprintf(fp, "\n %s :\n", sn_to_str(s->symnum));
  fprintf(fp, "      |");
  for (i = 0; i < n1; i++)
    fprintf(fp, s1, i);
  fprintf(fp, "\n    --+");
  for (i = 0; i < n1; i++)
    fprintf(fp, s2);
  fprintf(fp, "\n");

  for (i = 0; i < n0; i++) {
    fprintf(fp, "%5d |", i);
    for (j = 0; j < n1; j++) {
      v = function_value_2(Dp, s, i, j);
      if (v < 0)
	fprintf(fp, s3);
      else
	fprintf(fp, s1, v);
    }
    fprintf(fp, "\n");
  }
}  /* print_table_f2 */

/*************
 *
 *   print_table_r0()
 *
 *************/

void print_table_r0(FILE *fp, Macesym s)
{
  int v;
  char *s1;
  v = dp_atom_value(Dp, ATOM0(s));
  switch(v) {
  case ATOM_FALSE: s1 = "F"; break;
  case ATOM_TRUE:  s1 = "T"; break;
  default:         s1 = "?"; break;
  }
  fprintf(fp, "\n %s: %s\n", sn_to_str(s->symnum), s1);
}  /* print_table_r0 */

/*************
 *
 *   print_table_r1()
 *
 *************/

void print_table_r1(FILE *fp, Macesym s)
{
  int i, v;
  char *s1;
  int n0=s->args[0].n;

  fprintf(fp, "\n %s :\n", sn_to_str(s->symnum));
  fprintf(fp, "       ");
  for (i = 0; i < n0; i++)
    fprintf(fp, "%2d", i);
  fprintf(fp, "\n    ---");
  for (i = 0; i < n0; i++)
    fprintf(fp, "--");
  fprintf(fp, "\n       ");
  for (i = 0; i < n0; i++) {
    v = dp_atom_value(Dp, ATOM1(s, i));
    switch(v) {
    case ATOM_FALSE: s1 = "F"; break;
    case ATOM_TRUE:  s1 = "T"; break;
    default:         s1 = "?"; break;
    }
    fprintf(fp, "%2s", s1);
  }
  fprintf(fp, "\n");
}  /* print_table_r1 */

/*************
 *
 *   print_table_r2()
 *
 *************/

void print_table_r2(FILE *fp, Macesym s)
{
  int i, j, v;
  char *s1;
  int n0=s->args[0].n;
  int n1=s->args[1].n;

  fprintf(fp, "\n %s :\n", sn_to_str(s->symnum));
  fprintf(fp, "      |");
  for (i = 0; i < n1; i++)
    fprintf(fp, "%2d", i);
  fprintf(fp, "\n    --+");
  for (i = 0; i < n1; i++)
    fprintf(fp, "--");
  fprintf(fp, "\n");

  for (i = 0; i < n0; i++) {
    fprintf(fp, "%5d |", i);
    for (j = 0; j < n1; j++) {
      v = dp_atom_value(Dp, ATOM2(s, i, j));
      switch(v) {
      case ATOM_FALSE: s1 = "F"; break;
      case ATOM_TRUE:  s1 = "T"; break;
      default:         s1 = "?"; break;
      }
      fprintf(fp, "%2s", s1);
    }
    fprintf(fp, "\n");
  }
}  /* print_table_r2 */

/*************
 *
 *   print_model()
 *
 *   Print the current model as a first-order model.  The DP code
 *   knows about the propositional model, so we call dp_atom_value(Dp, )
 *   to get the information to print the first-order model.
 *
 *   Assume unassigned atoms are true, because dp determines that
 *   we have a model when all remaining clauses are positive.
 *
 *   Don't print equality or order relations.
 *
 *************/

/* DOCUMENTATION
This routine prints the current first-order model in readable format.
<P>
(This routine is almost obsolete.  See interp.h for routines
to print interpretations.)
*/

/* PUBLIC */
void print_model(FILE *fp)
{
  Macesym s;
  static int count;

#if 1
  fprintf(fp, "\n======================= Model #%d at %.2f seconds:\n",
	  ++count, user_time() / 1000.);
#endif

  for (s = Symbols; s != NULL; s = s->next) {

    if (s->function) {
      switch (s->arity) {
      case 1: print_table_f0(fp, s); break;
      case 2: print_table_f1(fp, s); break;
      case 3: print_table_f2(fp, s); break;
      default: fprintf(fp, "Dimension %d table for %s not printed\n",
		       s->arity-1, sn_to_str(s->symnum));
      }
    }
    else if (!s->equality && !s->order) {
      switch (s->arity) {
      case 0: print_table_r0(fp, s); break;
      case 1: print_table_r1(fp, s); break;
      case 2: print_table_r2(fp, s); break;
      default: fprintf(fp, "Dimension %d table for %s not printed\n",
		       s->arity, sn_to_str(s->symnum));
      }
    }
  }
  fprintf(fp, "end_of_model\n");
}  /* print_model */

/*************
 *
 *   f0_term()
 *
 *************/

Term f0_term(DP_state Dp, Macesym m)
{
  char s[10];
  Term t = get_rigid_term("function", 2);
  sprintf(s, "%d", function_value_0(Dp, m));

  t->args[0] = get_rigid_term(sn_to_str(m->symnum), 0);
  t->args[1] = listterm_cons(get_rigid_term(s, 0), get_nil_term());
  return t;
}  /* f0_term */

/*************
 *
 *   f1_term()
 *
 *************/

Term f1_term(DP_state Dp, Macesym m)
{
  int i;
  int n0 = m->args[0].n;

  char s[10];
  Term t = get_rigid_term("function", 2);
  Term t1 = get_rigid_term(sn_to_str(m->symnum), 1);
  Term a = get_nil_term();

  t->args[0] = t1;
  t1->args[0] = get_rigid_term("_", 0);

  for (i = 0; i < n0; i++) {
    sprintf(s, "%d", function_value_1(Dp, m, i));
    a = listterm_append(a, get_rigid_term(s, 0));
  }
  t->args[1] = a;
  return t;
}  /* f1_term */

/*************
 *
 *   f2_term()
 *
 *************/

Term f2_term(DP_state Dp, Macesym m)
{
  int i, j;
  int n0 = m->args[0].n;
  int n1 = m->args[1].n;

  char s[10];
  Term t = get_rigid_term("function", 2);
  Term t1 = get_rigid_term(sn_to_str(m->symnum), 2);
  Term a = get_nil_term();

  t->args[0] = t1;
  t1->args[0] = get_rigid_term("_", 0);
  t1->args[1] = get_rigid_term("_", 0);

  for (i = 0; i < n0; i++) {
    for (j = 0; j < n1; j++) {
      sprintf(s, "%d", function_value_2(Dp, m, i, j));
      a = listterm_append(a, get_rigid_term(s, 0));
    }
  }
  t->args[1] = a;
  return t;
}  /* f2_term */

/*************
 *
 *   f3_term()
 *
 *************/

Term f3_term(DP_state Dp, Macesym m)
{
  int i, j, k;
  int n0 = m->args[0].n;
  int n1 = m->args[1].n;
  int n2 = m->args[2].n;

  char s[10];
  Term t = get_rigid_term("function", 2);
  Term t1 = get_rigid_term(sn_to_str(m->symnum), 3);
  Term a = get_nil_term();

  t->args[0] = t1;
  t1->args[0] = get_rigid_term("_", 0);
  t1->args[1] = get_rigid_term("_", 0);
  t1->args[2] = get_rigid_term("_", 0);

  for (i = 0; i < n0; i++) {
    for (j = 0; j < n1; j++) {
      for (k = 0; k < n2; k++) {
	sprintf(s, "%d", function_value_3(Dp, m, i, j, k));
	a = listterm_append(a, get_rigid_term(s, 0));
      }
    }
  }
  t->args[1] = a;
  return t;
}  /* f3_term */

/*************
 *
 *   r0_term()
 *
 *************/

Term r0_term(DP_state Dp, Macesym m)
{
  char s[10];
  Term t = get_rigid_term("predicate", 2);
  sprintf(s, "%d", dp_atom_value(Dp, ATOM0(m)));

  t->args[0] = get_rigid_term(sn_to_str(m->symnum), 0);
  t->args[1] = listterm_cons(get_rigid_term(s, 0), get_nil_term());
  return t;
}  /* r0_term */

/*************
 *
 *   r1_term()
 *
 *************/

Term r1_term(DP_state Dp, Macesym m)
{
  int i;
  int n0=m->args[0].n;

  char s[10];
  Term t = get_rigid_term("predicate", 2);
  Term t1 = get_rigid_term(sn_to_str(m->symnum), 1);
  Term a = get_nil_term();

  t->args[0] = t1;
  t1->args[0] = get_rigid_term("_", 0);

  for (i = 0; i < n0; i++) {
    sprintf(s, "%d", dp_atom_value(Dp, ATOM1(m, i)));
    a = listterm_append(a, get_rigid_term(s, 0));
  }
  t->args[1] = a;
  return t;
}  /* r1_term */

/*************
 *
 *   r2_term()
 *
 *************/

Term r2_term(DP_state Dp, Macesym m)
{
  int i, j;
  int n0=m->args[0].n;
  int n1=m->args[1].n;

  char s[10];
  Term t = get_rigid_term("predicate", 2);
  Term t1 = get_rigid_term(sn_to_str(m->symnum), 2);
  Term a = get_nil_term();

  t->args[0] = t1;
  t1->args[0] = get_rigid_term("_", 0);
  t1->args[1] = get_rigid_term("_", 0);

  for (i = 0; i < n0; i++) {
    for (j = 0; j < n1; j++) {
      sprintf(s, "%d", dp_atom_value(Dp, ATOM2(m, i, j)));
      a = listterm_append(a, get_rigid_term(s, 0));
    }
  }
  t->args[1] = a;
  return t;
}  /* r2_term */

/*************
 *
 *   r3_term()
 *
 *************/

Term r3_term(DP_state Dp, Macesym m)
{
  int i, j, k;
  int n0=m->args[0].n;
  int n1=m->args[1].n;
  int n2=m->args[2].n;

  char s[10];
  Term t = get_rigid_term("predicate", 2);
  Term t1 = get_rigid_term(sn_to_str(m->symnum), 3);
  Term a = get_nil_term();

  t->args[0] = t1;
  t1->args[0] = get_rigid_term("_", 0);
  t1->args[1] = get_rigid_term("_", 0);
  t1->args[2] = get_rigid_term("_", 0);

  for (i = 0; i < n0; i++) {
    for (j = 0; j < n1; j++) {
      for (k = 0; k < n2; k++) {
	sprintf(s, "%d", dp_atom_value(Dp, ATOM3(m, i, j, k)));
	a = listterm_append(a, get_rigid_term(s, 0));
      }
    }
  }
  t->args[1] = a;
  return t;
}  /* r3_term */

/*************
 *
 *   r4_term()
 *
 *************/

Term r4_term(DP_state Dp, Macesym m)
{
  int i, j, k, l;
  int n0=m->args[0].n;
  int n1=m->args[1].n;
  int n2=m->args[2].n;
  int n3=m->args[3].n;

  char s[10];
  Term t = get_rigid_term("predicate", 2);
  Term t1 = get_rigid_term(sn_to_str(m->symnum), 4);
  Term a = get_nil_term();

  t->args[0] = t1;
  t1->args[0] = get_rigid_term("_", 0);
  t1->args[1] = get_rigid_term("_", 0);
  t1->args[2] = get_rigid_term("_", 0);
  t1->args[3] = get_rigid_term("_", 0);

  for (i = 0; i < n0; i++) {
    for (j = 0; j < n1; j++) {
      for (k = 0; k < n2; k++) {
	for (l = 0; l < n3; l++) {
	  sprintf(s, "%d", dp_atom_value(Dp, ATOM4(m, i, j, k, l)));
	  a = listterm_append(a, get_rigid_term(s, 0));
	}
      }
    }
  }
  t->args[1] = a;
  return t;
}  /* r4_term */

/*************
 *
 *   model_to_term()
 *
 *   Don't do equality relations.
 *
 *************/

static
Term model_to_term(Mace_state ms)
{
  DP_state dp = ms->dp;
  int domain_size;
  Macesym m;
  Term t = get_rigid_term("interpretation", 2);
  Term a;
  char s[10];

  if (Num_sorts > 1)
    fatal_error("smodel_to_term, sorted model");

  domain_size = ms->s[0].n;
  sprintf(s, "%d", domain_size);
  t->args[0] = get_rigid_term(s, 0);
  a = get_nil_term();

  for (m = ms->m; m != NULL; m = m->next) {
    if (!m->equality) {
      if (m->function) {
	switch (m->arity) {
	case 1: a = listterm_append(a, f0_term(dp, m)); break;
	case 2: a = listterm_append(a, f1_term(dp, m)); break;
	case 3: a = listterm_append(a, f2_term(dp, m)); break;
	case 4: a = listterm_append(a, f3_term(dp, m)); break;
	default: fatal_error("model_to_term, arity too big");
	}
      }
      else {
	switch (m->arity) {
	case 0: a = listterm_append(a, r0_term(dp, m)); break;
	case 1: a = listterm_append(a, r1_term(dp, m)); break;
	case 2: a = listterm_append(a, r2_term(dp, m)); break;
	case 3: a = listterm_append(a, r3_term(dp, m)); break;
	case 4: a = listterm_append(a, r4_term(dp, m)); break;
	default: fatal_error("model_to_term, arity too big");
	}
      }
    }
  }
  if (ms->elements_in_input) {
    int i;
    for (i = 0; i < domain_size; i++) {
      Term f = get_rigid_term("function", 2);
      char s[10];
      sprintf(s, "%d", i);
      f->args[0] = get_rigid_term(s, 0);
      f->args[1] = listterm_cons(get_rigid_term(s, 0), get_nil_term());
      a = listterm_append(a, f);
    }
  }
  t->args[1] = a;
  return t;
}  /* model_to_term */

/*************
 *
 *   clean_up_ms()
 *
 *************/

static
void clean_up_ms(Macesym m, Sort sorts)
{
  while (m != NULL) {
    Macesym m2 = m;
    m = m->next;
    free_macesym(m2);
  }
  my_free(sorts);
}  /* clean_up_ms */

/*************
 *
 *   mace_first_model()
 *
 *************/

/* DOCUMENTATION
This routine looks for the first model of a set of unsorted clauses.
The domain size is fixed.  If a model is found, a term representation
of the model is returned.  (NULL is returned if there is no model of
the requested size.)
<UL>
<LI> lst -- a list of clauses
<LI> domain_size -- we look for models with this number of elements
<LI> distinct_constants -- if TRUE, each constant in the input is
     assigned a distinct element of the domain.  If the number of
     constants is greater than the domain size, say n, then the
     first n constants are fixed, and the rest can vary.
     This option can block models, but it is well worth trying,
     because it really prunes the search by eliminating many
     isomorphic models.
<LI> elements_in_input -- if TRUE, then nonnegative integer constants
     in the input (up to domain_size-1) are assigned the domain
     elements of the same name.  Integer constants >= domain_size
     can take on any element of the domain.
<LI> max_seconds -- at most this many time will be spent
     (-1 means that there is no time limit).
<LI> max_megabytes -- at most this much memory will be used
     (-1 means that there is no memory limit).
<LI> flip_equations -- if TRUE, equations (positive unit) are
     used in both orientations.
<LI> subsume -- if TRUE, forward subsumption will be applied in the
     Davis-Putnam procedure.  (FALSE is usually faster.)
<LI> fp -- file to which ordinary processing messages are sent
     (this must not be NULL, but it may be "/dev/null").
<LI> msp -- this is the address of a pointer to a mace_state struct.
     If a model is found, the pointer is set to the state of the
     search.  It is used for subsequent calls to mace_next_model().
     If you get a model and don't call mace_next_model(), you should
     call mace_cancel() to free memory in the mace_state.
     If you don't get a model, don't call mace_cancel().
<LI> rcp -- pointer to the return code variable:
<UL>
<LI> OK_RETURN -- a model is returned.
<LI> NO_MODEL_RETURN -- there are no models (modulo distinct_constants).
<LI> TIME_OUT_RETURN -- max_seconds.
<LI> MEMORY_OUT_RETURN -- max_megabytes.
</UL>
</UL>
*/

/* PUBLIC */
Term mace_first_model(Clist lst,
		      int domain_size,
		      BOOL distinct_constants,
		      BOOL elements_in_input,
		      int max_seconds,
		      int max_megabytes,
		      BOOL flip_equations,
		      BOOL subsume,
		      FILE *fp,
		      Mace_state *msp,
		      int *rcp)
{
  Mace_state ms = get_mace_state();
  ms->elements_in_input = elements_in_input;
  ms->dp = dp_init(max_seconds, max_megabytes, subsume, fp);

  assert_propositions(ms, lst, domain_size,
		      distinct_constants, flip_equations, fp);
  *rcp = dp_first_model(ms->dp);
  if (*rcp == OK_RETURN) {
    Term t;
    t = model_to_term(ms);
    *msp = ms;
    return t;
  }
  else {
    clean_up_ms(ms->m, ms->s);
    free_mace_state(ms);
    return NULL;
  }
}  /* mace_first_model */

/*************
 *
 *   mace_next_model()
 *
 *************/

/* DOCUMENTATION
Get the next model.  Return codes are the same as for mace_first_model().
The max_seconds parameter given to mace_first_model() applies here as
well, and the time spent includes the call to mace_first_model.
The max_megabytes parameter given to mace_first_model() is irrelevant,
because mace_next_model() never allocates additional memory.
*/

/* PUBLIC */
Term mace_next_model(Mace_state ms, int *rcp)
{
  *rcp = dp_next_model(ms->dp);
  if (*rcp == OK_RETURN)
    return model_to_term(ms);
  else {
    clean_up_ms(ms->m, ms->s);
    free_mace_state(ms);
    return NULL;
  }
}  /* mace_next_model */

/*************
 *
 *   mace_cancel()
 *
 *************/

/* DOCUMENTATION
This routine should be called if you get some, but not all, models
from mace_first_model()/mace_next_model().  It frees memory used by
the Mace_state.
*/

/* PUBLIC */
void mace_cancel(Mace_state ms)
{
  dp_cancel(ms->dp);
  clean_up_ms(ms->m, ms->s);
  free_mace_state(ms);
}  /* mace_cancel */
