#include "memory.h"

static int Malloc_bytes=0;      /* btyes allocated by each malloc */
static int Max_megs=0;          /* max number of megabytes to malloc */

static char *Alloc_block=NULL;  /* location returned by most recent malloc */
static char *Alloc_pos;         /* current position in block */
static int Malloc_calls=0;      /* number of calls to malloc */

/*************
 *
 *   my_malloc()
 *
 *************/

/* DOCUMENTATION
This simply calls malloc().  The only difference between this
and malloc() is that malloc takes a size_t argument, and this
takes an int.
*/

/* PUBLIC */
void *my_malloc(int number_of_bytes)
{
  void *p = malloc((size_t) number_of_bytes);
  /* printf("%p my_malloc (%d).\n", p, number_of_bytes); fflush(stdout); */
  return p;
}  /* my_malloc */

/*************
 *
 *   my_free()
 *
 *************/

/* DOCUMENTATION
This simply calls free().
*/

/* PUBLIC */
void my_free(void *p)
{
  /* printf("%p my_free.\n", p); fflush(stdout); */
  free(p);
}  /* my_free */

/*************
 *
 *    void *tp_alloc(n)
 *
 *    Allocate n contiguous bytes, aligned on pointer boundry.
 *
 *************/

/* DOCUMENTATION
This routine allocates and returns a block of at least n bytes memory,
aligned so that it can be used for any kind of data.  (The size of the block
returned is a multiple of the size of a pointer.)  The way it works is
that a large block is obtained from malloc(), and this routine gives
out pieces until it has to get more from malloc(), and so on.
<P>
If the operating system runs out of memory, or the amount requested
is too great, or Max_megs is exceeded, the whole process just dies.
*/

/* PUBLIC */
void *tp_alloc(size_t n)
{
  char *return_block;
  int psize;

  /* Initialize some parameters if necessary. */

  if (Malloc_bytes == 0)
    Malloc_bytes = DEFAULT_MALLOC_BYTES;
  if (Max_megs == 0)
    Max_megs = DEFAULT_MAX_MEGS;
  
  /* If n is not a multiple of sizeof(void *), round up so that it is */
  
  psize = sizeof(void *);
  if (n % psize != 0)
    n = n + (psize - (n % psize));

  /* If first call or not enough in the current block, get a new block. */
  
  if (Alloc_block==NULL || Alloc_block + Malloc_bytes - Alloc_pos < n) {
    if (n > Malloc_bytes)
      fatal_error("tp_alloc, request too big");
    else if ((Malloc_calls+1)*Malloc_bytes > Max_megs*1024*1024)
      fatal_error("tp_alloc, Max_megs parameter exceeded");
    else {
      Alloc_pos = Alloc_block = malloc((size_t) Malloc_bytes);
      Malloc_calls++;
      if (!Alloc_pos)
	fatal_error("tp_alloc, operating system is out of memory");
    }
  }
  return_block = Alloc_pos;
  Alloc_pos += n;
  return(return_block);
}  /* tp_alloc */

/*************
 *
 *    void *tp_calloc(n)
 *
 *************/

/* DOCUMENTATION
This is just like tp_alloc(), except the memory is initialized to 0.
*/

/* PUBLIC */
void *tp_calloc(size_t n)
{
  int i;
  char *p = tp_alloc(n);

  for (i = 0; i < n; i++)
    p[i] = 0;
  return p;
}  /* tp_calloc */

/*************
 *
 *    int kbytes_malloced() -- How many K have been dynamically allocated?
 *
 *************/

/* DOCUMENTATION
This routine returns the number of kilobytes that tp_alloc()
has obtained from the operating system by my_malloc();
*/

/* PUBLIC */
int kbytes_malloced(void)
{
  return( (int) (Malloc_calls * (Malloc_bytes / 1024.)));
}  /* kbytes_malloced */

/*************
 *
 *   set_max_megs()
 *
 *************/

/* DOCUMENTATION
This routine changes the limit on the amount of memory obtained
from my_malloc() by tp_alloc().  The argument is in megabytes.
The default value is DEFAULT_MAX_MEGS.
*/

/* PUBLIC */
void set_max_megs(int megs)
{
  Max_megs = megs;
}  /* set_max_megs */

/*************
 *
 *   set_malloc_bytes()
 *
 *************/

/* DOCUMENTATION
This routine changes the size of the blocks obtained from my_malloc() by
tp_alloc().  The argument is in bytes.  The default value is
DEFAULT_MALLOC_BYTES.
*/

/* PUBLIC */
void set_malloc_bytes(int megs)
{
  Malloc_bytes = megs;
}  /* set_malloc_bytes */
