#include "interp.h"
#include "mace.h"
#include "clausify.h"

#include <unistd.h>  /* for gethostname() and getopt() */

#define VERSION "2.0"
#define VDATE "March 1998"

/* Global parameters from the command-line options. */

static int  Domain_size           = 0;
static BOOL First_order           = FALSE;
static BOOL Quasigroup_column     = FALSE;
static BOOL Sorted_input          = FALSE;

static BOOL Distinct_constants    = FALSE;
static BOOL Domain_elements       = FALSE;
static BOOL Subsume               = FALSE;
static BOOL Flip_equations        = FALSE;
static BOOL Print_models          = FALSE;
static BOOL Print_models_portable = FALSE;
static BOOL Quantified_input      = FALSE;

static int Max_models             = -1;
static int Max_seconds            = -1;
static int Max_megabytes          = -1;

/*************
 *
 *    void print_banner(argc, argv)
 *
 *************/

static void print_banner(int argc, char **argv)
{
  int i;
  char host[64];

  if (gethostname(host, 64) != 0)
    strcpy(host, "???");

  printf("----- MACE %s, %s -----\n", VERSION, VDATE);
  printf("The job began on %s, %s", host, get_date());
  printf("The command was \"");
  for(i = 0; i < argc; i++)
    printf("%s%s", argv[i], (i < argc-1 ? " " : ""));
  printf("\".\n\n");
}  /* print_banner */

/*************
 *
 *   bell()
 *
 *************/

static
void bell(FILE *fp)
{
  fprintf(fp, "%c", '\007');
}  /* bell */

/*************
 *
 *   usage_message()
 *
 *************/

void usage_message()
{
  printf("Input formulas  are taken from standard input.\n");
  printf("Input should be first-order iff -n option is used.\n");
  printf("First-order should be quantified iff -f option is used.\n");
  printf("\nCommand-line options for either propositional or first-order:\n");
  printf("   -p    | Print models as they are found.\n");
  printf("   -m n  | Stop when the n-th model is found.\n");
  printf("   -M n  | Use at most n megabutes.\n");
  printf("   -t n  | Stop after n seconds.\n");
  printf("   -s    | Perform unit subsumption. (Always done on input.)\n");
  printf("   -h    | Print this message.\n");
  printf("\nOptions for first-order problems only:\n");
  printf("   -n n  | Read first-order clauses and use domain size n.\n");
  printf("   -f    | Input are quantified formulas rather than clauses.\n");
  printf("   -P    | Print models in portable format as they are found.\n");
  printf("   -e    | Naturals in clauses are domain elements (up to n).\n");
  printf("   -u    | Assume constants are distinct (up to n).\n");
  printf("   -b    | Equations both ways.\n");
  printf("   -x    | Quasigroup constraint on last column of f/2.\n");
  printf("   -z    | Sorted first_order input.\n");

  printf("\n");
}  /* usage_message */

/*************
 *
 *   process_command_line_args()
 *
 *   return !error;
 *
 *************/

static
BOOL process_command_line_args(int argc, char **argv)
{
  extern char *optarg;

  int c, n;
  BOOL error = FALSE;

  while ((c = getopt(argc, argv, "fbspPhzuexm:t:M:n:d:")) != EOF) {
    switch (c) {
    case 'f':
      Quantified_input = TRUE;
      break;
    case 'u':
      Distinct_constants = TRUE;
      break;
    case 'e':
      Domain_elements = TRUE;
      break;
    case 's':
      Subsume = TRUE;
      break;
    case 'b':
      Flip_equations = TRUE;
      break;
    case 'p':
      Print_models = TRUE;
      break;
    case 'P':
      Print_models_portable = TRUE;
      break;
    case 'z':
      Sorted_input = TRUE;
      First_order = TRUE;
      break;
    case 'x':
      Quasigroup_column = TRUE;
      break;
    case 'm':
      n = atoi(optarg);
      Max_models = n;
      break;
    case 't':
      n = atoi(optarg);
      Max_seconds = n;
      break;
    case 'M':
      n = atoi(optarg);
      Max_megabytes = n;
      break;
    case 'n':
      n = atoi(optarg);
      Domain_size = n;
      First_order = TRUE;
      break;
    case 'h':
    case '?':
    default:
      error = TRUE;
      break;
    }
  }

  /* Check for inconsistent options. */

  if (First_order && !Sorted_input && Domain_size < 1) {
    fprintf(stderr, "\nDomain size must be >= 1.\n");
    error = TRUE;
  }

  if (Sorted_input && Distinct_constants) {
    fprintf(stderr, "\nSorted input and distinct constants are incompatible.\n");
    error = TRUE;
  }
	
  return (error == FALSE);
}  /* process_command_line_args */

/*************
 *
 *   read_all_clauses(fp)
 *
 *   This is used for propositional problems only.
 *   All errors are fatal.
 *
 *************/

#define MAX_PROP_LITS 1000

static
void read_all_clauses(FILE *fp, DP_state dp)
{
  int c[MAX_PROP_LITS], num_lits, lit, rc;

  do {
    num_lits = 0;
    do {
      rc = fscanf(fp, "%d", &lit);
      if (rc == EOF && num_lits != 0)
	fatal_error("read_all_clauses, last clause not terminated with 0");
      else if (rc == 0)
	fatal_error("propositional input must be all integers.");
      else if (lit != 0) {
	if (num_lits == MAX_PROP_LITS)
	  fatal_error("propositional clause has too many literals.");
	c[num_lits++] = lit;
      }
    } while (lit != 0 && rc != EOF);

    if (rc != EOF)
      rc = dp_append_clause(dp, c, num_lits);
  } while (rc != EOF);
}  /* read_all_clauses */

/*************
 *
 *   propositional()
 *
 *************/

static
void propositional(void)
{
  int models = 0;
  int next_message = 1;
  int rc;
  DP_state dp = dp_init(-1, 4, FALSE, stdout);
  
  read_all_clauses(stdin, dp);

  if (Max_models == 0)
    rc = NO_MODEL_RETURN;
  else
    rc = dp_first_model(dp);

  while (rc == OK_RETURN && (Max_models == -1 || models < Max_models)) {
    models++;
    if (Print_models || Print_models_portable) {
      int n, *lits, i;
      dp_get_model(dp, &lits, &n);
      printf("\nModel:");
      for (i = 0; i < n; i++)
	printf(" %d", lits[i]);
      printf("\n");
    }
    else if (next_message == models) {
      printf("\nThe %d%s model has been found.\n", models,
	     next_message == 1 ? "st" : "th");
      fflush(stdout);
      next_message *= 10;
    }
    if (Max_models == -1 || models < Max_models)
      rc = dp_next_model(dp);
  }

  if (rc == NO_MODEL_RETURN) {
    if (models == 0)
      printf("\nNo models were found.\n");
    else
      printf("\nAll models (%d) were found.\n", models);
  }
  else if (rc == TIME_OUT_RETURN)
    printf("\nTime out.\n");
  else if (rc == MEMORY_OUT_RETURN)
    printf("\nMemory out.\n");
  else if (rc == OK_RETURN) {
    dp_cancel(dp);
    printf("\nThe requested number of models (%d) was found.\n", models);
  }
  exit(0);
}  /* propositional */

/*************
 *
 *   check_terms()
 *
 *************/

static
void check_terms(Term t)
{
  if (CONSTANT(t)) {
    int e;
    BOOL rc = str_to_int(sn_to_str(SYMNUM(t)), &e);
    if (rc && e >= 0 && !Domain_elements) {
      bell(stderr);
      fprintf(stderr, "\nWARNING: the following constant looks suspicious: ");
      fwrite_term_nl(stderr, t);
      fprintf(stderr, "Did you forget the -e option?\n");
    }
    else if (rc && e >= 0 && Domain_elements && e >= Domain_size) {
      bell(stderr);
      fprintf(stderr, "\nWARNING: the following constant is greater than the "
	      "domain size: ");
      fwrite_term_nl(stderr, t);
      fprintf(stderr, "It will be treated as an ordinary constant.\n");
    }
  }
  else if (COMPLEX(t)) {
    if (cons_term(t)) {
      bell(stderr);
      fprintf(stderr, "\nWARNING: the following term is suspicious.\n");
      fwrite_term_nl(stderr, t);
      fprintf(stderr, "Do really wish to look for a model involving lists?\n");
    }
    else {
      int i;
      for (i = 0; i < t->arity; i++)
	check_terms(t->args[i]);
    }
  }
}  /* check_terms */

/*************
 *
 *   check_for_suspicious_things()
 *
 *************/

static
void check_for_suspicious_things(Clist clauses)
{
  Clist_pos p;
  for (p = clauses->first; p != NULL; p = p->next) {
    Literal lit;
    for (lit = p->c->first_literal; lit != NULL; lit = lit->next) {
      Term a = lit->atom;
      if (!Quantified_input) {
	if (is_symbol(SYMNUM(a), AND_SYM, 2) ||
	    is_symbol(SYMNUM(a), IFF_SYM, 2) ||
	    is_symbol(SYMNUM(a), IMP_SYM, 2) ||
	    is_symbol(SYMNUM(a), IMPBY_SYM, 2) ||
	    is_symbol(SYMNUM(a), QUANT_SYM, 1)) {
	  bell(stderr);
	  fprintf(stderr, "\nWARNING: the following clause looks more like a "
		  "general formula than a clause.\n");
	  fwrite_clause(stderr, p->c, FALSE);
	  fprintf(stderr, "Did you forget the -f option?\n");
	}
      }
      check_terms(a);
    }  /* for each literal */
  }
}  /* check_for_suspicious_things */

/*************
 *
 *    main
 *
 *************/

int main(int argc, char **argv)
{
  int rc, models;
  int next_message = 1;
  Mace_state ms;
  Clist clauses;
  Glist formulas = NULL;
  Term t;

  print_banner(argc, argv);

  rc = process_command_line_args(argc, argv);
  if (!rc) {
    usage_message();
    exit(0);
  }

  if (Domain_size < 1)
    propositional();  /* this exits process */

  declare_standard_parse_types();
  set_parse_type("v", 450, INFIX);
  set_parse_type("^", 450, INFIX);

  if (Quantified_input) {
    formulas = read_formula_list(stdin, stdout);
    fwrite_formula_list(stdout, formulas, "formulas_for_mace");
    clauses = clausify_list(formulas);
    name_clist(clauses, "clausified_formulas_for_mace");
    fwrite_clause_list(stdout, clauses, FALSE);
  }
  else {
    clauses = read_clause_list(stdin, stdout, "clauses_for_mace", FALSE);
    fwrite_clause_list(stdout, clauses, FALSE);
  }
  check_for_suspicious_things(clauses);

  if (Max_models == 0) {
    t = NULL;
    rc = NO_MODEL_RETURN;
  }
  else
    t = mace_first_model(clauses, Domain_size,
			 Distinct_constants, Domain_elements,
			 Max_seconds, Max_megabytes,
			 Flip_equations, Subsume,
			 stdout,
			 &ms, &rc);
  models = 0;
  while (t != NULL && (Max_models == -1 || models < Max_models)) {
    models++;
    if (Print_models || Print_models_portable) {
      printf("\n======================= Model #%d at %.2f seconds:\n",
            models, user_time() / 1000.);
      if (Print_models)
	fprint_interp_tabular(stdout, t);
      if (Print_models_portable)
	fprint_interp_from_term(stdout, t);
    }
    else if (next_message == models) {
      printf("\nThe %d%s model has been found.\n", models,
	     next_message == 1 ? "st" : "th");
      fflush(stdout);
      next_message *= 10;
    }
    if (1) {
      Interp i = compile_interp(t);
      if (Quantified_input) {
	Glist p;
	for (p = formulas; p != NULL; p = p->next) {
	  Formula f = p->u.v;
	  if (!eval_formula(f, i)) {
	    printf("FALSE: ");
	    fwrite_formula(stdout, f);
	    fatal_error("Formula false in model!!");
	  }
	}
      }
      else {
	Clist_pos p;
	for (p = clauses->first; p != NULL; p = p->next) {
	  Clause c = p->c;
	  if (!eval_clause(c, i)) {
	    printf("FALSE: ");
	    fwrite_clause(stdout, c, FALSE);
	    fatal_error("Clause false in model!!");
	  }
	}
      }
      zap_interp(i);
    }
    zap_term(t);
    if (Max_models == -1 || models < Max_models)
      t = mace_next_model(ms, &rc);
  }

  if (rc == NO_MODEL_RETURN) {
    if (models == 0)
      printf("\nNo models were found.\n");
    else
      printf("\nAll models were found.\n");
  }
  else if (rc == TIME_OUT_RETURN)
    printf("\nTime out.\n");
  else if (rc == MEMORY_OUT_RETURN)
    printf("\nMemory out.\n");
  else if (rc == OK_RETURN) {
    mace_cancel(ms);
    printf("\nThe requested number of models was found.\n");
  }

  clist_zap(clauses);
  if (Quantified_input)
    zap_formula_list(formulas);

#if 1
  fprint_term_mem(stdout, 1);
  fprint_glist_mem(stdout, 0);
  fprint_clause_mem(stdout, 0);
  fprint_clist_mem(stdout, 0);
  fprint_formula_mem(stdout, 0);
  fprint_putnam_mem(stdout, 0);
  fprint_interp_mem(stdout, 0);
  fprint_mace_mem(stdout, 0);
#endif
  exit(0);
}  /* main */

