#include "paramod.h"

/* Private definitions and types */

int Xflag;  /* termflag to mark AC-extended equality atoms */

/*************
 *
 *   paramodulate()
 *
 *   Construct a paramodulant.
 *
 *************/

static
Clause paramodulate(Literal from_lit,
		    Term alpha, Context from_subst,
		    Term into,  Context into_subst)
{
  Clause from_clause = alpha->container;
  Clause into_clause = into->container;
  Clause p = get_clause();
  Term beta = NULL;
  Literal lit;
  for (lit = from_clause->first_literal; lit != NULL; lit = lit->next) {
    if (lit != from_lit)
      append_literal(p, apply_lit(lit, from_subst));
    else
      beta = (lit->atom->args[0]==alpha?lit->atom->args[1]:lit->atom->args[0]);
  }
  for (lit = into_clause->first_literal; lit != NULL; lit = lit->next) {
    append_literal(p, new_literal(lit->sign,
				  apply_substitute(lit->atom,
						   beta, from_subst,
						   into, into_subst)));
  }
  upward_clause_links(p);
  return p;
}  /* paramodulate */

/*************
 *
 *   para_rule()
 *
 *************/

static
int para_rule(Term from_atom, Term into_atom)
{
  BOOL fx = term_flag(from_atom, Xflag);
  BOOL ix = term_flag(into_atom, Xflag);
  if (!fx && !ix)
    return PARA_JUST;
  else if (fx && !ix)
    return PARA_FX_JUST;
  else if (!fx && ix)
    return PARA_IX_JUST;
  else
    return PARA_FX_IX_JUST;
}  /* para_rule */

/*************
 *
 *   para_into()
 *
 *   Paramodulate from a given side of a given literal into a given
 *   term and/or its subterms.  The position vectors are for the
 *   justification only.  The parameter "where" should be
 *   PARA_ALL, PARA_TOP_ONLY, or PARA_ALL_EXCEPT_TOP.
 *
 *************/

static
void para_into(Literal from_lit, Term alpha, Glist from_vec,
	       Literal into_lit, Term into,
	       Glist into_vec, Para_loc where,
	       Glist *pp)
{
  if (!VARIABLE(into)) {
    int i;
    if (COMPLEX(into) && (where == PARA_ALL || where == PARA_ALL_EXCEPT_TOP)) {
      Glist g = ilist_prepend(into_vec, 0);
      if (is_assoc_comm(SYMNUM(into))) {
	/* Assume into is AC canonical (in particular, right associated).
	 * Recurse on AC args only.  That is, loop as if the term were flat.
	 * Also, position vectors are as if AC terms were flat.
	 */
	Term t = into;
	Term arg = t->args[0];
	while (arg != NULL) {
	  g->u.i += 1;
	  para_into(from_lit,alpha,from_vec,into_lit,arg,g,PARA_ALL,pp);
	  if (SYMNUM(t) == SYMNUM(into)) {
	    t = t->args[1];
	    arg = SYMNUM(t) == SYMNUM(into) ? t->args[0] : t;
	  }
	  else
	    arg = NULL;
	}
      }  /* AC into */
      else {
	for (i = 0; i < into->arity; i++) {
	  g->u.i += 1;
	  para_into(from_lit, alpha, from_vec,
		    into_lit, into->args[i], g, PARA_ALL, pp);
	}
      }  /* non-AC into */
      free_glist(g);  /* Frees first node only; into_vec is unchanged. */
    }
    if (where == PARA_ALL || where == PARA_TOP_ONLY) {
      Context cf = get_context();
      Context ci = get_context();
      Btu_state bt = unify_bt_first(alpha, cf, into, ci);
      while (bt != NULL) {
	Clause p = paramodulate(from_lit, alpha, cf, into, ci);
	int rule = para_rule(from_lit->atom, into_lit->atom);
	p->justification = para_just(rule, alpha->container, from_vec,
				     into->container, into_vec);
	*pp = plist_prepend(*pp, p);
	bt = unify_bt_next(bt);
      }
      free_context(cf);
      free_context(ci);
    }
  }
}  /* para_into */

/*************
 *
 *   para_lits_into_noneq()
 *
 *************/

static
void para_lits_into_noneq(Literal flit, Term alpha,
			  Literal ilit, BOOL extend_from,
			  Glist *pp)
{
  Term orig_from_atom = NULL;
  BOOL from_extended = FALSE;
  if (extend_from) {
    Term x = extend(flit->atom, alpha, Xflag);
    if (x != NULL) {
      orig_from_atom = flit->atom;
      flit->atom = x;
      from_extended = TRUE;
    }
  }
  if (!extend_from || from_extended) {
    /* Position vectors are constructed backward. */
    Glist from_vec = ilist_prepend(ilist_prepend(NULL,0),0);
    Glist into_vec = ilist_prepend(ilist_prepend(NULL,0),0);
    Term a = ilit->atom;
    int i;
    from_vec->next->u.i = literal_number(alpha->container, flit);
    from_vec->u.i = flit->atom->args[0] == alpha ? 1 : 2;  /* arg of flit */
    into_vec->next->u.i = literal_number(ilit->atom->container, ilit);
    /* The atom is not equality, so do nothing special. */
    for (i = 0; i < a->arity; i++) {
      into_vec->u.i += 1;  /* increment arg number */
      para_into(flit,alpha,from_vec,ilit,a->args[i],into_vec,PARA_ALL,pp);
    }
    zap_glist(from_vec);
    zap_glist(into_vec);
  }
  /* If from atom was extended, restore it to its original state. */
  if (from_extended) {
    zap_term(flit->atom);
    flit->atom = orig_from_atom;
  }
}  /* para_lits_into_noneq */

/*************
 *
 *   para_lits_into_eq()
 *
 *************/

static
void para_lits_into_eq(Literal flit, Term alpha,
		       Literal ilit, Term into_arg,
		       Para_loc where,
		       BOOL extend_from, BOOL extend_into,
		       Glist *pp)
{
  Term orig_from_atom = NULL;
  BOOL from_extended = FALSE;
  BOOL from_left = flit->atom->args[0] == alpha;

  Term orig_into_atom = NULL;
  BOOL into_extended = FALSE;
  BOOL into_left = ilit->atom->args[0] == into_arg;

  /* I'm not sure if self-paramodulation with extensions is necessary.
   * In either case, this code will not handle it, because literals
   * are extended in place.  If added later, clauses will probably have
   * to be copied (cloned) when extended.  
   */

  if (flit == ilit && (extend_from || extend_into))
    return;

  if (extend_from) {
    Term x = extend(flit->atom, alpha, Xflag);
    if (x != NULL) {
      orig_from_atom = flit->atom;
      flit->atom = x;
      alpha = (from_left ? x->args[0] : x->args[1]);
      from_extended = TRUE;
    }
  }

  if (extend_into && ilit->sign) {
    Term x = extend(ilit->atom, into_arg, Xflag);
    if (x != NULL) {
      orig_into_atom = ilit->atom;
      ilit->atom = x;
      into_arg = (into_left ? x->args[0] : x->args[1]);
      into_extended = TRUE;
    }
  }

  if ((!extend_from || from_extended) && (!extend_into || into_extended)) {
    /* Position vectors are constructed backward. */
    Glist from_vec = ilist_prepend(ilist_prepend(NULL,0),0);
    Glist into_vec = ilist_prepend(ilist_prepend(NULL,0),0);
   
    from_vec->next->u.i = literal_number(alpha->container, flit);
    from_vec->u.i = (from_left ? 1 : 2);  /* arg of flit */
    into_vec->next->u.i = literal_number(into_arg->container, ilit);
    into_vec->u.i = (into_left ? 1 : 2);  /* arg of ilit */

    para_into(flit,alpha,from_vec,ilit,into_arg,into_vec,where, pp);

    zap_glist(from_vec);
    zap_glist(into_vec);
  }

  /* If atoms were extended, restore them to their original states. */
  if (from_extended) {
    zap_term(flit->atom);
    flit->atom = orig_from_atom;
  }
  if (into_extended) {
    zap_term(ilit->atom);
    ilit->atom = orig_into_atom;
  }
}  /* para_lits_into_eq */

/*************
 *
 *   para_from_into()
 *
 *************/

/* DOCUMENTATION
Paramodulate from one clause into another.
The "where" parameter tells where to paramodulate into
equality literals:  PARA_ALL, PARA_TOP_ONLY, or PARA_ALL_EXCEPT_TOP.
"TOP" means the left or right argument of the equality.
<P>
The "extend_from" parameter means that all <I>from</I> literals must
be AC extensions, and
"extend_into" means that all <I>into</I> literals must
be AC extensions of positive equality literals.
<P>
For oriented equality atoms, we go from and into left sides only.
For nonoriented equality atoms, we go from and into both sides.
<P>
Without AC, the normal sequence for paramodulating C and D is
<UL>
<LI> C -> D, PARA_ALL;
<LI> D -> C, PARA_ALL_EXCEPT_TOP;
</UL>
With AC, the normal sequence for paramodulating C and D is (x means extension)
<UL>
<LI> C  -> D, PARA_ALL;
<LI> Cx -> D, PARA_ALL;
<LI> C  -> Dx, PARA_TOP_ONLY;
<LI> Cx -> Dx, PARA_TOP_ONLY;
<LI> D  -> C, PARA_ALL_EXCEPT_TOP;
<LI> Dx -> C, PARA_ALL_EXCEPT_TOP;
<LI> (into Cx is covered by the other cases).
</UL>
*/

/* PUBLIC */
Glist para_from_into(Clause from, Clause into, Para_loc where,
		    BOOL extend_from, BOOL extend_into)
{
  Literal flit;
  Glist paramodulants = NULL;
  Xflag = claim_term_flag();
  for (flit = from->first_literal; flit != NULL; flit = flit->next) {
    if (flit->sign && eq_atom(flit->atom)) {
      Term f1 = flit->atom->args[0];
      Term f2 = flit->atom->args[1];
      Literal ilit;
      for (ilit = into->first_literal; ilit != NULL; ilit = ilit->next) {
	if (eq_atom(ilit->atom)) {
	  Term i1 = ilit->atom->args[0];
	  Term i2 = ilit->atom->args[1];

	  para_lits_into_eq(flit,f1,ilit,i1,where,extend_from,extend_into,&paramodulants);
	  if (!oriented_eq(flit->atom))
	    para_lits_into_eq(flit,f2,ilit,i1,where,extend_from,extend_into,&paramodulants);
	  if (!oriented_eq(ilit->atom))
	    para_lits_into_eq(flit,f1,ilit,i2,where,extend_from,extend_into,&paramodulants);
	  if (!oriented_eq(flit->atom) && !oriented_eq(ilit->atom))
	    para_lits_into_eq(flit,f2,ilit,i2,where,extend_from,extend_into,&paramodulants);
	}
	else {
	  /* Into atom is non-equality.
	   * If we're supposed to extend_into, don't do anything.
	   */
	  if (!extend_into) {
	    para_lits_into_noneq(flit,f1,ilit,extend_from,&paramodulants);
	    if (!oriented_eq(flit->atom))
	      para_lits_into_noneq(flit,f2,ilit,extend_from,&paramodulants);
	  }
	}
      }
    }
  }
  release_term_flag(Xflag);
  return paramodulants;
}  /* para_from_into */

