#include "parautil.h"

/* Private definitions and types */

int Oriented_flag = -1;  /* termflag to mark oriented equality atoms */

/*************
 *
 *   eq_atom()
 *
 *************/

/* DOCUMENTATION
This function checks if an atom is an equality atom (positive or negative)
for the purposes of paramodulation and demodulation.
*/

/* PUBLIC */
BOOL eq_atom(Term a)
{
  return is_symbol(SYMNUM(a), "=", 2);
}  /* eq_atom */

/*************
 *
 *   pos_eq_unit()
 *
 *************/

/* DOCUMENTATION
This function checks if a clause is a positive equality unit
for the purposes of paramodulation and demodulation.
*/

/* PUBLIC */
BOOL pos_eq_unit(Clause c)
{
  return (number_of_literals(c) == 1 &&
	  c->first_literal->sign &&
	  eq_atom(c->first_literal->atom));
}  /* pos_eq_unit */

/*************
 *
 *   mark_oriented_eq()
 *
 *************/

/* DOCUMENTATION
This routine marks an atom as an oriented equality.
*/

/* PUBLIC */
void mark_oriented_eq(Term atom)
{
  if (Oriented_flag == -1)
    Oriented_flag = claim_term_flag();
  term_flag_set(atom, Oriented_flag);
}  /* mark_oriented_eq */

/*************
 *
 *   oriented_eq()
 *
 *************/

/* DOCUMENTATION
This function checks if an atom is an oriented equality atom.
(The terms are not actually compared.  Only the mark is checked.)
*/

/* PUBLIC */
BOOL oriented_eq(Term atom)
{
  if (Oriented_flag == -1)
    Oriented_flag = claim_term_flag();
  return term_flag(atom, Oriented_flag);
}  /* oriented_eq */

/*************
 *
 *   extendable_term()
 *
 *************/

static
BOOL extendable_term(Term top)
{
  if (!is_assoc_comm(SYMNUM(top)))
    return FALSE;
  else {
    BOOL ok = TRUE;
    Term t = top;
    Term arg = t->args[0];
    while (arg != NULL && ok) {
      if (VARIABLE(arg) && occurrences(top, arg) == 1)
	ok = FALSE;
      else if (SYMNUM(t) == SYMNUM(top)) {
	t = t->args[1];
	arg = SYMNUM(t) == SYMNUM(top) ? t->args[0] : t;
      }
      else
	arg = NULL;
    }
    return ok;
  }
}  /* extendable_term */

/*************
 *
 *   extend()
 *
 *************/

/* DOCUMENTATION
Atom is an equality, and arg is one of its two arguments.
If the arg side of the atom is AC-extendable, return extended atom,
else return NULL.
<P>
(A term is AC-extendable if the root symbol is AC, and it does not
have a lone variable as an AC argument.)
<P>
The extended atom is an entirely new copy.
*/

/* PUBLIC */
Term extend(Term atom, Term arg, int flag)
{
  if (extendable_term(arg)) {
    Clause c = arg->container;
    int varnum = greatest_variable_in_clause(c) + 1;
    Term newatom = build_binary_term(SYMNUM(atom),
				  build_binary_term(SYMNUM(arg),
						    get_variable_term(varnum),
						    copy_term(atom->args[0])),
				  build_binary_term(SYMNUM(arg),
						    get_variable_term(varnum),
						    copy_term(atom->args[1])));
    upward_term_links(newatom, c);
    if (oriented_eq(atom))
      mark_oriented_eq(newatom);
    term_flag_set(newatom, flag);
    return newatom;
  }
  else
    return NULL;
}  /* extend */

/*************
 *
 *   term_compare_orient()
 *
 *************/

static
Ordertype term_compare_orient(Term alpha, Term beta)
{
  if (lrpo(alpha, beta))
    return GREATER_THAN;
  else if (lrpo(beta, alpha))
    return LESS_THAN;
  else
    return NOT_COMPARABLE;
}  /* term_compare_orient */

/*************
 *
 *   orient_equalities()
 *
 *************/

/* DOCUMENTATION
For each equality literal (positive or negative) of Clause c,
compare the arguments; if the left is greater, mark the
atom as oriented, and if the the right is greater, flip the
arguments (add a note about flipping to the justification),
and mark the atom as oriented.
<P>
(<I>What ordering method will be used???</I>
*/

/* PUBLIC */
void orient_equalities(Clause c)
{
  Literal lit;
  int i;
  for (lit = c->first_literal, i = 1; lit != NULL; lit = lit->next, i++) {
    if (eq_atom(lit->atom)) {
      BOOL oriented = FALSE;
      BOOL flipped = FALSE;
      Term alpha = lit->atom->args[0];
      Term beta  = lit->atom->args[1];
      Ordertype rc = term_compare_orient(alpha, beta);
      if (rc == GREATER_THAN)
	oriented = TRUE;
      else if (rc == LESS_THAN) {
	oriented = TRUE;
	flipped = TRUE;
	lit->atom->args[0] = beta;
	lit->atom->args[1] = alpha;
	c->justification = ilist_append(ilist_append(c->justification,
						     FLIP_JUST), i);
      }
      if (oriented)
	mark_oriented_eq(lit->atom);
    }
  }
}  /* orient_equalities */

