#include "putnam.h"

/* Private definitions and types */

/* #define TRACE */  /* This turns on some debugging output. */

static int Utility_clock = -1;  /* for timing various perations */

/***************************************************************************/

/* OH NO! setjmp(),longjmp()!!  It makes me feel so unclean.
 * Please forgive me.  It is used in case we run out of memory.
 *
 * If you don't like it, you can disable it by not defining TP_SETJMP---in
 * that case, if memory trouble happens, a fatal error will occur.
 */

#define TP_SETJMP

#ifdef TP_SETJMP
#include <setjmp.h>
static jmp_buf Env;
#endif

/* None of these JMP symbols can be zero! */

#define JMP_TOO_BIG   1   /* memory request too great */
#define JMP_TOO_MUCH  2   /* total memory usage for DP exceeds limit */
#define JMP_OS_OUT    3   /* operating system cannot supply any more memory */

/***************************************************************************/

/* Return codes for some static routines.  Note that there is another set
 * of return codes for public routines and some other statics (see putnam.h).
 */

#define UNSATISFIABLE       0
#define NOT_DETERMINED      1

/* size of chunks malloced by dp_alloc() */

#define MALLOC_BYTES  1024*1024*4

/***************************************************************************/

/* Memory conservation is important, so the following integer types are
 * defined to be just as big as required.  All can safely be unsigned.
 *
 * ATOM_INT: To store integers representing atoms.  (The greatest number
 *      of atoms I have seen is 37026 in qg7.33).
 * LIT_INT:  To store the number of pos or neg literals in a clause.
 *      For large problems, this is usually limited by the domain size.
 */

#if 1  /* This is faster and uses less memory. */
#  define ATOM_INT          unsigned short
#  define ATOM_INT_MAX      USHRT_MAX
#  define LIT_INT           unsigned char
#  define LIT_INT_MAX       UCHAR_MAX
#else
#  define ATOM_INT          int
#  define ATOM_INT_MAX      INT_MAX
#  define LIT_INT           int
#  define LIT_INT_MAX       INT_MAX
#endif

typedef struct dpclause * DPclause;
typedef struct dpatom   * DPatom;
typedef struct dpnode   * DPnode;

/* dpclause -- These are kept in a list. */

struct dpclause {
  LIT_INT num_pos;      /* number of positive literals */
  LIT_INT num_neg;      /* number of negative literals */
  LIT_INT active_pos;   /* number of active (unassigned) positive literals */
  LIT_INT active_neg;   /* number of active (unassigned) negative literals */
  /* The following two are dynamically allocated arrays of ATOM_INT. */
  ATOM_INT *pos;        /* array of positive literals */
  ATOM_INT *neg;        /* array of negative literals (stored positively) */
  ATOM_INT subsumer;    /* 0 if not subsumed; else index of subsuming atom */
  DPclause next;        /* next clause in list */
};

/* dpatom -- These are kept in an array indexed by the atom (a positive
 * integer).  Given an atom, we get immediately to a node of this type,
 * then we can quickly access the clauses that contain the atom.
 */

struct dpatom {
  int exists;          /* If atom that indexes this node exists */
  int value;           /* ATOM_TRUE, ATOM_FALSE, ATOM_NOT_ASSIGNED */
  int enqueued_value;  /* ATOM_TRUE, ATOM_FALSE, ATOM_NOT_ASSIGNED */
  int num_pos_occ;     /* number of positive occurrences */
  int num_neg_occ;     /* number of negative occurrences */
  /* The following 2 are dynamically allocated arrays of ptrs to clauses. */
  DPclause *pos_occ;  /* array of pos occurrences (ptrs to clauses) */
  DPclause *neg_occ;  /* array of neg occurrences (ptrs to clauses) */
};

/* dpnode -- We don't do Davis-Putnam recursively (because we have to
 * suspend search when we find a model and restart later to get more
 * models).  The state of the search is maintained as a stack of these
 * dpnodes.  Each contains the literal (positive or negative), and the
 * unit propaggations that occur at the node.
 */

struct dpnode {
  int lit;           /* literal made true at this node */
  int *propagate;    /* array of literals made true by unit propagation */
  int num_propagate; /* size of array */
};

/* dp_state -- This is a giant structure that contains all of the data
 * on a Davis-Putnam search.  It contains the list of clauses, the array
 * of atoms, the unit propagation queue, the stack for simulating
 * DP recursion, statistics, and memory management stuff.
 */

struct dp_state {

  /* search parameters (first group set in init; others by separate call) */

  int       millisec_limit;    /* -1 for no limit */
  int       megabyte_limit;    /* -1 for no limit */

  int       subsume;           /* whether or not to mark subsumed clauses */
  FILE      *report_file;      /* file to which reports are sent */

  /* the list of clauses */

  DPclause  first_clause;
  DPclause  last_clause;
  DPclause  *selectable_clauses;  /* for select_atom() */

  /* the array of atoms */

  DPatom    atoms;
  int       greatest_atom;

  /* queue for unit propagation */

  int       *unit_queue;      /* queue for unit propagation */
  int       unit_queue_size;  /* one less than this will fit, because of 0 */
  int       unit_queue_first;
  int       unit_queue_last;

  /* stack for DP recursion */

  DPnode    stack;
  int       top;
  int       *work;  /* storage for propagate arrays in stack */

  /* if the caller asks for a model, it is placed here */

  int       *model;

  /* statistics on the DP search */

  int       num_clauses;
  int       lit_occ_input;
  int       preprocess_unit_assigns;
  int       lit_occ_after_sub;
  int       clauses_after_sub;
  int       preprocess_millisec;
  int       num_selectable_clauses;
  int       splits;
  int       unit_assigns;
  int       failed_paths;
  int       search_millisec;
  int       models;

  /* nonstandard memory management -- We don't use standard memory
   * management (from memory.h), because memory usage can be an very
   * high.  Instead, we do most of it locally, and free it to the OS
   * when done.  The method (with dp_alloc()) is similar to the
   * standard method (with tp_alloc()), except that we maintain a list
   * of the blocks malloced.
   */

  void      *alloc_block;    /* location returned by most recent malloc */
  void      *alloc_pos;      /* current position in block */
  Glist     blocks;          /* list of pointers to blocks we have malloced */
  int       malloc_calls;
  int       free_calls;
  int       bytes_dp_alloced;

  /* for standard memory management of this node (tp_alloc) */

  DP_state  next;             /* for avail list */

};  /* end of struct dp_state */

/*
 * memory management (standard method)
 */

static DP_state dp_state_avail;
static unsigned dp_state_gets, dp_state_frees, dp_state_avails;

/* end of counters for each type */

/*************
 *
 *    DP_state get_dp_state()
 *
 *************/

static
DP_state get_dp_state(void)
{
  DP_state p;

  dp_state_gets++;
  if (dp_state_avail == NULL)
    p = tp_alloc(sizeof(struct dp_state));
  else {
    dp_state_avails--;
    p = dp_state_avail;
    dp_state_avail = dp_state_avail->next;
  }
  /* initialization */	
  	
  p->millisec_limit = 0;
  p->megabyte_limit = 0;

  p->subsume = 0;
  p->report_file = NULL;

  p->first_clause = NULL;
  p->last_clause = NULL;
  p->selectable_clauses = NULL;

  p->atoms = NULL;
  p->greatest_atom = 0;

  p->unit_queue = NULL;
  p->unit_queue_size = 0;
  p->unit_queue_first = 0;
  p->unit_queue_last = 0;

  p->stack = NULL;
  p->top = 0;
  p->work = NULL;

  p->model = NULL;

  p->num_clauses = 0;
  p->lit_occ_input = 0;
  p->preprocess_unit_assigns = 0;
  p->lit_occ_after_sub = 0;
  p->clauses_after_sub = 0;
  p->preprocess_millisec = 0;
  p->num_selectable_clauses = 0;
  p->splits = 0;
  p->unit_assigns = 0;
  p->failed_paths = 0;
  p->search_millisec = 0;
  p->models = 0;

  p->alloc_block = NULL;
  p->alloc_pos = NULL;
  p->blocks = NULL;
  p->malloc_calls = 0;
  p->free_calls = 0;
  p->bytes_dp_alloced = 0;

  p->next = NULL;

  /* end of initialization */

  return(p);
}  /* get_dp_state */

/*************
 *
 *    free_dp_state()
 *
 *************/

static
void free_dp_state(DP_state p)
{
  dp_state_frees++;
  dp_state_avails++;
  p->next = dp_state_avail;
  dp_state_avail = p;
}  /* free_dp_state */

/* end of get and free routines for each type */

/*************
 *
 *   fprint_putnam_mem()
 *
 *************/

/* DOCUMENTATION
This routine prints (to FILE *fp) memory usage statistics for data types
associated with Davis-Putnam.
The Boolean argument heading tells whether to print a heading on the table.
(This does not include the vast majority of the DP memory usage,
which is managed privately.  See fprint_dp_report().)
*/

/* PUBLIC */
void fprint_putnam_mem(FILE *fp, BOOL heading)
{
  int n;
  if (heading)
    fprintf(fp, "  type (bytes each)        gets      frees     in use      avail      bytes\n");

  n = sizeof(struct dp_state);
  fprintf(fp, "dp_state (%4d)     %11u%11u%11u%11u%9.1f K\n",
	  n, dp_state_gets, dp_state_frees,
	  dp_state_gets - dp_state_frees, dp_state_avails,
	  (((dp_state_gets - dp_state_frees) + dp_state_avails) * n) / 1024.);

  /* end of printing for each type */
  
}  /* fprint_putnam_mem */

/*************
 *
 *   p_putnam_mem()
 *
 *************/

/* DOCUMENTATION
This routine prints (to stdout) memory usage statistics for data types
associated with Davis-Putnam.
(This does not include the vast majority of the DP memory usage,
which is managed privately.  See fprint_dp_report().)
*/

/* PUBLIC */
void p_putnam_mem()
{
  fprint_putnam_mem(stdout, 1);
}  /* p_putnam_mem */

/*
 *  end of memory management
 */

/*
 * SPECIAL (nonstandard) memory management
 */

/*************
 *
 *   jump_or_die()
 *
 *   A memory problem has occurred.  Either longjmp() or die.
 *
 *************/

static
void jump_or_die(int type)
{
#ifdef TP_SETJMP
  longjmp(Env, type);
#else
  switch (type) {
  case JMP_TOO_BIG:  fatal_error("dp_alloc, request too big");     break;
  case JMP_TOO_MUCH: fatal_error("dp_alloc, memory limit");        break;
  case JMP_OS_OUT:   fatal_error("dp_alloc, OS is out of memory"); break;
  default:           fatal_error("dp_alloc, unknown error");       break;
  }
#endif
}  /* jump_or_die */

/*************
 *
 *   memory_limit()
 *
 *************/

static
int memory_limit(DP_state dp)
{
  if (dp->megabyte_limit == -1)
    return 0;
  else
    /* Check if one more call would put us over the limit. */
    return (dp->malloc_calls+1)*MALLOC_BYTES > dp->megabyte_limit*1024*1024;
}  /* memory_limit */

/*************
 *
 *    void *dp_alloc()
 *
 *    Allocate n contiguous bytes, aligned on pointer boundry.
 *
 *************/

static
void *dp_alloc(DP_state dp, size_t n)
{
  char *return_block;
  int scale;

  /* If n is not a multiple of sizeof(void *), round up so that it is */
  
  scale = sizeof(void *);
  if (n % scale != 0)
    n = n + (scale - (n % scale));

  /* If first call or not enough in the current block, get a new block. */
  
  if (dp->alloc_block==NULL ||
      dp->alloc_block + MALLOC_BYTES - dp->alloc_pos < n) {
    if (n > MALLOC_BYTES)
      jump_or_die(JMP_TOO_BIG);
    else if (memory_limit(dp))
      jump_or_die(JMP_TOO_MUCH);
    else {
      dp->alloc_pos = dp->alloc_block = my_malloc(MALLOC_BYTES);
      dp->malloc_calls++;
      if (!dp->alloc_pos)
	jump_or_die(JMP_OS_OUT);
      else {
	/* Add new block to list, so that it can be freed later. */
	Glist g = get_glist();
	g->u.v = dp->alloc_block;
	g->next = dp->blocks;
	dp->blocks = g;
      }
    }
  }
  return_block = dp->alloc_pos;
  dp->alloc_pos += n;
  dp->bytes_dp_alloced += n;
  return(return_block);
}  /* dp_alloc */

/*
 *  end of SPECIAL memory management
 */

/*************
 *
 *   subsumed()
 *
 *   Check if any of the literals is assigned TRUE.
 *
 *************/

static
BOOL subsumed(DP_state dp, DPclause c)
{
  int i;
  for (i = 0; i < c->num_pos; i++)
    if (dp->atoms[c->pos[i]].value == ATOM_TRUE)
      return TRUE;
  for (i = 0; i < c->num_neg; i++)
    if (dp->atoms[c->neg[i]].value == ATOM_FALSE)
      return TRUE;
  return FALSE;
}  /* subsumed */

/*************
 *
 *   init_unit_queue()
 *
 *   Initialize the unit propagation queue.
 *
 *************/

static
void init_unit_queue(DP_state dp)
{
  dp->unit_queue_first = 0;
  dp->unit_queue_last = -1;
}  /* init_unit_queue */

/*************
 *
 *   unit_enqueue()
 *
 *   The given clause is assumed to have exactly 1 active literal.
 *   The clause may be subsumed.  The queue is not allowed to have
 *   duplicate or complementary literals.  Return either
 *   UNSATISFIABLE (if the negation of the literal is already in the queue) or
 *   NOT_DETERMINED.
 *
 *************/

static
int unit_enqueue(DP_state dp, DPclause c)
{
  ATOM_INT *ip;
  int lit, ev, i;

  /* First find the unit (*ip gets atom, lit gets literal).
   * If subsumption is disabled, then a clause may have "true"
   * literals that are included in the "active" count; in this
   * case, c's "active" literal is already true, so it is not
   * enqueued.
   */

  if (c->active_pos == 1) {
    for (ip = c->pos, i = 0;
	 i < c->num_pos && dp->atoms[*ip].value != ATOM_NOT_ASSIGNED;
	 ip++, i++);
    if (i == c->num_pos) {
      /* Subsumption must be off, and c is really subsumed. */
      return NOT_DETERMINED;
    }
    else
      lit = *ip;
  }
  else {
    for (ip = c->neg, i = 0;
	 i < c->num_neg && dp->atoms[*ip].value != ATOM_NOT_ASSIGNED;
	 ip++, i++);
    if (i == c->num_neg) {
      /* Subsumption must be off, and c is really subsumed. */
      return NOT_DETERMINED;
    }
    else
      lit = -(*ip);
  }

  /* Check if the atom is already in the queue. */

  ev = dp->atoms[*ip].enqueued_value;

  if (ev == ATOM_NOT_ASSIGNED) {
    /* Enqueue the literal. */
    dp->atoms[*ip].enqueued_value = (lit > 0 ? ATOM_TRUE : ATOM_FALSE);
    dp->unit_queue_last++;
    if (dp->unit_queue_last == dp->unit_queue_size)
      fatal_error("unit_enqueue: queue too small");
    dp->unit_queue[dp->unit_queue_last] = lit;
    return NOT_DETERMINED;
  }
  else if ((ev == ATOM_TRUE  && lit < 0) ||
	   (ev == ATOM_FALSE && lit > 0)) {
    /* already in the queue with opposite sign */
    return UNSATISFIABLE;
  }
  else {
    /* already in the queue with same sign */
    return NOT_DETERMINED;
  }
}  /* unit_enqueue */

/*************
 *
 *   unit_dequeue()
 *
 *   Also reset the enqueued_value field of the atom.
 *
 *************/

static
int unit_dequeue(DP_state dp)
{
  int lit;

  if (dp->unit_queue_last < dp->unit_queue_first)
    return 0;
  else {
    lit = dp->unit_queue[dp->unit_queue_first];
    dp->unit_queue_first++;
    dp->atoms[abs(lit)].enqueued_value = ATOM_NOT_ASSIGNED;
    return lit;
  }
}  /* unit_dequeue */

#if TRACE

/*************
 *
 *   p_dp_clause()
 *
 *************/

static
void p_dp_clause(DPclause c)
{
  int j;
    
  printf("( ");
  for (j = 0; j < c->num_neg; j++)
    printf("-%d ", c->neg[j]);
  for (j = 0; j < c->num_pos; j++)
    printf("%d ", c->pos[j]);
  printf(") active_neg=%d, active_pos=%d, subsumer=%d\n", c->active_neg,
	 c->active_pos, c->subsumer);
}  /* p_clause */

/*************
 *
 *   p_atom()
 *
 *************/

static
void p_atom(DP_state dp, int i)
{
  int j;
  DPatom a = dp->atoms + i;

  printf("DPatom=%d, value=%d, pos_occ=", i, a->value);
  for (j = 0; j < a->num_pos_occ; j++)
    printf("%x,", (unsigned) a->pos_occ[j]);
  printf(" neg_occ=");
  for (j = 0; j < a->num_neg_occ; j++)
    printf("%x,", (unsigned) a->neg_occ[j]);
  printf(".\n");
}  /* p_atom */

#endif

/*************
 *
 *   setup_unit_queue()
 *
 *************/

static
void setup_unit_queue(DP_state dp)
{
  int bytes;
  /* Allocate queue for unit propagation (dp->unit_queue is a global variable).
   * dp->greatest_atom must already be known.
   */

  dp->unit_queue_size = dp->greatest_atom + 1;
  bytes = dp->unit_queue_size * sizeof(DPclause *);
  dp->unit_queue = dp_alloc(dp, bytes);
  init_unit_queue(dp);
}  /* setup_unit_queue */

/*************
 *
 *   setup_model_array()
 *
 *************/

static
void setup_model_array(DP_state dp)
{
  int bytes = dp->greatest_atom * sizeof(int);
  dp->model = dp_alloc(dp, bytes);
}  /* setup_model_array */

/*************
 *
 *   setup_atom_array()
 *
 *   Assume all clauses have been inserted into the DP set.  This routine
 *   sets up atom array (which is used for accessing clauses that contain
 *   a given atom).
 *
 *************/

static
void setup_atom_array(DP_state dp)
{
  int i, j;
  DPclause c;
  DPatom a;

#ifdef TRACE
  for (c = dp->first_clause; c; c = c->next)
    p_dp_clause(c);
#endif
  /* Now set up the atom array (indexed with 1--dp->greatest_atom).
   * Each atom struct has lists of positive occurrences and
   * negative occurrences.  To save space, the lists are stored
   * as arrays; but we don't know how big the arrays should be
   * until we process the clauses.  So we make two passes through
   * the clauses.
   */

  i = (dp->greatest_atom+1) * sizeof(struct dpatom);
  dp->atoms = dp_alloc(dp, i);

  /* Some atoms (e.g. 0) don't exist, but that's okay. */
  for (i = 0, a = dp->atoms; i <= dp->greatest_atom; i++, a++) {
    a->value = ATOM_NOT_ASSIGNED;
    a->enqueued_value = ATOM_NOT_ASSIGNED;
    a->num_pos_occ = 0;
    a->pos_occ = NULL;
    a->num_neg_occ = 0;
    a->neg_occ = NULL;
    a->exists = 0;
  }

  /* For each literal of each clause, increment occurrence count
   * in atom structure.
   */

  for (c = dp->first_clause; c; c = c->next) {
    for (j = 0; j < c->num_pos; j++) {
      a = dp->atoms + c->pos[j];
      a->num_pos_occ++;
      a->exists = 1;
    }
    for (j = 0; j < c->num_neg; j++) {
      a = dp->atoms + c->neg[j];
      a->num_neg_occ++;
      a->exists = 1;
    }
  }

  /* For each atom, allocate occurrence arrays, and reset occurrence
   * counts to 0.
   */

  for (i = 0, a = dp->atoms; i <= dp->greatest_atom; i++, a++) {
    dp->lit_occ_input += a->num_pos_occ;
    a->pos_occ = dp_alloc(dp, a->num_pos_occ * sizeof(DPclause));
    a->num_pos_occ = 0;

    dp->lit_occ_input += a->num_neg_occ;
    a->neg_occ = dp_alloc(dp, a->num_neg_occ * sizeof(DPclause));
    a->num_neg_occ = 0;
  }

  /* For each literal of each clause, add clause to occurrence
   * array of atom.
   */

  for (c = dp->first_clause; c; c = c->next) {
    for (j = 0; j < c->num_pos; j++) {
      a = dp->atoms + c->pos[j];
      a->pos_occ[a->num_pos_occ++] = c;
    }
    for (j = 0; j < c->num_neg; j++) {
      a = dp->atoms + c->neg[j];
      a->neg_occ[a->num_neg_occ++] = c;
    }
  }
}  /* setup_atom_array */

/*************
 *
 *   setup_selection_array()
 *
 *   Initialize the mechanism for selecting atoms for splitting;
 *
 *************/

static
void setup_selection_array(DP_state dp)
{
  DPclause c;
  int j;
    
  /* We will select literals from (smallest) positive clauses for splitting.
   * To avoid scanning all clauses, we build an array of pointers
   * to all of the nonsubsumed clauses with 2 or more positive literals.
   * (Horn clauses will be handled by unit propagation.)
   */

  for (c = dp->first_clause, dp->num_selectable_clauses = 0; c; c = c->next) {
    /* Recall that even if subsumption is disabled, it is done
     * during preprocessing, which has just occurred.
     */
    if (!c->subsumer && c->num_pos >= 2)
      dp->num_selectable_clauses++;
  }

  /* Set up array of selectable clauses for select_atom(). */

  j = dp->num_selectable_clauses * sizeof(DPclause);
  dp->selectable_clauses = dp_alloc(dp, j);

  for (c = dp->first_clause, j = 0; c; c = c->next)
    if (!c->subsumer && c->num_pos >= 2)
      dp->selectable_clauses[j++] = c;
}  /* setup_selection_array */

/*************
 *
 *   select_atom()
 *
 *   Go through the array of selectable clauses, and select the first active
 *   literal in the first shortest nonsubsumed clause.
 *
 *************/

#define SUBSUMED(dp,c) (dp->subsume ? c->subsumer : subsumed(dp,c))

static
int select_atom(DP_state dp)
{
  ATOM_INT *ip;
  int min, atom, i;
  DPclause c;

  min = LIT_INT_MAX;
  atom = 0;
  for (i = 0; i < dp->num_selectable_clauses; i++) {
    c = dp->selectable_clauses[i];
    if (c->active_neg == 0 && c->active_pos < min && !SUBSUMED(dp,c)) {
      min = c->active_pos;
      for (ip = c->pos; dp->atoms[*ip].value != ATOM_NOT_ASSIGNED; ip++);
      atom = *ip;
    }
  }
  return atom;
}  /* select_atom */

/*************
 *
 *   setup_stack()
 *
 *************/

static
void setup_stack(DP_state dp)
{
  int bytes;
  bytes = dp->greatest_atom * sizeof(struct dpnode);
  dp->stack = dp_alloc(dp, bytes);
  bytes = dp->greatest_atom * sizeof(int);
  dp->work = dp_alloc(dp, bytes);
}  /* setup_stack */

/*************
 *
 *   assign(lit) - make a literal true
 *
 *   1. Make the assignment in the atom array.
 *   2. (optional) Mark nonsubsumed clauses that contain it as subsumed by it.
 *   3. Unit resolution on clauses that are not marked as subsumed.
 *      If a unit is found, enqueue it for unit propagation.
 *   4. If empty clause is found, return UNSATISFIABLE else NOT_DETERMINED.
 *
 *************/

static
int assign(DP_state dp, int lit, BOOL subsume)
{
  DPatom a;
  DPclause c;
  int atom, i, rc, n, np, nn;

  atom = abs(lit);
  a = dp->atoms + atom;

  if (a->value != ATOM_NOT_ASSIGNED)
    fatal_error("assign: atom already assigned");

  np = a->num_pos_occ;
  nn = a->num_neg_occ;
  a->value = (lit < 0 ? ATOM_FALSE : ATOM_TRUE);
  rc = NOT_DETERMINED;

  if (a->value == ATOM_TRUE) {  /* Assign true to atom */
    if (subsume) 
      for (i = 0; i < np; i++) {
	c = a->pos_occ[i];
	if (c->subsumer == 0)
	  c->subsumer = atom;
      }
    for (i = 0; i < nn; i++) {
      c = a->neg_occ[i];
      if (!c->subsumer) {
	c->active_neg--;
	n = c->active_pos + c->active_neg;
	if (n == 0)
	  rc = UNSATISFIABLE;
	else if (n == 1 && unit_enqueue(dp, c) == UNSATISFIABLE)
	  rc = UNSATISFIABLE;
      }
    }
  }

  else {  /* Assign false to atom */
    if (subsume)
      for (i = 0; i < nn; i++) {
	c = a->neg_occ[i];
	if (c->subsumer == 0)
	  c->subsumer = atom;
      }
    for (i = 0; i < np; i++) {
      c = a->pos_occ[i];
      if (!c->subsumer) {
	c->active_pos--;
	n = c->active_neg + c->active_pos;
	if (n == 0)
	  rc = UNSATISFIABLE;
	else if (n == 1 && unit_enqueue(dp, c) == UNSATISFIABLE)
	  rc = UNSATISFIABLE;
      }
    }
  }
  if (rc == UNSATISFIABLE)
    dp->failed_paths++;
  return rc;
}  /* assign */

/*************
 *
 *   unassign(lit) -- undo an assignment
 *
 *************/

static
void unassign(DP_state dp, int lit, int subsume)
{
  DPatom a;
  DPclause c;
  int i, atom, np, nn;
    
  atom = abs(lit);
  a = dp->atoms + atom;

  if (a->value == ATOM_NOT_ASSIGNED)
    fatal_error("unassign: atom not assigned");

  np = a->num_pos_occ;
  nn = a->num_neg_occ;
    
  if (a->value == ATOM_TRUE) {  /* Unassign true. */
    if (subsume)
      for (i = 0; i < np; i++) {
	c = a->pos_occ[i];
	if (c->subsumer == atom)
	  c->subsumer = 0;
      }
    for (i = 0; i < nn; i++) {
      c = a->neg_occ[i];
      if (!c->subsumer)
	c->active_neg++;
    }
  }
  else {  /* Unassign false. */
    if (subsume)
      for (i = 0; i < nn; i++) {
	c = a->neg_occ[i];
	if (c->subsumer == atom)
	  c->subsumer = 0;
      }
    for (i = 0; i < np; i++) {
      c = a->pos_occ[i];
      if (!c->subsumer)
	c->active_pos++;
    }
  }
  a->value = ATOM_NOT_ASSIGNED;
}  /* unassign */

/*************
 *
 *   delete_pointers_to_subsumed_clauses()
 *
 *   This routine is called after the initial unit propagation (before
 *   any splitting).  It goes through the occurrence lists in the
 *   atom array and deletes references to subsumed clauses.  The reason
 *   for this is simply to speed traversals during assign() and unassign().
 *
 *************/

static
void delete_pointers_to_subsumed_clauses(DP_state dp)
{
  int i, j, k;
  DPatom a;

  /* The set of occurrences, say positive, of an atom, is kept as
   * an array of pointers to clauses.  Pointers to subsumed clauses
   * are removed, and the the others are shifted left the apropriate
   * amount.  Same for negative occurrences.
   */

  for (i = 1; i <= dp->greatest_atom; i++) {
    a = dp->atoms + i;
    for (j = 0, k = 0; j < a->num_pos_occ; j++)
      if (a->pos_occ[j]->subsumer == 0)
	a->pos_occ[k++] = a->pos_occ[j];
    a->num_pos_occ = k;
    dp->lit_occ_after_sub += k;

    for (j = 0, k = 0; j < a->num_neg_occ; j++)
      if (a->neg_occ[j]->subsumer == 0)
	a->neg_occ[k++] = a->neg_occ[j];
    a->num_neg_occ = k;
    dp->lit_occ_after_sub += k;
  }
}  /* delete_pointers_to_subsumed_clauses */

/*************
 *
 *   assign_and_propagate()
 *
 *************/

static
int assign_and_propagate(DP_state dp, DPnode s)
{
  int rc, unit;
  int n = 0;

#ifdef TRACE
  printf("Assign: %d\n", s->lit);
#endif
  init_unit_queue(dp);
  rc = assign(dp, s->lit, dp->subsume);
  while (rc == NOT_DETERMINED && (unit = unit_dequeue(dp)) != 0) {
    dp->unit_assigns++;
#ifdef TRACE
    printf("UP assign: %d\n", unit);
#endif
    rc = assign(dp, unit, dp->subsume);
    /* Save the assignment so that it can be undone. */
    s->propagate[n++] = unit;
  }
  if (rc == UNSATISFIABLE)
    while ((unit = unit_dequeue(dp)) != 0);

  s->num_propagate = n;
  return rc;
}  /* assign_and_propagate */

/*************
 *
 *   time_limit()
 *
 *************/

static
BOOL time_limit(DP_state dp)
{
  if (dp->millisec_limit == -1)
    return FALSE;
  else
    return clock_value(Utility_clock)+dp->search_millisec > dp->millisec_limit;
}  /* time_limit */

/*************
 *
 *   davis_putnam()
 *
 *************/

#define GO_DP     0
#define BACK_DP   1
#define MODEL_DP  2
#define DONE_DP   3
#define TIME_DP   4

static
DP_rc davis_putnam(DP_state dp, BOOL start)
{
  int status, atom, rc;
  DPnode s;
  int *avail;

  clock_reset(Utility_clock);
  clock_start(Utility_clock);

  if (start) {
    status = GO_DP;
    avail = dp->work;
    dp->top = -1;
  }
  else {
    status = (dp->top == -1 ? DONE_DP : BACK_DP);
  }

  while (status == GO_DP || status == BACK_DP) {
    if (time_limit(dp))
      status = TIME_DP;
    else if (status == GO_DP) {             /* GO_DP */
      atom = select_atom(dp);
      if (atom == 0) {
	status = MODEL_DP;
	dp->models++;
      }
      else {
	dp->splits++;
	dp->top++;
	s = dp->stack + dp->top;
	s->propagate = (dp->top == 0 ? dp->work :
			dp->stack[dp->top-1].propagate +
			dp->stack[dp->top-1].num_propagate);
	s->lit = atom;
	rc = assign_and_propagate(dp, s);
	status = (rc == UNSATISFIABLE ? BACK_DP : GO_DP);
      }
    }  /* if GO_DP */
    else {                           /* BACK_DP */
      int i;
      s = dp->stack + dp->top;
      for (i = s->num_propagate-1; i >= 0; i--) {
#ifdef TRACE
	printf("UP unassign: %d\n", s->propagate[i]);
#endif
	unassign(dp, s->propagate[i], dp->subsume);
      }
#ifdef TRACE
      printf("Unassign: %d\n", s->lit);
#endif
      unassign(dp, s->lit, dp->subsume);
      if (s->lit > 0) {
	s->lit = -(s->lit);
	rc = assign_and_propagate(dp, s);
	status = (rc == UNSATISFIABLE ? BACK_DP : GO_DP);
      }
      else {
	dp->top--;
	status = (dp->top == -1 ? DONE_DP : BACK_DP);
      }
    }  /* if BACK_DP */
  }

  clock_stop(Utility_clock);
  dp->search_millisec += clock_value(Utility_clock);

  if (status == TIME_DP)
    return TIME_OUT_RETURN;
  else
    return (status == MODEL_DP ? OK_RETURN : NO_MODEL_RETURN);
}  /* davis_putnam */

/*************
 *
 *   dp_init()
 *
 *************/

/* DOCUMENTATION
This routine allocates a DP_state for a Davis-Putnam search.
<UL>
<LI> seconds_limit -- This is a limit on the number of CPU seconds
     allowed for the DP search.  Calls to dp_append_clause() are not
     included.  The timing is cumulative for dp_first_model(), dp_next_model()
     calls.  If the limit is exceeded, dp_first_model() or dp_next_model()
     will return TIME_OUT_RETURN.
<LI> megabyte_limit -- This is a limit on memory usage for the DP
     search.  If the limit is exceeded, dp_append_clause() or dp_first_model()
     will return MEMORY_OUT_RETURN.  (dp_next_model() does not allocate
     any memory.)
<LI> subsume -- Apply subsumption during the DP search (subsumption is
     always applied during preprocessing).  This is not recommended in
     general.
<LI> report_file -- File to which reports are sent; if NULL, no reports printed.
</UL>
*/

/* PUBLIC */
DP_state dp_init(int seconds_limit,
		 int megabyte_limit,
		 BOOL subsume,
		 FILE *report_file)
{
  DP_state dp = get_dp_state();

  dp->millisec_limit = seconds_limit * 1000;
  dp->megabyte_limit = megabyte_limit;
  dp->subsume = subsume;
  dp->report_file = report_file;

  if (Utility_clock == -1)
    Utility_clock = clock_init("davis-putnam utility clock");
  else
    clock_reset(Utility_clock);
  clock_start(Utility_clock);

  return dp;
}  /* dp_init */

/*************
 *
 *   dp_append_clause_jumper()
 *
 *   "jumper" means that we might longjmp() back through this routine.
 *
 *************/

static
void dp_append_clause_jumper(DP_state dp, int c[], int n)
{
  DPclause d;
  int i, np, nn;

  if (dp->atoms != NULL)
    fatal_error("dp_append_clause_jumper, atoms not NULL");

  /* Find out how many pos and neg literals, so that arrays can be allocated.
   */

  np = nn = 0;
  for (i = 0; i < n; i++) {
    if (c[i] > 0)
      np++;
    else
      nn++;
    if (abs(c[i]) > dp->greatest_atom) {
      dp->greatest_atom = abs(c[i]);
      if (abs(c[i]) >= ATOM_INT_MAX)
	fatal_error("dp_append_clause, atom_too_big.");
    }
  }

  /* Allocate the clause and the literal arrays. */
	
  d = dp_alloc(dp, sizeof(struct dpclause));
  d->next = NULL; d->subsumer = 0;
  d->num_pos = np; d->active_pos = np;
  d->num_neg = nn; d->active_neg = nn;
  if (dp->last_clause)
    dp->last_clause->next = d;
  else
    dp->first_clause = d;
  dp->last_clause = d;

  d->pos = dp_alloc(dp, np * sizeof(ATOM_INT));  /* ok if np==0 */
  d->neg = dp_alloc(dp, nn * sizeof(ATOM_INT));  /* ok if nn==0 */

  /* Fill in the literal arrays (neg literals are stored positively). */

  np = nn = 0;
  for (i = 0; i < n; i++) {
    if (c[i] > 0)
      d->pos[np++] = c[i];
    else
      d->neg[nn++] = -c[i];
  }
  dp->num_clauses++;
}  /* dp_append_clause_jumper */

/*************
 *
 *   dp_append_clause()
 *
 *   Take a propositional clause (array of nonzero integers) and
 *   insert it into the set of DP clauses.
 *
 *************/

/* DOCUMENTATION
This routine adds a propositional clause to a Davis-Putnam set.
The clause is an array of non-zero integers (atoms are positive
integers).
(If the set of clauses has n atoms, it is most efficient to name
them 1, 2, ..., n.)
<P>
Clauses cannot be added to the set after dp_first_model() has been
called.  If you try to do so, a fatal error will occur.
<P>
The return codes are
<UL>
<LI> OK_RETURN -- success
<LI> MEMORY_OUT_RETURN -- the memory limit was exceeded (dp deallocated).
</UL>
*/

/* PUBLIC */
DP_rc dp_append_clause(DP_state dp, int clause[], int number_of_literals)
{
#ifdef TP_SETJMP
  if (setjmp(Env) == 0) {
    dp_append_clause_jumper(dp, clause, number_of_literals);
    return OK_RETURN;
  }
  else {
    dp_cancel(dp);
    return MEMORY_OUT_RETURN;
  }
#else
  /* If we try to use too much memory, a fatal error will occur. */
  dp_append_clause_jumper(dp, clause, number_of_literals);
  return OK_RETURN;
#endif
}  /* dp_append_clause */

/*************
 *
 *   dp_first_model_jumper()
 *
 *   "jumper" means that we might longjmp() back through this routine.
 *
 *************/

static
DP_rc dp_first_model_jumper(DP_state dp)
{
  int rc;
  DP_rc dp_rc;
  int unit, n;
  DPclause c;

  setup_unit_queue(dp);
  setup_atom_array(dp);

  /* Initial unit propagation and subsumption. */

  rc = NOT_DETERMINED;
  for (c = dp->first_clause; c && rc == NOT_DETERMINED; c = c->next)
    if (c->num_pos + c->num_neg == 1)
      rc = unit_enqueue(dp, c);

  while (rc == NOT_DETERMINED && (unit = unit_dequeue(dp)) != 0) {
    dp->preprocess_unit_assigns++;
    /* Always mark subsumed clauses while preprocessing. */
    rc = assign(dp, unit, 1);
  }
  clock_stop(Utility_clock);  /* started in dp_init() */
  dp->preprocess_millisec = clock_value(Utility_clock);

  if (rc == UNSATISFIABLE)
    dp_rc = NO_MODEL_RETURN;
  else {
    if (dp->preprocess_unit_assigns > 0)
      delete_pointers_to_subsumed_clauses(dp);
    else
      dp->lit_occ_after_sub = dp->lit_occ_input;

    for (c = dp->first_clause, n = 0; c; c = c->next)
      if (c->subsumer)
	n++;

    dp->clauses_after_sub = dp->num_clauses - n;

    /* Done with initial unit propagation and subsumption. */

    setup_selection_array(dp);
    setup_stack(dp);
    setup_model_array(dp);

    if (dp->report_file != NULL)
      fprint_dp_preprocess_report(dp->report_file, dp);

    dp_rc = davis_putnam(dp, 1);
  }

  if (dp_rc != OK_RETURN)
    dp_cancel(dp);
  return dp_rc;
}  /* dp_first_model_jumper */

/*************
 *
 *   dp_first_model()
 *
 *************/

/* DOCUMENTATION
This routine starts the Davis-Putnam search.  The return codes are
<UL>
<LI> OK_RETURN -- a model has been found (see dp_get_model()).
<LI> NO_MODEL_RETURN -- there are no models (dp deallocated).
<LI> TIME_OUT_RETURN -- the time limit was exceeded (dp deallocated).
<LI> MEMORY_OUT_RETURN -- the memory limit was exceeded (dp deallocated).
</UL>
*/

/* PUBLIC */
DP_rc dp_first_model(DP_state dp)
{
#ifdef TP_SETJMP
  if (setjmp(Env) == 0)
    return dp_first_model_jumper(dp);
  else {
    dp_cancel(dp);
    return MEMORY_OUT_RETURN;
  }
#else
  /* If we try to use too much memory, a fatal error will occur. */
  return dp_first_model_jumper(dp);
#endif
}  /* dp_first_model */

/*************
 *
 *   dp_next_model()
 *
 *   We don't allocate any more memory, so setjmp/longjmp are not needed.
 *
 *************/

/* DOCUMENTATION
This routine continues the Davis-Putnam search.  The return codes are
<UL>
<LI> OK_RETURN -- another model has been found (see dp_get_model()).
<LI> NO_MODEL_RETURN -- there are no more models (dp deallocated).
<LI> TIME_OUT_RETURN -- the time limit was exceeded (dp deallocated).
</UL>
*/

/* PUBLIC */
DP_rc dp_next_model(DP_state dp)
{
  DP_rc rc;
  rc = davis_putnam(dp, 0);
  if (rc != OK_RETURN)
    dp_cancel(dp);
  return rc;
}  /* dp_next_model */

/*************
 *
 *   dp_get_model()
 *
 *************/

/* DOCUMENTATION
This routine can be called to retrieve the current model after
dp_first_model() or dp_next_model() returns with OK_RETURN.
The form of the model is an array (*literals) of nonzero integers (sorted
by absolute value) giving the literals true in the model.
The size of the array (*num_literals) is the number of atoms in the set of
clauses.
*/

/* PUBLIC */
void dp_get_model(DP_state dp, int **literals, int *num_literals)
{
  int i, n;
  DPatom a;

  n = 0;

  for (i = 0, a = dp->atoms; i <= dp->greatest_atom; i++, a++) {
    if (a->exists) {
      if (a->value == ATOM_TRUE)
	dp->model[n] = i;
      else if (a->value == ATOM_FALSE)
	dp->model[n] = -i;
      else  /* ATOM_NOT_ASSIGNED */
	/* assume false because of selection method */
	dp->model[n] = -i;
      n++;
    }
  }
  *num_literals = n;
  *literals = dp->model;
}  /* get_model */

/*************
 *
 *   dp_cancel()
 *
 *************/

/* DOCUMENTATION
This routine deallocates a DP_state.  It should be called if
<UL>
<LI> you find some, but not all models;
<LI> you have inserted some clauses with dp_append_clause(), but decide
     not to look for models.
</UL>
This routine should <I>not</I> be called if
<UL>
<LI> a routine returned NO_MODEL_RETURN, TIME_OUT_RETURN, or MEMORY_OUT_RETURN.
</UL>
If the "final_report" flag was given to dp_init(), the
report is printed (to stdout).
*/

/* PUBLIC */
void dp_cancel(DP_state dp)
{
  Glist g1;

  g1 = dp->blocks;
  while (g1 != NULL) {
    Glist g2 = g1;
    g1 = g1->next;
    my_free(g2->u.v);
    free_glist(g2);
    dp->free_calls++;
  }
  dp->blocks = NULL;

  if (dp->report_file != NULL)
    fprint_dp_report(dp->report_file, dp);

  free_dp_state(dp);
}  /* dp_cancel */

/*************
 *
 *   fprint_dp_preprocess_report()
 *
 *************/

/* DOCUMENTATION
This routine prints (to FILE *fp) a report containing all statistics
for a Davis-Putnam search.
*/

/* PUBLIC */
void fprint_dp_preprocess_report(FILE *fp, DP_state dp)
{
  fprintf(fp, "\nReport After All Preprocessing\n\n");

  fprintf(fp, "Input:\n");
  fprintf(fp, "   num_clauses:             %8d\n", dp->num_clauses);
  fprintf(fp, "   lit_occ_input:           %8d\n", dp->lit_occ_input);
  fprintf(fp, "   greatest_atom:           %8d\n", dp->greatest_atom);

  fprintf(fp, "Unit preprocess:\n");
  fprintf(fp, "   preprocess_unit_assigns: %8d\n", dp->preprocess_unit_assigns);
  fprintf(fp, "   clauses_after_sub:       %8d\n", dp->clauses_after_sub);
  fprintf(fp, "   lit_occ_after_sub:       %8d\n", dp->lit_occ_after_sub);
  fprintf(fp, "   num_selectable_clauses:  %8d\n", dp->num_selectable_clauses);
  fprintf(fp, "   input & preprocess time: %8.2f\n", dp->preprocess_millisec / 1000.);
  fprintf(fp, "   malloc calls (total megs):%7d (%d)\n", dp->malloc_calls,
	  (MALLOC_BYTES * dp->malloc_calls) / (1024*1024));
  fprintf(fp, "   bytes_dp_alloced:        %8d\n", dp->bytes_dp_alloced);

  fflush(fp);
}  /* fprint_dp_preprocess_report */

/*************
 *
 *   fprint_dp_report()
 *
 *************/

/* DOCUMENTATION
This routine prints (to FILE *fp) a report containing all statistics
for a Davis-Putnam search.
*/

/* PUBLIC */
void fprint_dp_report(FILE *fp, DP_state dp)
{
  fprintf(fp, "\nComplete Report\n\n");

  fprintf(fp, "Input:\n");
  fprintf(fp, "   num_clauses:             %8d\n", dp->num_clauses);
  fprintf(fp, "   lit_occ_input:           %8d\n", dp->lit_occ_input);
  fprintf(fp, "   greatest_atom:           %8d\n", dp->greatest_atom);

  fprintf(fp, "Unit preprocess:\n");
  fprintf(fp, "   preprocess_unit_assigns: %8d\n", dp->preprocess_unit_assigns);
  fprintf(fp, "   clauses_after_sub:       %8d\n", dp->clauses_after_sub);
  fprintf(fp, "   lit_occ_after_sub:       %8d\n", dp->lit_occ_after_sub);
  fprintf(fp, "   num_selectable_clauses:  %8d\n", dp->num_selectable_clauses);
  fprintf(fp, "   input & preprocess time: %8.2f\n", dp->preprocess_millisec / 1000.);

  fprintf(fp, "Decide:\n");
  fprintf(fp, "   splits:                  %8d\n", dp->splits);
  fprintf(fp, "   unit_assigns:            %8d\n", dp->unit_assigns);
  fprintf(fp, "   failed_paths:            %8d\n", dp->failed_paths);
  fprintf(fp, "   models:                  %8d\n", dp->models);
  fprintf(fp, "   decide time:             %8.2f\n", dp->search_millisec / 1000.);

  fprintf(fp, "Memory:\n");
  fprintf(fp, "   malloc calls (total megs):%7d (%d)\n", dp->malloc_calls,
	  (MALLOC_BYTES * dp->malloc_calls) / (1024*1024));
  fprintf(fp, "   free calls:              %8d\n", dp->free_calls);
  fprintf(fp, "   bytes_dp_alloced:        %8d\n", dp->bytes_dp_alloced);

  fflush(fp);
}  /* fprint_dp_report */

/*************
 *
 *   dp_atom_value()
 *
 *************/

/* DOCUMENTATION
This routine returns the current value of an atom.
It assumes you have a model
(that is, dp_first_model() or dp_next_model() just returned OK_RETURN).
The return is ATOM_TRUE or ATOM_FALSE.
If the atom is out of range, ATOM_FALSE is returned.
*/

/* PUBLIC */
int dp_atom_value(DP_state dp, int atom)
{
  if (atom < 1 || atom > dp->greatest_atom)
    return ATOM_FALSE;
  else if (dp->atoms[atom].value == ATOM_TRUE)
    return ATOM_TRUE;
  else
    /* Recall that we have a model, so "not assigned" means false. */
    return ATOM_FALSE;
}  /* dp_atom_value */

/*************
 *
 *   dp_p_model()
 *
 *************/

/* DOCUMENTATION
*/

/* PUBLIC */
void dp_p_model(DP_state dp)
{
  int i;
  for (i = 1; i <= dp->greatest_atom; i++)
    printf("atom %d=%d.\n", i, dp->atoms[i].value);
}  /* dp_p_model */
