#ifndef TP_PUTNAM_H
#define TP_PUTNAM_H

#include "clock.h"
#include "glist.h"

/* INTRODUCTION
This package is an implementation of the Davis-Putnam procedure for
testing satisfiability of propositional CNF formulas.  It uses
a straightforward algorithm.  The selection function is to
use the first literal in the first, shortest positive clause.
<P>
The search is incremental, returning models as they are found.
How to use the package:
<UL>
<LI> dp_init() -- initialization.
<LI> dp_append_clause() -- assert clauses.
<LI> dp_first_model() -- look for a model.
<LI> dp_get_model() -- (optional) retrieve a model which has been found.
<LI> dp_next_model() -- (optional) look for another model.
<LI> dp_cancel() -- use this if you don't want all models.
</UL>
If you go through the complete sequence of models with calls to
dp_first_model() and dp_next_model(), you don't necessarily get all of
the models.  The reason is that if the search gets to a state with no
positive clauses, it simply makes all remaining atoms false, and
reports a model.  For example, with the two clauses (-1 | -2) and
(-3), one model (everything false) will be reported, but there are
actually 2 more models.
*/

/* Public definitions */

typedef struct dp_state * DP_state;

/* return codes for first_model(), next_model(), and dp_insert_clause() */

typedef enum { OK_RETURN,
	       NO_MODEL_RETURN,
	       TIME_OUT_RETURN,
	       MEMORY_OUT_RETURN,
             } DP_rc;

/* values of propositional variables */

#define ATOM_FALSE          0
#define ATOM_TRUE           1
#define ATOM_NOT_ASSIGNED   2

/* End of public definitions */

/* Public function prototypes from putnam.c */

void fprint_putnam_mem(FILE *fp, BOOL heading);

void p_putnam_mem();

DP_state dp_init(int seconds_limit,
		 int megabyte_limit,
		 BOOL subsume,
		 FILE *report_file);

DP_rc dp_append_clause(DP_state dp, int clause[], int number_of_literals);

DP_rc dp_first_model(DP_state dp);

DP_rc dp_next_model(DP_state dp);

void dp_get_model(DP_state dp, int **literals, int *num_literals);

void dp_cancel(DP_state dp);

void fprint_dp_preprocess_report(FILE *fp, DP_state dp);

void fprint_dp_report(FILE *fp, DP_state dp);

int dp_atom_value(DP_state dp, int atom);

void dp_p_model(DP_state dp);

#endif  /* conditional compilation of whole file */
