#include "subsume.h"

/* Private definitions and types */

/*************
 *
 *   subsume_literals()
 *
 *   This is a recursive routine that tries to map a list of literals
 *   into a clause d.  If successful, the trail is updated; if failure,
 *   the trail is unchanged.
 *
 *   This version uses ordinary unification.
 *
 *************/

static
BOOL subsume_literals(Literal clit, Context subst, Clause d, Trail *trp)
{
  BOOL subsumed = FALSE;
  Literal dlit;

  if (clit == NULL)
    return TRUE;
  else {
    for (dlit=d->first_literal; !subsumed && dlit!=NULL; dlit=dlit->next) {
      /* printf("testing "); fprint_term(stdout, clit->atom);
      printf("  "); fprint_term(stdout, dlit->atom); printf("\n"); */
      if (clit->sign == dlit->sign) {
	Trail mark = *trp;
	if (match(clit->atom, subst, dlit->atom, trp)) {
	  /* printf("                    Match\n"); */
	  if (subsume_literals(clit->next, subst, d, trp))
	    subsumed = TRUE;
	  else {
	    /* undo bindings from this match only */
	    /* printf("                    UNDO\n"); */
	    undo_subst_2(*trp, mark);
	    *trp = mark;
	  }
	}
      }
    }
    return subsumed;
  }
}  /* subsume_literals */

/*************
 *
 *   subsume_literals()
 *
 *   This version uses backtrack unification.
 *
 *************/

static
BOOL subsume_bt_literals(Literal clit, Context subst,
			 Clause d, Glist *gp)
{
  if (clit == NULL)
    return TRUE;
  else {
    BOOL subsumed = FALSE;
    Literal dlit;
    *gp = plist_prepend(*gp, NULL);

    for (dlit=d->first_literal; !subsumed && dlit!=NULL; dlit=dlit->next) {
      /* printf("testing "); fprint_term(stdout, clit->atom);
      printf("  "); fprint_term(stdout, dlit->atom); printf("\n"); */
      if (clit->sign == dlit->sign) {
	Btm_state bt = match_bt_first(clit->atom, subst, dlit->atom, FALSE);
	while (bt != NULL && !subsumed) {
	  (*gp)->u.v = bt;
	  if (subsume_bt_literals(clit->next, subst, d, gp))
	    subsumed = TRUE;
	  else {
	    bt = match_bt_next(bt);
	  }
	}
      }
    }
    if (subsumed)
      return TRUE;
    else {
      *gp = glist_pop(*gp);
      return FALSE;
    }
  }
}  /* Subsume_bt_literals */

/*************
 *
 *   subsumes()
 *
 *************/

/* DOCUMENTATION
This routine checks if Clause c subsumes Clause d.
Ordinary unification is used; in particular, symmetry of
equality is not built-in.
<P>
*/

/* PUBLIC */
BOOL subsumes(Clause c, Clause d)
{
  Context subst = get_context();
  Trail tr = NULL;
  int rc = subsume_literals(c->first_literal, subst, d, &tr);
  if (rc)
    undo_subst(tr);
  free_context(subst);
  return rc;
}  /* subsumes */

/*************
 *
 *   subsumes_bt()
 *
 *************/

/* DOCUMENTATION
This routine checks if Clause c subsumes Clause d.
Backtrack unification is used; in particular, AC and
commutative/symmetric matching are applied where appropriate.
*/

/* PUBLIC */
BOOL subsumes_bt(Clause c, Clause d)
{
  Context subst = get_context();
  Glist g = NULL;
  int rc = subsume_bt_literals(c->first_literal, subst, d, &g);
  if (rc) {
    /* Cancel the list (stack) of btm_states */
    while (g != NULL) {
      Btm_state bt = g->u.v;
      match_bt_cancel(bt);
      g = glist_pop(g);
    }
  }
  free_context(subst);
  return rc;
}  /* subsumes_bt */

/*************
 *
 *   forward_subsume()
 *
 *************/

/* DOCUMENTATION
*/

/* PUBLIC */
Clause forward_subsume(Clause d, Lindex idx)
{
  Literal dlit;
  Clause subsumer = NULL;
  Context subst = get_context();
  int nd = number_of_literals(d);

  for (dlit=d->first_literal; dlit!=NULL && subsumer==NULL; dlit=dlit->next) {
    Mindex mdx = dlit->sign ? idx->pos : idx->neg;
    Mindex_pos pos;
    Term catom = mindex_retrieve_first(dlit->atom, mdx, GENERALIZATION,
				       NULL, subst, FALSE, &pos);
    BOOL backtrack = lindex_backtrack(idx);
    while (catom != NULL && subsumer == NULL) {
      Clause c = catom->container;
      if (atom_number(c, catom) == 1) {
	int nc = number_of_literals(c);
	/* If c is a unit the we already know it subsumes d; otherwise,
	 * do a full subsumption check on the clauses.  (We don't let
	 * a clause subsume a shorter one, because that would cause
	 * factors to be deleted.)
	 */
	if (nc == 1 || (nc <= nd && (backtrack
				     ? subsumes_bt(c,d)
				     : subsumes(c,d)))) {
	  subsumer = c;
	  mindex_retrieve_cancel(pos);
	}
      }
      if (subsumer == NULL)
	catom = mindex_retrieve_next(pos);
    }
  }
  free_context(subst);
  return subsumer;
}  /* forward_subsume */

/*************
 *
 *   back_subsume()
 *
 *************/

/* DOCUMENTATION
*/

/* PUBLIC */
Glist back_subsume(Clause c, Lindex idx)
{
  int nc = number_of_literals(c);

  if (nc == 0)
    return NULL;
  else {
    Glist subsumees = NULL;
    Context subst = get_context();
    Literal clit = c->first_literal;

    Mindex mdx = clit->sign ? idx->pos : idx->neg;
    Mindex_pos pos;

    Term datom = mindex_retrieve_first(clit->atom, mdx, INSTANCE,
				       subst, NULL, FALSE, &pos);
    BOOL backtrack = lindex_backtrack(idx);
    while (datom != NULL) {
      Clause d = datom->container;
      if (d != c) {  /* in case c is already in idx */
	int nd = number_of_literals(d);
	/* If c is a unit the we already know it subsumes d; otherwise,
	 * do a full subsumption check on the clauses.  (We don't let
	 * a clause subsume a shorter one.)
	 */
	if (nc == 1 || (nc <= nd && (backtrack
				     ? subsumes_bt(c, d)
				     : subsumes(c, d))))
	  subsumees = insert_clause_into_glist(subsumees, d, FALSE);
      }
      datom = mindex_retrieve_next(pos);
    }
    free_context(subst);
    return subsumees;
  }
}  /* back_subsume */

/*************
 *
 *   any_subsumes()
 *
 *************/

/* DOCUMENTATION
Return the first clause in Glist g that subsumes Clause d.
(We don't consider clasues in g that are longer than Clause d.)
*/

/* PUBLIC */
Clause any_subsumes(Glist g, Clause d, Uniftype utype)
{
  int dn = number_of_literals(d);
  Glist p = g;
  Clause subsumer = NULL;

  while (subsumer == NULL && p != NULL) {
    Clause c = p->u.v;
    if (number_of_literals(c) <= dn && (utype == BACKTRACK_UNIF
					? subsumes_bt(c,d)
					: subsumes_bt(c,d)))
      subsumer = c;
    else
      p = p->next;
  }
  return subsumer;
}  /* any_subsumes */

/*************
 *
 *   unit_conflict()
 *
 *************/

/* DOCUMENTATION
*/

/* PUBLIC */
Clause unit_conflict(Clause c, Lindex idx)
{
  if (number_of_literals(c) != 1)
    return NULL;
  else {
    Clause conflicter = NULL;
    Context subst1 = get_context();
    Context subst2 = get_context();
    Mindex mdx = c->first_literal->sign ? idx->neg : idx->pos;
    Mindex_pos pos;
    Term atom = mindex_retrieve_first(c->first_literal->atom, mdx,
				      UNIFY, subst1, subst2, FALSE, &pos);

    while (atom != NULL && conflicter == NULL) {
      Clause d = atom->container;
      if (number_of_literals(d) == 1) {
	conflicter = d;
	mindex_retrieve_cancel(pos);
      }
      else 
	atom = mindex_retrieve_next(pos);
    }
    free_context(subst1);
    free_context(subst2);
    if (conflicter == NULL)
      return NULL;
    else {
      Clause empty_clause = get_clause();
      empty_clause->justification = binary_res_just(c, 1, conflicter, 1);
      return empty_clause;
    }
  }
}  /* unit_conflict */

