#include "term.h"

/* Private definitions and types */

/*
 * memory management
 */

/*
 * Note that this is unusual, because there is an array of avail lists
 * for terms of different arities.  Terms with arities >= MAX_TERM_AVAIL-1
 * all go into the same list.
 */

#define MAX_TERM_AVAIL 5

static Term term_avail[MAX_TERM_AVAIL];
static unsigned term_gets, term_frees, term_avails, term_mem;

/* end of counters for each type */

/*************
 *
 *   Term get_term(arity)
 *
 *************/

static
Term get_term(int arity)
{
  Term p = NULL;
  
  term_gets++;
  
  if (arity < 0 || arity > MAX_ARITY) {
    fatal_error("get_term, arity out of range.");
  }
  else if (arity == 0) {
    if (term_avail[0] != NULL) {
      term_avails--;
      p = term_avail[0];
      term_avail[0] = (Term) p->args;
    }
    else {
      p = tp_alloc(sizeof(struct term));
      p->arity = 0;
      term_mem += sizeof(struct term);
    }
    p->args = NULL;
  }
  else if (arity < MAX_TERM_AVAIL-1) {
    if (term_avail[arity] != NULL) {
      term_avails--;
      p = term_avail[arity];
      term_avail[arity] = p->args[0];
    }
    else {
      p = tp_alloc(sizeof(struct term));
      p->args = tp_alloc(sizeof(Term) * arity);
      p->arity = arity;
      term_mem += sizeof(struct term) + sizeof(Term) * arity;
    }
  }
  else {  /* must use multiple-arity list */
    Term t2;
    for (p = term_avail[MAX_TERM_AVAIL-1], t2 = NULL;
	 p && p->arity != arity;
	 t2 = p, p = p->args[0]);
    if (p != NULL) {
      term_avails--;
      if (t2 != NULL)
	t2->args[0] = p->args[0];
      else
	term_avail[MAX_TERM_AVAIL-1] = p->args[0];
    }
    else {
      p = tp_alloc(sizeof(struct term));
      p->args = tp_alloc(sizeof(Term) * arity);
      p->arity = arity;
      term_mem += sizeof(struct term) + sizeof(Term) * arity;
    }
  }
  
  p->private_symbol = 0;
  p->private_flags = 0;
  p->u.i = 0;
  p->container = NULL;
  return p;
}  /* get_term */

/*************
 *
 *    free_term()
 *
 *************/

/* DOCUMENTATION
This routine frees a term node only.  If you wish to recursively
free all of the subterms as well, call zap_term(t) instead.
*/

/* PUBLIC */
void free_term(Term p)
{
  term_frees++;
  term_avails++;
  if (p->arity == 0) {
    p->args = (struct term **) term_avail[0];
    term_avail[0] = p;
  }
  else if (p->arity < MAX_TERM_AVAIL-1) {
    p->args[0] = term_avail[p->arity];
    term_avail[p->arity] = p;
  }
  else {
    p->args[0] = term_avail[MAX_TERM_AVAIL-1];
    term_avail[MAX_TERM_AVAIL-1] = p;
  }
}  /* free_term */

/* end of get and free routines for each type */

/*************
 *
 *   fprint_term_mem()
 *
 *************/

/* DOCUMENTATION
This routine prints (to FILE *fp) memory usage statistics for Terms.
The Boolean argument heading tells whether to print a heading on the table.
*/

/* PUBLIC */
void fprint_term_mem(FILE *fp, BOOL heading)
{
  int n;
  
  if (heading)
    fprintf(fp, "  type (bytes each)        gets      frees     in use      avail      bytes\n");
  
  n = sizeof(struct term);
  fprintf(fp, "term (%4d)         %11u%11u%11u%11u%9.1f K\n",
	  n, term_gets, term_frees, term_gets - term_frees, term_avails,
	  term_mem / 1024.);
  
  /* end of printing for each type */
  
}  /* fprint_term_mem */

/*************
 *
 *   p_term_mem()
 *
 *************/

/* DOCUMENTATION
This routine prints memory usage statistics for Terms to stdout.
*/

/* PUBLIC */
void p_term_mem(void)
{
  fprint_term_mem(stdout, 1);
}  /* p__mem */

/*
 *  end of memory management
 */

/*************
 *
 *   get_variable_term()
 *
 *************/

/* DOCUMENTATION
This routine allocates and returns a term of type VARIABLE.
The index of the variable is set to var_num, which  should
be an integer >= 0.
*/

/* PUBLIC */
Term get_variable_term(int var_num)
{
  Term t = get_term(0);

  if (var_num < 0 || var_num > MAX_VAR)
    fatal_error("get_variable_term,  var_num out of range.");
  t->private_symbol = var_num;
  return t;
}  /* get_variable_term */

/*************
 *
 *   set_variable()
 *
 *************/

/* DOCUMENTATION
This routine takes a VARIABLE Term and sets the index to
var_num, which should be an integer >= 0.
*/

/* PUBLIC */
void set_variable(Term t, int var_num)
{
  if (var_num < 0 || var_num > MAX_VAR) {
    fatal_error("set_variable, var_num out of range.");
  }
  else {
    t->private_symbol = var_num;
  }
}  /* get_variable_term */

/*************
 *
 *   get_rigid_term_like()
 *
 *************/

/* DOCUMENTATION
This routine allocates and returns a term node with the same
symbol and arity as the given Term t.
*/

/* PUBLIC */
Term get_rigid_term_like(Term t)
{
  Term t1 = get_term(t->arity);
  t1->private_symbol = t->private_symbol;
  return t1;
}  /* get_rigid_term_like */

/*************
 *
 *   get_rigid_term()
 *
 *************/

/* DOCUMENTATION
This routine allocates and returns a term node with the given
symbol and arity.  If you already have a similar term node, say t,
(containing the symbol and arity you need) call get_rigid_term_like(t)
instead.
*/

/* PUBLIC */
Term get_rigid_term(char *sym, int arity)
{
  Term t1 = get_term(arity);
  t1->private_symbol = -str_to_sn(sym, arity);
  return t1;
}  /* get_rigid_term */

/*************
 *
 *   get_rigid_term_dangerously()
 *
 *************/

/* DOCUMENTATION
This routine can be used to allocate a term node if all you have is
the symbol ID and arity.  <I>If the arity is not correct
for the symbol ID, terrible things will happen!</I> 
<P>
If you have a similar term, use get_rigid_term_like() instead.
If you can afford the time to access the symbol table,
use sn_to_str() and get_rigid_term() instead.
*/

/* PUBLIC */
Term get_rigid_term_dangerously(int symnum, int arity)
{
  Term t1 = get_term(arity);
  t1->private_symbol = -symnum;
  return t1;
}  /* get_rigid_term_dangerously */

/*************
 *
 *    zap_term(term)
 *
 *    Free a term and all of its subterms.
 *
 *************/

/* DOCUMENTATION
This routine frees a term t and all of its subterms.  You should not
refer to t after calling zap_term(t).
*/

/* PUBLIC */
void zap_term(Term t)
{
  int i;
  for (i = 0; i < t->arity; i++)
    zap_term(t->args[i]);
  free_term(t);
}  /* zap_term */

/*************
 *
 *    int term_ident(term1, term2) -- Compare two terms.
 *
 *    If identical return 1); else return 0.  The bits
 *    field is not checked.
 *
 *************/

/* DOCUMENTATION
This function checks if two terms are identical.  Only the 
structure and symbols are checked---any extra fields such as
bits or u are NOT checked.
*/

/* PUBLIC */
BOOL term_ident(Term t1, Term t2)
{
  if (t1->private_symbol != t2->private_symbol)
    return 0;
  else {
    int i;
    for (i = 0; i < t1->arity; i++)
      if (!term_ident(t1->args[i], t2->args[i]))
	return 0;
    return 1;
  }
}  /* term_ident */  

/*************
 *
 *    Term copy_term(term) -- Return a copy of the term.
 *
 *    The bits field is not copied.
 *
 *************/

/* DOCUMENTATION
This routine copies a term.  Only the symbols and structure
are copied---any extra fields such as bits or u are
NOT copied.
*/

/* PUBLIC */
Term copy_term(Term t)
{
  if (VARIABLE(t))
    return get_variable_term(VARNUM(t));
  else {
    int i;
    Term t2 = get_rigid_term_like(t);
    for (i = 0; i < t->arity; i++)
      t2->args[i] = copy_term(t->args[i]);
    return t2;
  }
}  /* copy_term */

/*************
 *
 *    int ground_term(t) -- is a term ground?
 *
 *************/

/* DOCUMENTATION
This function checks if a term is ground, that is, has no variables.
*/

/* PUBLIC */
BOOL ground_term(Term t)
{
  if (VARIABLE(t))
    return FALSE;
  else {
    int i;
    for (i = 0; i < t->arity; i++)
      if (!ground_term(t->args[i]))
	return FALSE;
    return TRUE;
  }
}  /* ground_term */

/*************
 *
 *   biggest_variable()
 *
 *************/

/* DOCUMENTATION
This routine returns the greatest variable index of any variable int
the given term t.  If t is ground, -1 is returned.
*/

/* PUBLIC */
int biggest_variable(Term t)
{
  if (VARIABLE(t))
    return VARNUM(t);
  else {
    int i, max, v;
    for (max = -1, i = 0; i < t->arity; i++) {
      v = biggest_variable(t->args[i]);
      max = (v > max ? v : max);
    }
    return max;
  }
}  /* biggest_variable */

/*************
 *
 *    symbol_count
 *
 *************/

/* DOCUMENTATION
This routine returns the total number of symbols (i.e., the number of
nodes) in the given term t.
*/

/* PUBLIC */
int symbol_count(Term t)
{
  int i;
  int count = 0;
  for (i = 0; i < ARITY(t); i++)
    count += symbol_count(ARG(t,i));
  return count+1;
}  /* symbol_count  */

/*************
 *
 *     int occurs_in(t1, t2) -- Does t1 occur in t2 (including t1==t2)?
 *
 *     term_ident is used to check identity.
 *
 *************/

/* DOCUMENTATION
This function checks if Term t2 is identical to a subterm of Term t1,
including the case term_ident(t1,t2).  All identity checks are done
with term_ident(), so extra fields such as bits or u are not
checked.
*/

/* PUBLIC */
BOOL occurs_in(Term t1, Term t2)
{
  if (term_ident(t1, t2))
    return TRUE;
  else {
    int i;
    for (i = 0; i < t2->arity; i++)
      if (occurs_in(t1, t2->args[i]))
	return TRUE;
    return FALSE;
  }
}  /* occurs_in */

/*************
 *
 *  fprint_term(fp, t)
 *
 *************/

/* DOCUMENTATION
This routine prints (to FILE *fp) a term.  A newline is NOT printed.
*/

/* PUBLIC */
void fprint_term(FILE *fp, Term t)
{
  if (VARIABLE(t))
    fprintf(fp, "v%d", VARNUM(t));
  else {
    fprint_sym(fp, SYMNUM(t));
    if (COMPLEX(t)) {
      int i;
      fprintf(fp, "(");
      for (i = 0; i < ARITY(t); i++) {
	fprint_term(fp, ARG(t,i));
	if (i < ARITY(t)-1)
	  fprintf(fp, ",");
      }
      fprintf(fp, ")");
    }
  }
  fflush(fp);
}  /* print_term */

/*************
 *
 *  sprint_term(sb, t)
 *
 *************/

/* DOCUMENTATION
This (recursive) routine appends the string representation of a term to
a String_buf.  A newline is not included.
*/

/* PUBLIC */
void sprint_term(String_buf sb, Term t)
{
  if (VARIABLE(t)) {
    char s[MAX_NAME];
    sprintf(s, "v%d", VARNUM(t));
    sb_append(sb, s);
  }
  else {
    sprint_sym(sb, SYMNUM(t));
    if (COMPLEX(t)) {
      int i;
      sb_append(sb, "(");
      for (i = 0; i < t->arity; i++) {
	sprint_term(sb, t->args[i]);
	if (i < t->arity-1)
	  sb_append(sb, ",");
      }
      sb_append(sb, ")");
    }
  }
}  /* sprint_term */

/*************
 *
 *  p_term(t)
 *
 *************/

/* DOCUMENTATION
This routine prints a term, followed by '\n' and fflush, to stdout.
If you don't want the newline, use fprint_term() instead.
If you want the term put into a string, use sprint_term() instead.
*/

/* PUBLIC */
void p_term(Term t)
{
  fprint_term(stdout, t);
  printf("\n");
  fflush(stdout);
}  /* p_term */

/*************
 *
 *  all_args_vars(t)
 *
 *************/

/* DOCUMENTATION
This Boolean routine checks if all argumets of Term t are VARIABLEs.
(It is true also if t is a VARIABLE.)
*/

/* PUBLIC */
BOOL all_args_vars(Term t)
{
  if (VARIABLE(t))
    return TRUE;
  else {
    int i;
    for (i = 0; i < t->arity; i++)
      if (!VARIABLE(t->args[i]))
	return FALSE;
    return TRUE;
  }
}  /* all_args_vars */

/*************
 *
 *    build_binary_term()
 *
 *************/

/* DOCUMENTATION
Build and return a binary term with SYMNUM sn, first term a1, and
second term a2.
<P>
WARNING: if sn is not a binary symbol, bad things will happen!
*/

/* PUBLIC */
Term build_binary_term(int sn, Term a1, Term a2)
{
    Term t = get_rigid_term_dangerously(sn, 2);
    t->args[0] = a1;
    t->args[1] = a2;
    return(t);
}  /* build_binary_term */

/*************
 *
 *    build_unary_term()
 *
 *************/

/* DOCUMENTATION
Build and return a unary term with SYMNUM sn and argument term a.
<P>
WARNING: if sn is not a unary symbol, bad things will happen!
*/

/* PUBLIC */
Term build_unary_term(int sn, Term a)
{
    Term t = get_rigid_term_dangerously(sn, 1);
    t->args[0] = a;
    return(t);
}  /* build_unary_term */

/*************
 *
 *   subst_term()
 *
 *************/

/* DOCUMENTATION
In term t, replace all occurrences of Term target with <I>copies of</I>
Term replacement.  Free all of the replaced terms;
*/

/* PUBLIC */
Term subst_term(Term t, Term target, Term replacement)
{
  if (term_ident(t, target)) {
    zap_term(t);
    return copy_term(replacement);
  }
  else {
    int i;
    for (i = 0; i < t->arity; i++)
      t->args[i] = subst_term(t->args[i], target, replacement);
    return t;
  }
}  /* subst_term */

/*************
 *
 *   subst_var_term()
 *
 *************/

/* DOCUMENTATION
In Term t, replace all CONSTANT terms containing SYMNUM symnum
with a variable containing VARNUM varnum.  Free the replaced constants
and return the result.
*/

/* PUBLIC */
Term subst_var_term(Term t, int symnum, int varnum)
{
  if (CONSTANT(t) && SYMNUM(t) == symnum) {
    Term v = get_variable_term(varnum);
    zap_term(t);
    return v;
  }
  else {
    int i;
    for (i = 0; i < t->arity; i++)
      t->args[i] = subst_var_term(t->args[i], symnum, varnum);
    return t;
  }
}  /* subst_var_term */

/*************
 *
 *   greatest_variable()
 *
 *************/

/* DOCUMENTATION
This routine returns the greatest variable index in a term.
If the term is ground, -1 is returned.
*/

/* PUBLIC */
int greatest_variable(Term t)
{
  if (VARIABLE(t))
    return VARNUM(t);
  else {
    int i, max, v;
    for (max = -1, i = 0; i < t->arity; i++) {
      v = greatest_variable(t->args[i]);
      max = (v > max ? v : max);
    }
    return max;
  }
}  /* greatest_variable */

/*************
 *
 *   greatest_symnum_in_term()
 *
 *************/

/* DOCUMENTATION
This function returns the greatest SYMNUM (of a CONSTANT or COMPLEX term)
in the given Term t.
If the term is a VARIABLE, return -1.
*/

/* PUBLIC */
int greatest_symnum_in_term(Term t)
{
  if (VARIABLE(t))
    return -1;
  else {
    int max = SYMNUM(t);
    int i;
    for (i = 0; i < t->arity; i++) {
      int sm = greatest_symnum_in_term(t->args[i]);
      max = (sm > max ? sm : max);
    }
    return max;
  }
}  /* greatest_symnum_in_term */

/*************
 *
 *   upward_term_links()
 *
 *************/

/* DOCUMENTATION
In the given Term t, make the "container" field of t and each subterm
point to (void *) p.
*/

/* PUBLIC */
void upward_term_links(Term t, void *p)
{
  int i;
  t->container = p;
  for (i = 0; i < t->arity; i++)
    upward_term_links(t->args[i], p);
}  /* upward_term_links */

/*************
 *
 *   occurrences()
 *
 *************/

/* DOCUMENTATION
This function returns the number of occurrences of Term target in Term t.
The checks are made with term_ident().
*/

/* PUBLIC */
int occurrences(Term t, Term target)
{
  if (term_ident(t, target))
    return 1;
  else {
    int n = 0;
    int i;
    for (i = 0; i < t->arity; i++)
      n += occurrences(t->args[i], target);
    return n;
  }
}  /* occurrences */

/*************
 *
 *   trm_set_vars_recurse()
 *
 *   There might be another (static) copy of this routine in clause.c.
 *
 *************/

static
Term trm_set_vars_recurse(Term t, char **varnames, int max_vars, BOOL *rcp)
{
  if (CONSTANT(t)) {
    char *name = sn_to_str(SYMNUM(t));
    if (variable_name(name)) {
      int i = 0;
      while (i < max_vars && varnames[i] != NULL && varnames[i] != name)
	i++;
      if (i == max_vars) 
	*rcp = FALSE;
      else {
	if (varnames[i] == NULL)
	  varnames[i] = name;
	free_term(t);
	t = get_variable_term(i);
      }
    }
  }
  else {
    int i;
    for (i = 0; i < t->arity && *rcp; i++)
      t->args[i] = trm_set_vars_recurse(t->args[i], varnames, max_vars, rcp);
  }
  return t;
}  /* trm_set_vars_recurse */

/*************
 *
 *   term_set_variables()
 *
 *************/

/* DOCUMENTATION
This routine traverses a term and changes the constants
that should be variables, into variables.  On input, the term
should have no variables.  The new variables are numbered
0, 1, 2 ... according the the first occurrence, reading from the
left.
<P>
If there are not more than max_vars distinct variables,
1 is returned;  if there are more than max_vars variables,
the transformation is not completed, and 0 is returned.
<P>
<I>If you are dealing with clauses, use clause_set_variables()
instead.</I>
*/

#define CSTACK_ARRAY_SIZE    100

/* PUBLIC */
BOOL term_set_variables(Term t, int max_vars)
{
  char *a[CSTACK_ARRAY_SIZE], **vmap;
  int i;
  BOOL ok = TRUE;

  if (max_vars > CSTACK_ARRAY_SIZE)
    vmap = my_malloc((max_vars * sizeof(char *)));
  else
    vmap = a;

  for (i = 0; i < max_vars; i++)
    vmap[i] = NULL;

  for (i = 0; i < t->arity && ok; i++) 
    t->args[i] = trm_set_vars_recurse(t->args[i], vmap, max_vars, &ok);

  if (max_vars > CSTACK_ARRAY_SIZE)
    my_free(vmap);
  return ok;
}  /* term_set_variables */

/*************
 *
 *   nat_to_term()
 *
 *************/

/* DOCUMENTATION
This routine takes a nonnegative integer and returns
a constant Term with the string representation of the
integer as the constant symbol.
*/

/* PUBLIC */
Term nat_to_term(int n)
{
  char s[100];
  if (n < 0)
    fatal_error("nat_to_term: negative term");
  itoa(n, s);
  return get_rigid_term(s, 0);
}  /* nat_to_term */

/*************
 *
 *   natural_constant(t)
 *
 *************/

/* DOCUMENTATION
This routine takes a term, and if the term represents
an nonnegative integer, that integer is returned;
otherwise, -1 is returned.
*/

/* PUBLIC */
int natural_constant(Term t)
{
  if (!CONSTANT(t))
    return -1;
  else {
    int i;
    if (!str_to_int(sn_to_str(SYMNUM(t)), &i))
      return -1;
    else if (i < 0)
      return -1;
    else
      return i;
  }
}  /* natural_constant */

/*************
 *
 *   arg_position()
 *
 *************/

/* DOCUMENTATION
If the given terms are in a parent-child relatioship,
return the argument position (index) of the child.
Otherwise, return -1.
*/

/* PUBLIC */
int arg_position(Term parent, Term child)
{
  int i;
  for (i = 0; i < ARITY(parent); i++) {
    if (ARG(parent,i) == child)
      return i;
  }
  return -1;
}  /* arg_position */
